<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2011 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/lib/utf8lib.php - utility-routines for UTF-8
 *
 * This file deals with the idiosyncrasies of UTF-8.
 *
 * Reference:
 * The Unicode Consortium, The Unicode Standard, Version 6.0.0,
 * (Mountain View, CA: The Unicode Consortium, 2011, ISBN 978-1-936213-01-6)
 * <http://www.unicode.org/versions/Unicode6.0.0>, chapter 3 (Conformance), page 94
 *
 * Summary of the way valid code points are stored in 1 to 4 byte sequences.
 *
 * bits   code-points                     1st byte   2nd byte   3rd byte   4th byte
 * ----   -----------------------------   ---------  ---------  ---------  ---------
 *    7   0000 0000 0000 0000 0xxx xxxx   0xxx.xxxx
 *   11   0000 0000 0000 0yyy yyxx xxxx   110y.yyyy  10xx.xxxx
 *   16   0000 0000 zzzz yyyy yyxx xxxx   1110.zzzz  10yy.yyyy  10xx.xxxx
 *   21   000w wwzz zzzz yyyy yyxx xxxx   1111.0www  10zz.zzzz  10yy.yyyy  10xx.xxxx
 *
 * bits    byte 1   byte 2   byte 3   byte 4  comments
 * ----   -------  -------  -------  -------  --------
 *    7   00 - 7F                             U+0000 - U+007F
 *        80 - BF                             --> ill-formed
 *        C0 - C1  80 - BF                    --> overlong 2-byte
 *   11   C2 - DF  80 - BF                    U+0080 - U+07FF
 *        E0       80 - 9F* 80 - BF           --> overlong 3-byte
 *   16   E0       A0 - BF* 80 - BF           U+0800 - U+0FFF
 *   16   E1 - EC  80 - BF  80 - BF           U+1000 - U+CFFF
 *   16   ED       80 - 9F* 80 - BF           U+D000 - U+D7FF
 *        ED       A0 - BF* 80 - BF           --> surrogates (U+D800 - U+DFFF)
 *   16   EE - EF  80 - BF  80 - BF           U+E000 -   U+FFFF
 *        F0       80 - 8F* 80 - BF  80 - BF  --> overlong 4-byte
 *   21   F0       90 - BF* 80 - BF  80 - BF  U+10000 - U+3FFFF
 *   21   F1 - F3  80 - BF  80 - BF  80 - BF  U+40000 - U+FFFFF
 *   21   F4       80 - 8F* 80 - BF  80 - BF  U+100000 - U+10FFFF
 *        F4       90 - BF* 80 - BF  80 - BF  --> invalid planes 11 - 13
 *        F5 - F7  80 - BF  80 - BF  80 - BF  --> invalid planes 14 - 1F
 *        F8 - FB                             --> invalid 5-byte sequence
 *        FC - FD                             --> invalid 6-byte sequence
 *        FE - FF                             --> disallowed (BOM)
 *
 * Note:
 * Non-standard continuation ranges are marked with * (only for byte 2)
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2011 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wascore
 * @version $Id: utf8lib.php,v 1.2 2011-05-06 16:07:56 pfokker Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

if (!defined('USE_MBSTRING')) {
    if (defined('MB_CASE_LOWER')) {
        define('USE_MBSTRING',TRUE);
        mb_internal_encoding('UTF-8');
    } else {
        define('USE_MBSTRING',FALSE);
    }
}


/** check an arbitrary string for UTF-8 conformity
 *
 * return TRUE if string is valid UTF-8, FALSE otherwise.
 *
 * @param string $str the string to check
 * @return bool TRUE if valid UTF-8, FALSE otherwise
 */
function utf8_validate($str) {
    $pattern = '/^([\\x00-\\x7F]'.                         // ASCII (including ctrl-chars)
               '|[\\xC2-\\xDF][\\x80-\\xBF]'.              // non-overlong 2-byte
               '|\\xE0[\\xA0-\\xBF][\\x80-\\xBF]'.         // 3-byte excluding overlongs
               '|[\\xE1-\\xEC\\xEE\\xEF][\\x80-\\xBF]{2}'. // 3-byte (plain)
               '|\\xED[\\x80-\\x9F][\\x80-\\xBF]'.         // 3-byte excluding surrogates
               '|\\xF0[\\x90-\\xBF][\\x80-\\xBF]{2}'.      // 4-byte excluding overlongs
               '|[\\xF1-\\xF3][\\x80-\\xBF]{3}'.           // 4-byte planes 4-15
               '|\\xF4[\\x80-\\x8F][\\x80-\\xBF]{2})*$/';  // 4-byte plane 16
    return (preg_match($pattern,$str) === 1);
} // utf8_validate()


/** calculate the number of code points encoded in an UTF-8 string
 *
 * This routine uses a trick to calculate the number of code points
 * available in string $str. By first converting the UTF-8 string to ISO-8859-1
 * all multi-byte code points (ie. all non-ASCII) are converted to a single
 * byte: the correct ISO-8859-1 character where possible and a '?' for
 * characters not available in the ISO-8859-1 reportoire.
 * Note that utf8_decode() does NOT work very well on ill-formed UTF-8 strings,
 * e.g. it happily interprets codes in invalid planes and translates truncated
 * sequences to chr(0). The good news is that it works for valid UTF-8.
 *
 * @param string $utf8str a valid UTF-8 string to examine
 * @return int the number of code points in string $str
 */
function utf8_strlen($utf8str) {
    return strlen(utf8_decode($utf8str));
} // utf8_strlen()


/** fold a UTF-8 string to lower case
 *
 * this routine tries to use the multibyte routine for folding,
 * but if it is not available we fall back to our own translation
 * table $UTF8_UPPER_LOWER which is derived straight from the
 * Unicode Character Database.
 *
 * Informal benchmarking yielded no significant speed difference
 * between mb_strtolower() and strtr(). However, our table needs
 * memory to store the 1100+ pairs and for mb_strtolower() this is
 * already taken care of. OTOH: depending on the version of mb_string,
 * our table (UCD 6.0.0 February 2011) may be more complete and up
 * to date. Oh well.
 *
 * @param string $utf8str a valid UTF-8 string to examine
 * @return string the lowercase equivalent of $utf8str
 */
function utf8_strtolower($utf8str) {
    if (USE_MBSTRING) {
        return mb_strtolower($utf8str,'UTF-8');
    } else {
        global $UTF8_UPPER_LOWER;
        return strtr($utf8str,$UTF8_UPPER_LOWER);
    }
} // utf8_strtolower()


/** fold a UTF-8 string to upper case (sort of)
 *
 * this routine tries to use the multibyte routine for folding,
 * but if it is not available we fall back to our own translation
 * table $UTF8_LOWER_UPPER which is derived from the
 * Unicode Character Database. Note that this table has some
 * quirks because the underlying table is injective.
 *
 * @param string $utf8str a valid UTF-8 string to examine
 * @return string the uppercase equivalent of $utf8str
 */
function utf8_strtoupper($utf8str) {
    if (USE_MBSTRING) {
        return mb_strtoupper($utf8str,'UTF-8');
    } else {
        global $UTF8_LOWER_UPPER;
        return strtr($utf8str,$UTF8_LOWER_UPPER);
    }
} // utf8_strtoupper()


/** compare two UTF8 strings in a case-INsensitive way
 *
 * This compares two UTF-8 strings caseINsensitive. We
 * do this by comparing the lowercase variant of the strings
 * with the regular strcmp. We use lowercasing because that
 * translation table is 'better' than the upper case one.
 *
 * Note that this is a quick and dirty approach: we simply
 * use the multibyte strings as-is rather than extracting
 * the actual code points and applying some sort of collation
 * because that would make it much more complicated.
 *
 * @param string $utf8str1 first string
 * @param string $utf8str2 second string
 * @return int result < 0 if string 1 < string 2, result > 0 if string 1 > string 2, 0 when equal
 */
function utf8_strcasecmp($utf8str1,$utf8str2) {
    return strcmp(utf8_strtolower($utf8str1),utf8_strtolower($utf8str2));
} // utf8_strcasecmp()


/** return part of a UTF-8 string
 *
 * this routine returns a valid UTF-8 substring from $utf8str
 * based on the $start and $length parameters in a way comparable
 * to substr(). If available, we use the mb_substr() replacement,
 * otherwise we perform the requested actions ourselves.
 *
 * If we do it ourselves, we prefix variable names with 'c_' for characters
 * and 'b_' for bytes. A UTF-8 character takes up to 4 bytes.
 *
 * @param string $utf8str a valid UTF-8 string to examine
 * @param int $start an offset expressed in characters (not bytes)
 * @param int $length the length of the string to return (also expressed in characters)
 * @return string the requested substring of $utf8str or FALSE if $start points beyond the string
 */
function utf8_substr($utf8str,$start,$length=NULL) {
    if (USE_MBSTRING) {
        return mb_substr($utf8str,$start,(is_null($length)) ? utf8_strlen($utf8str) : $length,'UTF-8');
    }

    $c_strlen = utf8_strlen($utf8str);
    if ($start >= $c_strlen) {
        return FALSE;
    }
    // convert the negative offset from end of string into a non-negative offset from start of string
    $c_start = ($start < 0) ? max($c_strlen + $start, 0) : $start;

    // calculate the byte offset equivalent with the requested character offset
    $b_start = 0;     
    for ($c_index = 0; $c_index < $c_start; ++$c_index) {
        $byte = ord($utf8str[$b_start++]);
        if (($byte & 0x80) == 0x00) {       // 0xxx.xxxx => plain ascii; we're done here
            continue;
        } elseif (($byte & 0xE0) == 0xC0) { // 110x.xxxx => start of 2-byte sequence
            ++$b_start;
        } elseif (($byte & 0xF0) == 0xE0) { // 1110.xxxx => start of 3-byte sequence
            $b_start += 2;
        } elseif (($byte & 0xF8) == 0xF0) { // 1111.0xxx => start of 4-byte sequence
            $b_start += 3;
        } // else invalid byte shouldn't happen with valid UTF-8
    }
    // If no length was specified, they want the part from $b_start to end of string so we're done here
    if (is_null($length)) {
        return substr($utf8str,$b_start);
    }
    // convert the negative length into a non-negative offset from start of string
    $c_end = ($length < 0) ? max($c_strlen + $length,0) : min($c_start + $length,$c_strlen);

    // continue counting characters from where we were
    for ($b_end = $b_start; $c_index < $c_end; ++$c_index) {
        $byte = ord($utf8str[$b_end++]);
        if (($byte & 0x80) == 0x00) {       // 0xxx.xxxx => plain ascii; we're done here
            continue;
        } elseif (($byte & 0xE0) == 0xC0) { // 110x.xxxx => start of 2-byte sequence
            ++$b_end;
        } elseif (($byte & 0xF0) == 0xE0) { // 1110.xxxx => start of 3-byte sequence
            $b_end += 2;
        } elseif (($byte & 0xF8) == 0xF0) { // 1111.0xxx => start of 4-byte sequence
            $b_end += 3;
        } // else invalid byte (shouldn't happen with valid UTF-8)
    }
    return substr($utf8str,$b_start,$b_end-$b_start);
} // utf8_substr()


/** map some UTF-8 characters to comparable ASCII strings
 *
 * this maps a lot of UTF-8 characters to more or less comparable
 * ASCII strings, e.g. an A-acute is mapped to 'A', etc.
 * Handy when trying to construct readable filenames from
 * letters with diacritics, eg. e-acute 'l' e-grave 'v' 'e' (French
 * for pupil) maps to 'e' 'l' 'e' 'v' e' rather than 'l' 'v' 'e'
 * if the diacriticals would simply be 'eaten'.
 *
 * Note that UTF-8 characters that are NOT mapped to ASCII are
 * retained, i.e. the result is not plain ASCII but an UTF-8 string
 * with as much characters mapped to ASCII as possible.
 *
 * @param string $utf8str input string possibly with letters with diacriticals etc.
 * @return string a 'best-effort' ASCII-approximation of $utf8str
 * @uses $UTF8_ASCII
 */
function utf8_strtoascii($utf8str) {
    global $UTF8_ASCII;
    return strtr($utf8str,$UTF8_ASCII);
} // utf8_strtoascii()


/** UTF-8 mapping to ASCII
 *
 * This array derived from UnicodeData.txt
 * (part of the Unicode Character Database).
 * It maps UTF-8 encoded characters to the
 * comparable ASCII characters without diacriticals.
 * This is useful to keep (file)names with say
 * A-acute readable by mapping that character to 'A'.
 *
 */
global $UTF8_ASCII;
if (empty($UTF8_ASCII)) { $UTF8_ASCII = array(
  "\xC2\xAA"=>'a',"\xC2\xB2"=>'2',"\xC2\xB3"=>'3',"\xC2\xB9"=>'1',"\xC2\xBA"=>'o',"\xC2\xBC"=>'14',"\xC2\xBD"=>'12',
  "\xC2\xBE"=>'34',"\xC3\x80"=>'A',"\xC3\x81"=>'A',"\xC3\x82"=>'A',"\xC3\x83"=>'A',"\xC3\x84"=>'A',"\xC3\x85"=>'A',
  "\xC3\x87"=>'C',"\xC3\x88"=>'E',"\xC3\x89"=>'E',"\xC3\x8A"=>'E',"\xC3\x8B"=>'E',"\xC3\x8C"=>'I',"\xC3\x8D"=>'I',
  "\xC3\x8E"=>'I',"\xC3\x8F"=>'I',"\xC3\x91"=>'N',"\xC3\x92"=>'O',"\xC3\x93"=>'O',"\xC3\x94"=>'O',"\xC3\x95"=>'O',
  "\xC3\x96"=>'O',"\xC3\x99"=>'U',"\xC3\x9A"=>'U',"\xC3\x9B"=>'U',"\xC3\x9C"=>'U',"\xC3\x9D"=>'Y',"\xC3\xA0"=>'a',
  "\xC3\xA1"=>'a',"\xC3\xA2"=>'a',"\xC3\xA3"=>'a',"\xC3\xA4"=>'a',"\xC3\xA5"=>'a',"\xC3\xA7"=>'c',"\xC3\xA8"=>'e',
  "\xC3\xA9"=>'e',"\xC3\xAA"=>'e',"\xC3\xAB"=>'e',"\xC3\xAC"=>'i',"\xC3\xAD"=>'i',"\xC3\xAE"=>'i',"\xC3\xAF"=>'i',
  "\xC3\xB1"=>'n',"\xC3\xB2"=>'o',"\xC3\xB3"=>'o',"\xC3\xB4"=>'o',"\xC3\xB5"=>'o',"\xC3\xB6"=>'o',"\xC3\xB9"=>'u',
  "\xC3\xBA"=>'u',"\xC3\xBB"=>'u',"\xC3\xBC"=>'u',"\xC3\xBD"=>'y',"\xC3\xBF"=>'y',"\xC4\x80"=>'A',"\xC4\x81"=>'a',
  "\xC4\x82"=>'A',"\xC4\x83"=>'a',"\xC4\x84"=>'A',"\xC4\x85"=>'a',"\xC4\x86"=>'C',"\xC4\x87"=>'c',"\xC4\x88"=>'C',
  "\xC4\x89"=>'c',"\xC4\x8A"=>'C',"\xC4\x8B"=>'c',"\xC4\x8C"=>'C',"\xC4\x8D"=>'c',"\xC4\x8E"=>'D',"\xC4\x8F"=>'d',
  "\xC4\x92"=>'E',"\xC4\x93"=>'e',"\xC4\x94"=>'E',"\xC4\x95"=>'e',"\xC4\x96"=>'E',"\xC4\x97"=>'e',"\xC4\x98"=>'E',
  "\xC4\x99"=>'e',"\xC4\x9A"=>'E',"\xC4\x9B"=>'e',"\xC4\x9C"=>'G',"\xC4\x9D"=>'g',"\xC4\x9E"=>'G',"\xC4\x9F"=>'g',
  "\xC4\xA0"=>'G',"\xC4\xA1"=>'g',"\xC4\xA2"=>'G',"\xC4\xA3"=>'g',"\xC4\xA4"=>'H',"\xC4\xA5"=>'h',"\xC4\xA8"=>'I',
  "\xC4\xA9"=>'i',"\xC4\xAA"=>'I',"\xC4\xAB"=>'i',"\xC4\xAC"=>'I',"\xC4\xAD"=>'i',"\xC4\xAE"=>'I',"\xC4\xAF"=>'i',
  "\xC4\xB0"=>'I',"\xC4\xB2"=>'IJ',"\xC4\xB3"=>'ij',"\xC4\xB4"=>'J',"\xC4\xB5"=>'j',"\xC4\xB6"=>'K',"\xC4\xB7"=>'k',
  "\xC4\xB9"=>'L',"\xC4\xBA"=>'l',"\xC4\xBB"=>'L',"\xC4\xBC"=>'l',"\xC4\xBD"=>'L',"\xC4\xBE"=>'l',"\xC4\xBF"=>'L',
  "\xC5\x80"=>'l',"\xC5\x83"=>'N',"\xC5\x84"=>'n',"\xC5\x85"=>'N',"\xC5\x86"=>'n',"\xC5\x87"=>'N',"\xC5\x88"=>'n',
  "\xC5\x89"=>'n',"\xC5\x8C"=>'O',"\xC5\x8D"=>'o',"\xC5\x8E"=>'O',"\xC5\x8F"=>'o',"\xC5\x90"=>'O',"\xC5\x91"=>'o',
  "\xC5\x94"=>'R',"\xC5\x95"=>'r',"\xC5\x96"=>'R',"\xC5\x97"=>'r',"\xC5\x98"=>'R',"\xC5\x99"=>'r',"\xC5\x9A"=>'S',
  "\xC5\x9B"=>'s',"\xC5\x9C"=>'S',"\xC5\x9D"=>'s',"\xC5\x9E"=>'S',"\xC5\x9F"=>'s',"\xC5\xA0"=>'S',"\xC5\xA1"=>'s',
  "\xC5\xA2"=>'T',"\xC5\xA3"=>'t',"\xC5\xA4"=>'T',"\xC5\xA5"=>'t',"\xC5\xA8"=>'U',"\xC5\xA9"=>'u',"\xC5\xAA"=>'U',
  "\xC5\xAB"=>'u',"\xC5\xAC"=>'U',"\xC5\xAD"=>'u',"\xC5\xAE"=>'U',"\xC5\xAF"=>'u',"\xC5\xB0"=>'U',"\xC5\xB1"=>'u',
  "\xC5\xB2"=>'U',"\xC5\xB3"=>'u',"\xC5\xB4"=>'W',"\xC5\xB5"=>'w',"\xC5\xB6"=>'Y',"\xC5\xB7"=>'y',"\xC5\xB8"=>'Y',
  "\xC5\xB9"=>'Z',"\xC5\xBA"=>'z',"\xC5\xBB"=>'Z',"\xC5\xBC"=>'z',"\xC5\xBD"=>'Z',"\xC5\xBE"=>'z',"\xC5\xBF"=>'s',
  "\xC6\xA0"=>'O',"\xC6\xA1"=>'o',"\xC6\xAF"=>'U',"\xC6\xB0"=>'u',"\xC7\x84"=>'DZ',"\xC7\x85"=>'Dz',"\xC7\x86"=>'dz',
  "\xC7\x87"=>'LJ',"\xC7\x88"=>'Lj',"\xC7\x89"=>'lj',"\xC7\x8A"=>'NJ',"\xC7\x8B"=>'Nj',"\xC7\x8C"=>'nj',"\xC7\x8D"=>'A',
  "\xC7\x8E"=>'a',"\xC7\x8F"=>'I',"\xC7\x90"=>'i',"\xC7\x91"=>'O',"\xC7\x92"=>'o',"\xC7\x93"=>'U',"\xC7\x94"=>'u',
  "\xC7\x95"=>'U',"\xC7\x96"=>'u',"\xC7\x97"=>'U',"\xC7\x98"=>'u',"\xC7\x99"=>'U',"\xC7\x9A"=>'u',"\xC7\x9B"=>'U',
  "\xC7\x9C"=>'u',"\xC7\x9E"=>'A',"\xC7\x9F"=>'a',"\xC7\xA0"=>'A',"\xC7\xA1"=>'a',"\xC7\xA6"=>'G',"\xC7\xA7"=>'g',
  "\xC7\xA8"=>'K',"\xC7\xA9"=>'k',"\xC7\xAA"=>'O',"\xC7\xAB"=>'o',"\xC7\xAC"=>'O',"\xC7\xAD"=>'o',"\xC7\xB0"=>'j',
  "\xC7\xB1"=>'DZ',"\xC7\xB2"=>'Dz',"\xC7\xB3"=>'dz',"\xC7\xB4"=>'G',"\xC7\xB5"=>'g',"\xC7\xB8"=>'N',"\xC7\xB9"=>'n',
  "\xC7\xBA"=>'A',"\xC7\xBB"=>'a',"\xC8\x80"=>'A',"\xC8\x81"=>'a',"\xC8\x82"=>'A',"\xC8\x83"=>'a',"\xC8\x84"=>'E',
  "\xC8\x85"=>'e',"\xC8\x86"=>'E',"\xC8\x87"=>'e',"\xC8\x88"=>'I',"\xC8\x89"=>'i',"\xC8\x8A"=>'I',"\xC8\x8B"=>'i',
  "\xC8\x8C"=>'O',"\xC8\x8D"=>'o',"\xC8\x8E"=>'O',"\xC8\x8F"=>'o',"\xC8\x90"=>'R',"\xC8\x91"=>'r',"\xC8\x92"=>'R',
  "\xC8\x93"=>'r',"\xC8\x94"=>'U',"\xC8\x95"=>'u',"\xC8\x96"=>'U',"\xC8\x97"=>'u',"\xC8\x98"=>'S',"\xC8\x99"=>'s',
  "\xC8\x9A"=>'T',"\xC8\x9B"=>'t',"\xC8\x9E"=>'H',"\xC8\x9F"=>'h',"\xC8\xA6"=>'A',"\xC8\xA7"=>'a',"\xC8\xA8"=>'E',
  "\xC8\xA9"=>'e',"\xC8\xAA"=>'O',"\xC8\xAB"=>'o',"\xC8\xAC"=>'O',"\xC8\xAD"=>'o',"\xC8\xAE"=>'O',"\xC8\xAF"=>'o',
  "\xC8\xB0"=>'O',"\xC8\xB1"=>'o',"\xC8\xB2"=>'Y',"\xC8\xB3"=>'y',"\xCA\xB0"=>'h',"\xCA\xB2"=>'j',"\xCA\xB3"=>'r',
  "\xCA\xB7"=>'w',"\xCA\xB8"=>'y',"\xCB\xA1"=>'l',"\xCB\xA2"=>'s',"\xCB\xA3"=>'x',"\xCD\xBE"=>';',"\xE1\xB4\xAC"=>'A',
  "\xE1\xB4\xAE"=>'B',"\xE1\xB4\xB0"=>'D',"\xE1\xB4\xB1"=>'E',"\xE1\xB4\xB3"=>'G',"\xE1\xB4\xB4"=>'H',
  "\xE1\xB4\xB5"=>'I',"\xE1\xB4\xB6"=>'J',"\xE1\xB4\xB7"=>'K',"\xE1\xB4\xB8"=>'L',"\xE1\xB4\xB9"=>'M',
  "\xE1\xB4\xBA"=>'N',"\xE1\xB4\xBC"=>'O',"\xE1\xB4\xBE"=>'P',"\xE1\xB4\xBF"=>'R',"\xE1\xB5\x80"=>'T',
  "\xE1\xB5\x81"=>'U',"\xE1\xB5\x82"=>'W',"\xE1\xB5\x83"=>'a',"\xE1\xB5\x87"=>'b',"\xE1\xB5\x88"=>'d',
  "\xE1\xB5\x89"=>'e',"\xE1\xB5\x8D"=>'g',"\xE1\xB5\x8F"=>'k',"\xE1\xB5\x90"=>'m',"\xE1\xB5\x92"=>'o',
  "\xE1\xB5\x96"=>'p',"\xE1\xB5\x97"=>'t',"\xE1\xB5\x98"=>'u',"\xE1\xB5\x9B"=>'v',"\xE1\xB5\xA2"=>'i',
  "\xE1\xB5\xA3"=>'r',"\xE1\xB5\xA4"=>'u',"\xE1\xB5\xA5"=>'v',"\xE1\xB6\x9C"=>'c',"\xE1\xB6\xA0"=>'f',
  "\xE1\xB6\xBB"=>'z',"\xE1\xB8\x80"=>'A',"\xE1\xB8\x81"=>'a',"\xE1\xB8\x82"=>'B',"\xE1\xB8\x83"=>'b',
  "\xE1\xB8\x84"=>'B',"\xE1\xB8\x85"=>'b',"\xE1\xB8\x86"=>'B',"\xE1\xB8\x87"=>'b',"\xE1\xB8\x88"=>'C',
  "\xE1\xB8\x89"=>'c',"\xE1\xB8\x8A"=>'D',"\xE1\xB8\x8B"=>'d',"\xE1\xB8\x8C"=>'D',"\xE1\xB8\x8D"=>'d',
  "\xE1\xB8\x8E"=>'D',"\xE1\xB8\x8F"=>'d',"\xE1\xB8\x90"=>'D',"\xE1\xB8\x91"=>'d',"\xE1\xB8\x92"=>'D',
  "\xE1\xB8\x93"=>'d',"\xE1\xB8\x94"=>'E',"\xE1\xB8\x95"=>'e',"\xE1\xB8\x96"=>'E',"\xE1\xB8\x97"=>'e',
  "\xE1\xB8\x98"=>'E',"\xE1\xB8\x99"=>'e',"\xE1\xB8\x9A"=>'E',"\xE1\xB8\x9B"=>'e',"\xE1\xB8\x9C"=>'E',
  "\xE1\xB8\x9D"=>'e',"\xE1\xB8\x9E"=>'F',"\xE1\xB8\x9F"=>'f',"\xE1\xB8\xA0"=>'G',"\xE1\xB8\xA1"=>'g',
  "\xE1\xB8\xA2"=>'H',"\xE1\xB8\xA3"=>'h',"\xE1\xB8\xA4"=>'H',"\xE1\xB8\xA5"=>'h',"\xE1\xB8\xA6"=>'H',
  "\xE1\xB8\xA7"=>'h',"\xE1\xB8\xA8"=>'H',"\xE1\xB8\xA9"=>'h',"\xE1\xB8\xAA"=>'H',"\xE1\xB8\xAB"=>'h',
  "\xE1\xB8\xAC"=>'I',"\xE1\xB8\xAD"=>'i',"\xE1\xB8\xAE"=>'I',"\xE1\xB8\xAF"=>'i',"\xE1\xB8\xB0"=>'K',
  "\xE1\xB8\xB1"=>'k',"\xE1\xB8\xB2"=>'K',"\xE1\xB8\xB3"=>'k',"\xE1\xB8\xB4"=>'K',"\xE1\xB8\xB5"=>'k',
  "\xE1\xB8\xB6"=>'L',"\xE1\xB8\xB7"=>'l',"\xE1\xB8\xB8"=>'L',"\xE1\xB8\xB9"=>'l',"\xE1\xB8\xBA"=>'L',
  "\xE1\xB8\xBB"=>'l',"\xE1\xB8\xBC"=>'L',"\xE1\xB8\xBD"=>'l',"\xE1\xB8\xBE"=>'M',"\xE1\xB8\xBF"=>'m',
  "\xE1\xB9\x80"=>'M',"\xE1\xB9\x81"=>'m',"\xE1\xB9\x82"=>'M',"\xE1\xB9\x83"=>'m',"\xE1\xB9\x84"=>'N',
  "\xE1\xB9\x85"=>'n',"\xE1\xB9\x86"=>'N',"\xE1\xB9\x87"=>'n',"\xE1\xB9\x88"=>'N',"\xE1\xB9\x89"=>'n',
  "\xE1\xB9\x8A"=>'N',"\xE1\xB9\x8B"=>'n',"\xE1\xB9\x8C"=>'O',"\xE1\xB9\x8D"=>'o',"\xE1\xB9\x8E"=>'O',
  "\xE1\xB9\x8F"=>'o',"\xE1\xB9\x90"=>'O',"\xE1\xB9\x91"=>'o',"\xE1\xB9\x92"=>'O',"\xE1\xB9\x93"=>'o',
  "\xE1\xB9\x94"=>'P',"\xE1\xB9\x95"=>'p',"\xE1\xB9\x96"=>'P',"\xE1\xB9\x97"=>'p',"\xE1\xB9\x98"=>'R',
  "\xE1\xB9\x99"=>'r',"\xE1\xB9\x9A"=>'R',"\xE1\xB9\x9B"=>'r',"\xE1\xB9\x9C"=>'R',"\xE1\xB9\x9D"=>'r',
  "\xE1\xB9\x9E"=>'R',"\xE1\xB9\x9F"=>'r',"\xE1\xB9\xA0"=>'S',"\xE1\xB9\xA1"=>'s',"\xE1\xB9\xA2"=>'S',
  "\xE1\xB9\xA3"=>'s',"\xE1\xB9\xA4"=>'S',"\xE1\xB9\xA5"=>'s',"\xE1\xB9\xA6"=>'S',"\xE1\xB9\xA7"=>'s',
  "\xE1\xB9\xA8"=>'S',"\xE1\xB9\xA9"=>'s',"\xE1\xB9\xAA"=>'T',"\xE1\xB9\xAB"=>'t',"\xE1\xB9\xAC"=>'T',
  "\xE1\xB9\xAD"=>'t',"\xE1\xB9\xAE"=>'T',"\xE1\xB9\xAF"=>'t',"\xE1\xB9\xB0"=>'T',"\xE1\xB9\xB1"=>'t',
  "\xE1\xB9\xB2"=>'U',"\xE1\xB9\xB3"=>'u',"\xE1\xB9\xB4"=>'U',"\xE1\xB9\xB5"=>'u',"\xE1\xB9\xB6"=>'U',
  "\xE1\xB9\xB7"=>'u',"\xE1\xB9\xB8"=>'U',"\xE1\xB9\xB9"=>'u',"\xE1\xB9\xBA"=>'U',"\xE1\xB9\xBB"=>'u',
  "\xE1\xB9\xBC"=>'V',"\xE1\xB9\xBD"=>'v',"\xE1\xB9\xBE"=>'V',"\xE1\xB9\xBF"=>'v',"\xE1\xBA\x80"=>'W',
  "\xE1\xBA\x81"=>'w',"\xE1\xBA\x82"=>'W',"\xE1\xBA\x83"=>'w',"\xE1\xBA\x84"=>'W',"\xE1\xBA\x85"=>'w',
  "\xE1\xBA\x86"=>'W',"\xE1\xBA\x87"=>'w',"\xE1\xBA\x88"=>'W',"\xE1\xBA\x89"=>'w',"\xE1\xBA\x8A"=>'X',
  "\xE1\xBA\x8B"=>'x',"\xE1\xBA\x8C"=>'X',"\xE1\xBA\x8D"=>'x',"\xE1\xBA\x8E"=>'Y',"\xE1\xBA\x8F"=>'y',
  "\xE1\xBA\x90"=>'Z',"\xE1\xBA\x91"=>'z',"\xE1\xBA\x92"=>'Z',"\xE1\xBA\x93"=>'z',"\xE1\xBA\x94"=>'Z',
  "\xE1\xBA\x95"=>'z',"\xE1\xBA\x96"=>'h',"\xE1\xBA\x97"=>'t',"\xE1\xBA\x98"=>'w',"\xE1\xBA\x99"=>'y',
  "\xE1\xBA\x9A"=>'a',"\xE1\xBA\x9B"=>'s',"\xE1\xBA\xA0"=>'A',"\xE1\xBA\xA1"=>'a',"\xE1\xBA\xA2"=>'A',
  "\xE1\xBA\xA3"=>'a',"\xE1\xBA\xA4"=>'A',"\xE1\xBA\xA5"=>'a',"\xE1\xBA\xA6"=>'A',"\xE1\xBA\xA7"=>'a',
  "\xE1\xBA\xA8"=>'A',"\xE1\xBA\xA9"=>'a',"\xE1\xBA\xAA"=>'A',"\xE1\xBA\xAB"=>'a',"\xE1\xBA\xAC"=>'A',
  "\xE1\xBA\xAD"=>'a',"\xE1\xBA\xAE"=>'A',"\xE1\xBA\xAF"=>'a',"\xE1\xBA\xB0"=>'A',"\xE1\xBA\xB1"=>'a',
  "\xE1\xBA\xB2"=>'A',"\xE1\xBA\xB3"=>'a',"\xE1\xBA\xB4"=>'A',"\xE1\xBA\xB5"=>'a',"\xE1\xBA\xB6"=>'A',
  "\xE1\xBA\xB7"=>'a',"\xE1\xBA\xB8"=>'E',"\xE1\xBA\xB9"=>'e',"\xE1\xBA\xBA"=>'E',"\xE1\xBA\xBB"=>'e',
  "\xE1\xBA\xBC"=>'E',"\xE1\xBA\xBD"=>'e',"\xE1\xBA\xBE"=>'E',"\xE1\xBA\xBF"=>'e',"\xE1\xBB\x80"=>'E',
  "\xE1\xBB\x81"=>'e',"\xE1\xBB\x82"=>'E',"\xE1\xBB\x83"=>'e',"\xE1\xBB\x84"=>'E',"\xE1\xBB\x85"=>'e',
  "\xE1\xBB\x86"=>'E',"\xE1\xBB\x87"=>'e',"\xE1\xBB\x88"=>'I',"\xE1\xBB\x89"=>'i',"\xE1\xBB\x8A"=>'I',
  "\xE1\xBB\x8B"=>'i',"\xE1\xBB\x8C"=>'O',"\xE1\xBB\x8D"=>'o',"\xE1\xBB\x8E"=>'O',"\xE1\xBB\x8F"=>'o',
  "\xE1\xBB\x90"=>'O',"\xE1\xBB\x91"=>'o',"\xE1\xBB\x92"=>'O',"\xE1\xBB\x93"=>'o',"\xE1\xBB\x94"=>'O',
  "\xE1\xBB\x95"=>'o',"\xE1\xBB\x96"=>'O',"\xE1\xBB\x97"=>'o',"\xE1\xBB\x98"=>'O',"\xE1\xBB\x99"=>'o',
  "\xE1\xBB\x9A"=>'O',"\xE1\xBB\x9B"=>'o',"\xE1\xBB\x9C"=>'O',"\xE1\xBB\x9D"=>'o',"\xE1\xBB\x9E"=>'O',
  "\xE1\xBB\x9F"=>'o',"\xE1\xBB\xA0"=>'O',"\xE1\xBB\xA1"=>'o',"\xE1\xBB\xA2"=>'O',"\xE1\xBB\xA3"=>'o',
  "\xE1\xBB\xA4"=>'U',"\xE1\xBB\xA5"=>'u',"\xE1\xBB\xA6"=>'U',"\xE1\xBB\xA7"=>'u',"\xE1\xBB\xA8"=>'U',
  "\xE1\xBB\xA9"=>'u',"\xE1\xBB\xAA"=>'U',"\xE1\xBB\xAB"=>'u',"\xE1\xBB\xAC"=>'U',"\xE1\xBB\xAD"=>'u',
  "\xE1\xBB\xAE"=>'U',"\xE1\xBB\xAF"=>'u',"\xE1\xBB\xB0"=>'U',"\xE1\xBB\xB1"=>'u',"\xE1\xBB\xB2"=>'Y',
  "\xE1\xBB\xB3"=>'y',"\xE1\xBB\xB4"=>'Y',"\xE1\xBB\xB5"=>'y',"\xE1\xBB\xB6"=>'Y',"\xE1\xBB\xB7"=>'y',
  "\xE1\xBB\xB8"=>'Y',"\xE1\xBB\xB9"=>'y',"\xE1\xBF\xAF"=>'`',"\xE2\x80\xA4"=>'.',"\xE2\x80\xA5"=>'..',
  "\xE2\x80\xA6"=>'...',"\xE2\x80\xBC"=>'!!',"\xE2\x81\x87"=>'??',"\xE2\x81\x88"=>'?!',"\xE2\x81\x89"=>'!?',
  "\xE2\x81\xB1"=>'i',"\xE2\x81\xB4"=>'4',"\xE2\x81\xB5"=>'5',"\xE2\x81\xB6"=>'6',"\xE2\x81\xB7"=>'7',
  "\xE2\x81\xB8"=>'8',"\xE2\x81\xB9"=>'9',"\xE2\x81\xBA"=>'+',"\xE2\x81\xBC"=>'=',"\xE2\x81\xBD"=>'(',
  "\xE2\x81\xBE"=>')',"\xE2\x81\xBF"=>'n',"\xE2\x82\x81"=>'1',"\xE2\x82\x82"=>'2',"\xE2\x82\x83"=>'3',
  "\xE2\x82\x84"=>'4',"\xE2\x82\x85"=>'5',"\xE2\x82\x86"=>'6',"\xE2\x82\x87"=>'7',"\xE2\x82\x88"=>'8',
  "\xE2\x82\x89"=>'9',"\xE2\x82\x8A"=>'+',"\xE2\x82\x8C"=>'=',"\xE2\x82\x8D"=>'(',"\xE2\x82\x8E"=>')',
  "\xE2\x82\x90"=>'a',"\xE2\x82\x91"=>'e',"\xE2\x82\x92"=>'o',"\xE2\x82\x93"=>'x',"\xE2\x82\x95"=>'h',
  "\xE2\x82\x96"=>'k',"\xE2\x82\x97"=>'l',"\xE2\x82\x98"=>'m',"\xE2\x82\x99"=>'n',"\xE2\x82\x9A"=>'p',
  "\xE2\x82\x9B"=>'s',"\xE2\x82\x9C"=>'t',"\xE2\x82\xA8"=>'Rs',"\xE2\x84\x80"=>'a/c',"\xE2\x84\x81"=>'a/s',
  "\xE2\x84\x82"=>'C',"\xE2\x84\x83"=>'C',"\xE2\x84\x85"=>'c/o',"\xE2\x84\x86"=>'c/u',"\xE2\x84\x89"=>'F',
  "\xE2\x84\x8A"=>'g',"\xE2\x84\x8B"=>'H',"\xE2\x84\x8C"=>'H',"\xE2\x84\x8D"=>'H',"\xE2\x84\x8E"=>'h',
  "\xE2\x84\x90"=>'I',"\xE2\x84\x91"=>'I',"\xE2\x84\x92"=>'L',"\xE2\x84\x93"=>'l',"\xE2\x84\x95"=>'N',
  "\xE2\x84\x96"=>'No',"\xE2\x84\x99"=>'P',"\xE2\x84\x9A"=>'Q',"\xE2\x84\x9B"=>'R',"\xE2\x84\x9C"=>'R',
  "\xE2\x84\x9D"=>'R',"\xE2\x84\xA0"=>'SM',"\xE2\x84\xA1"=>'TEL',"\xE2\x84\xA2"=>'TM',"\xE2\x84\xA4"=>'Z',
  "\xE2\x84\xA8"=>'Z',"\xE2\x84\xAA"=>'K',"\xE2\x84\xAB"=>'A',"\xE2\x84\xAC"=>'B',"\xE2\x84\xAD"=>'C',
  "\xE2\x84\xAF"=>'e',"\xE2\x84\xB0"=>'E',"\xE2\x84\xB1"=>'F',"\xE2\x84\xB3"=>'M',"\xE2\x84\xB4"=>'o',
  "\xE2\x84\xB9"=>'i',"\xE2\x84\xBB"=>'FAX',"\xE2\x85\x85"=>'D',"\xE2\x85\x86"=>'d',"\xE2\x85\x87"=>'e',
  "\xE2\x85\x88"=>'i',"\xE2\x85\x89"=>'j',"\xE2\x85\x90"=>'17',"\xE2\x85\x91"=>'19',"\xE2\x85\x92"=>'110',
  "\xE2\x85\x93"=>'13',"\xE2\x85\x94"=>'23',"\xE2\x85\x95"=>'15',"\xE2\x85\x96"=>'25',"\xE2\x85\x97"=>'35',
  "\xE2\x85\x98"=>'45',"\xE2\x85\x99"=>'16',"\xE2\x85\x9A"=>'56',"\xE2\x85\x9B"=>'18',"\xE2\x85\x9C"=>'38',
  "\xE2\x85\x9D"=>'58',"\xE2\x85\x9E"=>'78',"\xE2\x85\x9F"=>'1',"\xE2\x85\xA0"=>'I',"\xE2\x85\xA1"=>'II',
  "\xE2\x85\xA2"=>'III',"\xE2\x85\xA3"=>'IV',"\xE2\x85\xA4"=>'V',"\xE2\x85\xA5"=>'VI',"\xE2\x85\xA6"=>'VII',
  "\xE2\x85\xA7"=>'VIII',"\xE2\x85\xA8"=>'IX',"\xE2\x85\xA9"=>'X',"\xE2\x85\xAA"=>'XI',"\xE2\x85\xAB"=>'XII',
  "\xE2\x85\xAC"=>'L',"\xE2\x85\xAD"=>'C',"\xE2\x85\xAE"=>'D',"\xE2\x85\xAF"=>'M',"\xE2\x85\xB0"=>'i',
  "\xE2\x85\xB1"=>'ii',"\xE2\x85\xB2"=>'iii',"\xE2\x85\xB3"=>'iv',"\xE2\x85\xB4"=>'v',"\xE2\x85\xB5"=>'vi',
  "\xE2\x85\xB6"=>'vii',"\xE2\x85\xB7"=>'viii',"\xE2\x85\xB8"=>'ix',"\xE2\x85\xB9"=>'x',"\xE2\x85\xBA"=>'xi',
  "\xE2\x85\xBB"=>'xii',"\xE2\x85\xBC"=>'l',"\xE2\x85\xBD"=>'c',"\xE2\x85\xBE"=>'d',"\xE2\x85\xBF"=>'m',
  "\xE2\x86\x89"=>'03',"\xE2\x89\xA0"=>'=',"\xE2\x89\xAE"=>'<',"\xE2\x89\xAF"=>'>',"\xE2\x91\xA0"=>'1',
  "\xE2\x91\xA1"=>'2',"\xE2\x91\xA2"=>'3',"\xE2\x91\xA3"=>'4',"\xE2\x91\xA4"=>'5',"\xE2\x91\xA5"=>'6',
  "\xE2\x91\xA6"=>'7',"\xE2\x91\xA7"=>'8',"\xE2\x91\xA8"=>'9',"\xE2\x91\xA9"=>'10',"\xE2\x91\xAA"=>'11',
  "\xE2\x91\xAB"=>'12',"\xE2\x91\xAC"=>'13',"\xE2\x91\xAD"=>'14',"\xE2\x91\xAE"=>'15',"\xE2\x91\xAF"=>'16',
  "\xE2\x91\xB0"=>'17',"\xE2\x91\xB1"=>'18',"\xE2\x91\xB2"=>'19',"\xE2\x91\xB3"=>'20',"\xE2\x91\xB4"=>'(1)',
  "\xE2\x91\xB5"=>'(2)',"\xE2\x91\xB6"=>'(3)',"\xE2\x91\xB7"=>'(4)',"\xE2\x91\xB8"=>'(5)',"\xE2\x91\xB9"=>'(6)',
  "\xE2\x91\xBA"=>'(7)',"\xE2\x91\xBB"=>'(8)',"\xE2\x91\xBC"=>'(9)',"\xE2\x91\xBD"=>'(10)',"\xE2\x91\xBE"=>'(11)',
  "\xE2\x91\xBF"=>'(12)',"\xE2\x92\x80"=>'(13)',"\xE2\x92\x81"=>'(14)',"\xE2\x92\x82"=>'(15)',"\xE2\x92\x83"=>'(16)',
  "\xE2\x92\x84"=>'(17)',"\xE2\x92\x85"=>'(18)',"\xE2\x92\x86"=>'(19)',"\xE2\x92\x87"=>'(20)',"\xE2\x92\x88"=>'1.',
  "\xE2\x92\x89"=>'2.',"\xE2\x92\x8A"=>'3.',"\xE2\x92\x8B"=>'4.',"\xE2\x92\x8C"=>'5.',"\xE2\x92\x8D"=>'6.',
  "\xE2\x92\x8E"=>'7.',"\xE2\x92\x8F"=>'8.',"\xE2\x92\x90"=>'9.',"\xE2\x92\x91"=>'10.',"\xE2\x92\x92"=>'11.',
  "\xE2\x92\x93"=>'12.',"\xE2\x92\x94"=>'13.',"\xE2\x92\x95"=>'14.',"\xE2\x92\x96"=>'15.',"\xE2\x92\x97"=>'16.',
  "\xE2\x92\x98"=>'17.',"\xE2\x92\x99"=>'18.',"\xE2\x92\x9A"=>'19.',"\xE2\x92\x9B"=>'20.',"\xE2\x92\x9C"=>'(a)',
  "\xE2\x92\x9D"=>'(b)',"\xE2\x92\x9E"=>'(c)',"\xE2\x92\x9F"=>'(d)',"\xE2\x92\xA0"=>'(e)',"\xE2\x92\xA1"=>'(f)',
  "\xE2\x92\xA2"=>'(g)',"\xE2\x92\xA3"=>'(h)',"\xE2\x92\xA4"=>'(i)',"\xE2\x92\xA5"=>'(j)',"\xE2\x92\xA6"=>'(k)',
  "\xE2\x92\xA7"=>'(l)',"\xE2\x92\xA8"=>'(m)',"\xE2\x92\xA9"=>'(n)',"\xE2\x92\xAA"=>'(o)',"\xE2\x92\xAB"=>'(p)',
  "\xE2\x92\xAC"=>'(q)',"\xE2\x92\xAD"=>'(r)',"\xE2\x92\xAE"=>'(s)',"\xE2\x92\xAF"=>'(t)',"\xE2\x92\xB0"=>'(u)',
  "\xE2\x92\xB1"=>'(v)',"\xE2\x92\xB2"=>'(w)',"\xE2\x92\xB3"=>'(x)',"\xE2\x92\xB4"=>'(y)',"\xE2\x92\xB5"=>'(z)',
  "\xE2\x92\xB6"=>'A',"\xE2\x92\xB7"=>'B',"\xE2\x92\xB8"=>'C',"\xE2\x92\xB9"=>'D',"\xE2\x92\xBA"=>'E',
  "\xE2\x92\xBB"=>'F',"\xE2\x92\xBC"=>'G',"\xE2\x92\xBD"=>'H',"\xE2\x92\xBE"=>'I',"\xE2\x92\xBF"=>'J',
  "\xE2\x93\x80"=>'K',"\xE2\x93\x81"=>'L',"\xE2\x93\x82"=>'M',"\xE2\x93\x83"=>'N',"\xE2\x93\x84"=>'O',
  "\xE2\x93\x85"=>'P',"\xE2\x93\x86"=>'Q',"\xE2\x93\x87"=>'R',"\xE2\x93\x88"=>'S',"\xE2\x93\x89"=>'T',
  "\xE2\x93\x8A"=>'U',"\xE2\x93\x8B"=>'V',"\xE2\x93\x8C"=>'W',"\xE2\x93\x8D"=>'X',"\xE2\x93\x8E"=>'Y',
  "\xE2\x93\x8F"=>'Z',"\xE2\x93\x90"=>'a',"\xE2\x93\x91"=>'b',"\xE2\x93\x92"=>'c',"\xE2\x93\x93"=>'d',
  "\xE2\x93\x94"=>'e',"\xE2\x93\x95"=>'f',"\xE2\x93\x96"=>'g',"\xE2\x93\x97"=>'h',"\xE2\x93\x98"=>'i',
  "\xE2\x93\x99"=>'j',"\xE2\x93\x9A"=>'k',"\xE2\x93\x9B"=>'l',"\xE2\x93\x9C"=>'m',"\xE2\x93\x9D"=>'n',
  "\xE2\x93\x9E"=>'o',"\xE2\x93\x9F"=>'p',"\xE2\x93\xA0"=>'q',"\xE2\x93\xA1"=>'r',"\xE2\x93\xA2"=>'s',
  "\xE2\x93\xA3"=>'t',"\xE2\x93\xA4"=>'u',"\xE2\x93\xA5"=>'v',"\xE2\x93\xA6"=>'w',"\xE2\x93\xA7"=>'x',
  "\xE2\x93\xA8"=>'y',"\xE2\x93\xA9"=>'z',"\xE2\xA9\xB4"=>'::=',"\xE2\xA9\xB5"=>'==',"\xE2\xA9\xB6"=>'===',
  "\xE2\xB1\xBC"=>'j',"\xE2\xB1\xBD"=>'V',"\xE3\x88\x80"=>'()',"\xE3\x88\x81"=>'()',"\xE3\x88\x82"=>'()',
  "\xE3\x88\x83"=>'()',"\xE3\x88\x84"=>'()',"\xE3\x88\x85"=>'()',"\xE3\x88\x86"=>'()',"\xE3\x88\x87"=>'()',
  "\xE3\x88\x88"=>'()',"\xE3\x88\x89"=>'()',"\xE3\x88\x8A"=>'()',"\xE3\x88\x8B"=>'()',"\xE3\x88\x8C"=>'()',
  "\xE3\x88\x8D"=>'()',"\xE3\x88\x8E"=>'()',"\xE3\x88\x8F"=>'()',"\xE3\x88\x90"=>'()',"\xE3\x88\x91"=>'()',
  "\xE3\x88\x92"=>'()',"\xE3\x88\x93"=>'()',"\xE3\x88\x94"=>'()',"\xE3\x88\x95"=>'()',"\xE3\x88\x96"=>'()',
  "\xE3\x88\x97"=>'()',"\xE3\x88\x98"=>'()',"\xE3\x88\x99"=>'()',"\xE3\x88\x9A"=>'()',"\xE3\x88\x9B"=>'()',
  "\xE3\x88\x9C"=>'()',"\xE3\x88\x9D"=>'()',"\xE3\x88\x9E"=>'()',"\xE3\x88\xA0"=>'()',"\xE3\x88\xA1"=>'()',
  "\xE3\x88\xA2"=>'()',"\xE3\x88\xA3"=>'()',"\xE3\x88\xA4"=>'()',"\xE3\x88\xA5"=>'()',"\xE3\x88\xA6"=>'()',
  "\xE3\x88\xA7"=>'()',"\xE3\x88\xA8"=>'()',"\xE3\x88\xA9"=>'()',"\xE3\x88\xAA"=>'()',"\xE3\x88\xAB"=>'()',
  "\xE3\x88\xAC"=>'()',"\xE3\x88\xAD"=>'()',"\xE3\x88\xAE"=>'()',"\xE3\x88\xAF"=>'()',"\xE3\x88\xB0"=>'()',
  "\xE3\x88\xB1"=>'()',"\xE3\x88\xB2"=>'()',"\xE3\x88\xB3"=>'()',"\xE3\x88\xB4"=>'()',"\xE3\x88\xB5"=>'()',
  "\xE3\x88\xB6"=>'()',"\xE3\x88\xB7"=>'()',"\xE3\x88\xB8"=>'()',"\xE3\x88\xB9"=>'()',"\xE3\x88\xBA"=>'()',
  "\xE3\x88\xBB"=>'()',"\xE3\x88\xBC"=>'()',"\xE3\x88\xBD"=>'()',"\xE3\x88\xBE"=>'()',"\xE3\x88\xBF"=>'()',
  "\xE3\x89\x80"=>'()',"\xE3\x89\x81"=>'()',"\xE3\x89\x82"=>'()',"\xE3\x89\x83"=>'()',"\xE3\x89\x90"=>'PTE',
  "\xE3\x89\x91"=>'21',"\xE3\x89\x92"=>'22',"\xE3\x89\x93"=>'23',"\xE3\x89\x94"=>'24',"\xE3\x89\x95"=>'25',
  "\xE3\x89\x96"=>'26',"\xE3\x89\x97"=>'27',"\xE3\x89\x98"=>'28',"\xE3\x89\x99"=>'29',"\xE3\x89\x9A"=>'30',
  "\xE3\x89\x9B"=>'31',"\xE3\x89\x9C"=>'32',"\xE3\x89\x9D"=>'33',"\xE3\x89\x9E"=>'34',"\xE3\x89\x9F"=>'35',
  "\xE3\x8A\xB1"=>'36',"\xE3\x8A\xB2"=>'37',"\xE3\x8A\xB3"=>'38',"\xE3\x8A\xB4"=>'39',"\xE3\x8A\xB5"=>'40',
  "\xE3\x8A\xB6"=>'41',"\xE3\x8A\xB7"=>'42',"\xE3\x8A\xB8"=>'43',"\xE3\x8A\xB9"=>'44',"\xE3\x8A\xBA"=>'45',
  "\xE3\x8A\xBB"=>'46',"\xE3\x8A\xBC"=>'47',"\xE3\x8A\xBD"=>'48',"\xE3\x8A\xBE"=>'49',"\xE3\x8A\xBF"=>'50',
  "\xE3\x8B\x80"=>'1',"\xE3\x8B\x81"=>'2',"\xE3\x8B\x82"=>'3',"\xE3\x8B\x83"=>'4',"\xE3\x8B\x84"=>'5',
  "\xE3\x8B\x85"=>'6',"\xE3\x8B\x86"=>'7',"\xE3\x8B\x87"=>'8',"\xE3\x8B\x88"=>'9',"\xE3\x8B\x89"=>'10',
  "\xE3\x8B\x8A"=>'11',"\xE3\x8B\x8B"=>'12',"\xE3\x8B\x8C"=>'Hg',"\xE3\x8B\x8D"=>'erg',"\xE3\x8B\x8E"=>'eV',
  "\xE3\x8B\x8F"=>'LTD',"\xE3\x8D\x99"=>'1',"\xE3\x8D\x9A"=>'2',"\xE3\x8D\x9B"=>'3',"\xE3\x8D\x9C"=>'4',
  "\xE3\x8D\x9D"=>'5',"\xE3\x8D\x9E"=>'6',"\xE3\x8D\x9F"=>'7',"\xE3\x8D\xA0"=>'8',"\xE3\x8D\xA1"=>'9',
  "\xE3\x8D\xA2"=>'10',"\xE3\x8D\xA3"=>'11',"\xE3\x8D\xA4"=>'12',"\xE3\x8D\xA5"=>'13',"\xE3\x8D\xA6"=>'14',
  "\xE3\x8D\xA7"=>'15',"\xE3\x8D\xA8"=>'16',"\xE3\x8D\xA9"=>'17',"\xE3\x8D\xAA"=>'18',"\xE3\x8D\xAB"=>'19',
  "\xE3\x8D\xAC"=>'20',"\xE3\x8D\xAD"=>'21',"\xE3\x8D\xAE"=>'22',"\xE3\x8D\xAF"=>'23',"\xE3\x8D\xB0"=>'24',
  "\xE3\x8D\xB1"=>'hPa',"\xE3\x8D\xB2"=>'da',"\xE3\x8D\xB3"=>'AU',"\xE3\x8D\xB4"=>'bar',"\xE3\x8D\xB5"=>'oV',
  "\xE3\x8D\xB6"=>'pc',"\xE3\x8D\xB7"=>'dm',"\xE3\x8D\xB8"=>'dm2',"\xE3\x8D\xB9"=>'dm3',"\xE3\x8D\xBA"=>'IU',
  "\xE3\x8E\x80"=>'pA',"\xE3\x8E\x81"=>'nA',"\xE3\x8E\x82"=>'A',"\xE3\x8E\x83"=>'mA',"\xE3\x8E\x84"=>'kA',
  "\xE3\x8E\x85"=>'KB',"\xE3\x8E\x86"=>'MB',"\xE3\x8E\x87"=>'GB',"\xE3\x8E\x88"=>'cal',"\xE3\x8E\x89"=>'kcal',
  "\xE3\x8E\x8A"=>'pF',"\xE3\x8E\x8B"=>'nF',"\xE3\x8E\x8C"=>'F',"\xE3\x8E\x8D"=>'g',"\xE3\x8E\x8E"=>'mg',
  "\xE3\x8E\x8F"=>'kg',"\xE3\x8E\x90"=>'Hz',"\xE3\x8E\x91"=>'kHz',"\xE3\x8E\x92"=>'MHz',"\xE3\x8E\x93"=>'GHz',
  "\xE3\x8E\x94"=>'THz',"\xE3\x8E\x95"=>'l',"\xE3\x8E\x96"=>'ml',"\xE3\x8E\x97"=>'dl',"\xE3\x8E\x98"=>'kl',
  "\xE3\x8E\x99"=>'fm',"\xE3\x8E\x9A"=>'nm',"\xE3\x8E\x9B"=>'m',"\xE3\x8E\x9C"=>'mm',"\xE3\x8E\x9D"=>'cm',
  "\xE3\x8E\x9E"=>'km',"\xE3\x8E\x9F"=>'mm2',"\xE3\x8E\xA0"=>'cm2',"\xE3\x8E\xA1"=>'m2',"\xE3\x8E\xA2"=>'km2',
  "\xE3\x8E\xA3"=>'mm3',"\xE3\x8E\xA4"=>'cm3',"\xE3\x8E\xA5"=>'m3',"\xE3\x8E\xA6"=>'km3',"\xE3\x8E\xA7"=>'ms',
  "\xE3\x8E\xA8"=>'ms2',"\xE3\x8E\xA9"=>'Pa',"\xE3\x8E\xAA"=>'kPa',"\xE3\x8E\xAB"=>'MPa',"\xE3\x8E\xAC"=>'GPa',
  "\xE3\x8E\xAD"=>'rad',"\xE3\x8E\xAE"=>'rads',"\xE3\x8E\xAF"=>'rads2',"\xE3\x8E\xB0"=>'ps',"\xE3\x8E\xB1"=>'ns',
  "\xE3\x8E\xB2"=>'s',"\xE3\x8E\xB3"=>'ms',"\xE3\x8E\xB4"=>'pV',"\xE3\x8E\xB5"=>'nV',"\xE3\x8E\xB6"=>'V',
  "\xE3\x8E\xB7"=>'mV',"\xE3\x8E\xB8"=>'kV',"\xE3\x8E\xB9"=>'MV',"\xE3\x8E\xBA"=>'pW',"\xE3\x8E\xBB"=>'nW',
  "\xE3\x8E\xBC"=>'W',"\xE3\x8E\xBD"=>'mW',"\xE3\x8E\xBE"=>'kW',"\xE3\x8E\xBF"=>'MW',"\xE3\x8F\x80"=>'k',
  "\xE3\x8F\x81"=>'M',"\xE3\x8F\x82"=>'a.m.',"\xE3\x8F\x83"=>'Bq',"\xE3\x8F\x84"=>'cc',"\xE3\x8F\x85"=>'cd',
  "\xE3\x8F\x86"=>'Ckg',"\xE3\x8F\x87"=>'Co.',"\xE3\x8F\x88"=>'dB',"\xE3\x8F\x89"=>'Gy',"\xE3\x8F\x8A"=>'ha',
  "\xE3\x8F\x8B"=>'HP',"\xE3\x8F\x8C"=>'in',"\xE3\x8F\x8D"=>'KK',"\xE3\x8F\x8E"=>'KM',"\xE3\x8F\x8F"=>'kt',
  "\xE3\x8F\x90"=>'lm',"\xE3\x8F\x91"=>'ln',"\xE3\x8F\x92"=>'log',"\xE3\x8F\x93"=>'lx',"\xE3\x8F\x94"=>'mb',
  "\xE3\x8F\x95"=>'mil',"\xE3\x8F\x96"=>'mol',"\xE3\x8F\x97"=>'PH',"\xE3\x8F\x98"=>'p.m.',"\xE3\x8F\x99"=>'PPM',
  "\xE3\x8F\x9A"=>'PR',"\xE3\x8F\x9B"=>'sr',"\xE3\x8F\x9C"=>'Sv',"\xE3\x8F\x9D"=>'Wb',"\xE3\x8F\x9E"=>'Vm',
  "\xE3\x8F\x9F"=>'Am',"\xE3\x8F\xA0"=>'1',"\xE3\x8F\xA1"=>'2',"\xE3\x8F\xA2"=>'3',"\xE3\x8F\xA3"=>'4',
  "\xE3\x8F\xA4"=>'5',"\xE3\x8F\xA5"=>'6',"\xE3\x8F\xA6"=>'7',"\xE3\x8F\xA7"=>'8',"\xE3\x8F\xA8"=>'9',
  "\xE3\x8F\xA9"=>'10',"\xE3\x8F\xAA"=>'11',"\xE3\x8F\xAB"=>'12',"\xE3\x8F\xAC"=>'13',"\xE3\x8F\xAD"=>'14',
  "\xE3\x8F\xAE"=>'15',"\xE3\x8F\xAF"=>'16',"\xE3\x8F\xB0"=>'17',"\xE3\x8F\xB1"=>'18',"\xE3\x8F\xB2"=>'19',
  "\xE3\x8F\xB3"=>'20',"\xE3\x8F\xB4"=>'21',"\xE3\x8F\xB5"=>'22',"\xE3\x8F\xB6"=>'23',"\xE3\x8F\xB7"=>'24',
  "\xE3\x8F\xB8"=>'25',"\xE3\x8F\xB9"=>'26',"\xE3\x8F\xBA"=>'27',"\xE3\x8F\xBB"=>'28',"\xE3\x8F\xBC"=>'29',
  "\xE3\x8F\xBD"=>'30',"\xE3\x8F\xBE"=>'31',"\xE3\x8F\xBF"=>'gal',"\xEF\xAC\x80"=>'ff',"\xEF\xAC\x81"=>'fi',
  "\xEF\xAC\x82"=>'fl',"\xEF\xAC\x83"=>'ffi',"\xEF\xAC\x84"=>'ffl',"\xEF\xAC\x85"=>'st',"\xEF\xAC\x86"=>'st',
  "\xEF\xAC\xA9"=>'+',"\xEF\xB8\x90"=>',',"\xEF\xB8\x93"=>':',"\xEF\xB8\x94"=>';',"\xEF\xB8\x95"=>'!',
  "\xEF\xB8\x96"=>'?',"\xEF\xB8\x99"=>'...',"\xEF\xB8\xB0"=>'..',"\xEF\xB8\xB3"=>'_',"\xEF\xB8\xB4"=>'_',
  "\xEF\xB8\xB5"=>'(',"\xEF\xB8\xB6"=>')',"\xEF\xB8\xB7"=>'{',"\xEF\xB8\xB8"=>'}',"\xEF\xB9\x87"=>'[',
  "\xEF\xB9\x88"=>']',"\xEF\xB9\x8D"=>'_',"\xEF\xB9\x8E"=>'_',"\xEF\xB9\x8F"=>'_',"\xEF\xB9\x90"=>',',
  "\xEF\xB9\x92"=>'.',"\xEF\xB9\x94"=>';',"\xEF\xB9\x95"=>':',"\xEF\xB9\x96"=>'?',"\xEF\xB9\x97"=>'!',
  "\xEF\xB9\x99"=>'(',"\xEF\xB9\x9A"=>')',"\xEF\xB9\x9B"=>'{',"\xEF\xB9\x9C"=>'}',"\xEF\xB9\x9F"=>'#',
  "\xEF\xB9\xA0"=>'&',"\xEF\xB9\xA1"=>'*',"\xEF\xB9\xA2"=>'+',"\xEF\xB9\xA3"=>'-',"\xEF\xB9\xA4"=>'<',
  "\xEF\xB9\xA5"=>'>',"\xEF\xB9\xA6"=>'=',"\xEF\xB9\xA8"=>'\\',"\xEF\xB9\xA9"=>'$',"\xEF\xB9\xAA"=>'%',
  "\xEF\xB9\xAB"=>'@',"\xEF\xBC\x81"=>'!',"\xEF\xBC\x82"=>'"',"\xEF\xBC\x83"=>'#',"\xEF\xBC\x84"=>'$',
  "\xEF\xBC\x85"=>'%',"\xEF\xBC\x86"=>'&',"\xEF\xBC\x87"=>'\'',"\xEF\xBC\x88"=>'(',"\xEF\xBC\x89"=>')',
  "\xEF\xBC\x8A"=>'*',"\xEF\xBC\x8B"=>'+',"\xEF\xBC\x8C"=>',',"\xEF\xBC\x8D"=>'-',"\xEF\xBC\x8E"=>'.',
  "\xEF\xBC\x8F"=>'/',"\xEF\xBC\x91"=>'1',"\xEF\xBC\x92"=>'2',"\xEF\xBC\x93"=>'3',"\xEF\xBC\x94"=>'4',
  "\xEF\xBC\x95"=>'5',"\xEF\xBC\x96"=>'6',"\xEF\xBC\x97"=>'7',"\xEF\xBC\x98"=>'8',"\xEF\xBC\x99"=>'9',
  "\xEF\xBC\x9A"=>':',"\xEF\xBC\x9B"=>';',"\xEF\xBC\x9C"=>'<',"\xEF\xBC\x9D"=>'=',"\xEF\xBC\x9E"=>'>',
  "\xEF\xBC\x9F"=>'?',"\xEF\xBC\xA0"=>'@',"\xEF\xBC\xA1"=>'A',"\xEF\xBC\xA2"=>'B',"\xEF\xBC\xA3"=>'C',
  "\xEF\xBC\xA4"=>'D',"\xEF\xBC\xA5"=>'E',"\xEF\xBC\xA6"=>'F',"\xEF\xBC\xA7"=>'G',"\xEF\xBC\xA8"=>'H',
  "\xEF\xBC\xA9"=>'I',"\xEF\xBC\xAA"=>'J',"\xEF\xBC\xAB"=>'K',"\xEF\xBC\xAC"=>'L',"\xEF\xBC\xAD"=>'M',
  "\xEF\xBC\xAE"=>'N',"\xEF\xBC\xAF"=>'O',"\xEF\xBC\xB0"=>'P',"\xEF\xBC\xB1"=>'Q',"\xEF\xBC\xB2"=>'R',
  "\xEF\xBC\xB3"=>'S',"\xEF\xBC\xB4"=>'T',"\xEF\xBC\xB5"=>'U',"\xEF\xBC\xB6"=>'V',"\xEF\xBC\xB7"=>'W',
  "\xEF\xBC\xB8"=>'X',"\xEF\xBC\xB9"=>'Y',"\xEF\xBC\xBA"=>'Z',"\xEF\xBC\xBB"=>'[',"\xEF\xBC\xBC"=>'\\',
  "\xEF\xBC\xBD"=>']',"\xEF\xBC\xBE"=>'^',"\xEF\xBC\xBF"=>'_',"\xEF\xBD\x80"=>'`',"\xEF\xBD\x81"=>'a',
  "\xEF\xBD\x82"=>'b',"\xEF\xBD\x83"=>'c',"\xEF\xBD\x84"=>'d',"\xEF\xBD\x85"=>'e',"\xEF\xBD\x86"=>'f',
  "\xEF\xBD\x87"=>'g',"\xEF\xBD\x88"=>'h',"\xEF\xBD\x89"=>'i',"\xEF\xBD\x8A"=>'j',"\xEF\xBD\x8B"=>'k',
  "\xEF\xBD\x8C"=>'l',"\xEF\xBD\x8D"=>'m',"\xEF\xBD\x8E"=>'n',"\xEF\xBD\x8F"=>'o',"\xEF\xBD\x90"=>'p',
  "\xEF\xBD\x91"=>'q',"\xEF\xBD\x92"=>'r',"\xEF\xBD\x93"=>'s',"\xEF\xBD\x94"=>'t',"\xEF\xBD\x95"=>'u',
  "\xEF\xBD\x96"=>'v',"\xEF\xBD\x97"=>'w',"\xEF\xBD\x98"=>'x',"\xEF\xBD\x99"=>'y',"\xEF\xBD\x9A"=>'z',
  "\xEF\xBD\x9B"=>'{',"\xEF\xBD\x9C"=>'|',"\xEF\xBD\x9D"=>'}',"\xEF\xBD\x9E"=>'~',"\xF0\x9D\x90\x80"=>'A',
  "\xF0\x9D\x90\x81"=>'B',"\xF0\x9D\x90\x82"=>'C',"\xF0\x9D\x90\x83"=>'D',"\xF0\x9D\x90\x84"=>'E',
  "\xF0\x9D\x90\x85"=>'F',"\xF0\x9D\x90\x86"=>'G',"\xF0\x9D\x90\x87"=>'H',"\xF0\x9D\x90\x88"=>'I',
  "\xF0\x9D\x90\x89"=>'J',"\xF0\x9D\x90\x8A"=>'K',"\xF0\x9D\x90\x8B"=>'L',"\xF0\x9D\x90\x8C"=>'M',
  "\xF0\x9D\x90\x8D"=>'N',"\xF0\x9D\x90\x8E"=>'O',"\xF0\x9D\x90\x8F"=>'P',"\xF0\x9D\x90\x90"=>'Q',
  "\xF0\x9D\x90\x91"=>'R',"\xF0\x9D\x90\x92"=>'S',"\xF0\x9D\x90\x93"=>'T',"\xF0\x9D\x90\x94"=>'U',
  "\xF0\x9D\x90\x95"=>'V',"\xF0\x9D\x90\x96"=>'W',"\xF0\x9D\x90\x97"=>'X',"\xF0\x9D\x90\x98"=>'Y',
  "\xF0\x9D\x90\x99"=>'Z',"\xF0\x9D\x90\x9A"=>'a',"\xF0\x9D\x90\x9B"=>'b',"\xF0\x9D\x90\x9C"=>'c',
  "\xF0\x9D\x90\x9D"=>'d',"\xF0\x9D\x90\x9E"=>'e',"\xF0\x9D\x90\x9F"=>'f',"\xF0\x9D\x90\xA0"=>'g',
  "\xF0\x9D\x90\xA1"=>'h',"\xF0\x9D\x90\xA2"=>'i',"\xF0\x9D\x90\xA3"=>'j',"\xF0\x9D\x90\xA4"=>'k',
  "\xF0\x9D\x90\xA5"=>'l',"\xF0\x9D\x90\xA6"=>'m',"\xF0\x9D\x90\xA7"=>'n',"\xF0\x9D\x90\xA8"=>'o',
  "\xF0\x9D\x90\xA9"=>'p',"\xF0\x9D\x90\xAA"=>'q',"\xF0\x9D\x90\xAB"=>'r',"\xF0\x9D\x90\xAC"=>'s',
  "\xF0\x9D\x90\xAD"=>'t',"\xF0\x9D\x90\xAE"=>'u',"\xF0\x9D\x90\xAF"=>'v',"\xF0\x9D\x90\xB0"=>'w',
  "\xF0\x9D\x90\xB1"=>'x',"\xF0\x9D\x90\xB2"=>'y',"\xF0\x9D\x90\xB3"=>'z',"\xF0\x9D\x90\xB4"=>'A',
  "\xF0\x9D\x90\xB5"=>'B',"\xF0\x9D\x90\xB6"=>'C',"\xF0\x9D\x90\xB7"=>'D',"\xF0\x9D\x90\xB8"=>'E',
  "\xF0\x9D\x90\xB9"=>'F',"\xF0\x9D\x90\xBA"=>'G',"\xF0\x9D\x90\xBB"=>'H',"\xF0\x9D\x90\xBC"=>'I',
  "\xF0\x9D\x90\xBD"=>'J',"\xF0\x9D\x90\xBE"=>'K',"\xF0\x9D\x90\xBF"=>'L',"\xF0\x9D\x91\x80"=>'M',
  "\xF0\x9D\x91\x81"=>'N',"\xF0\x9D\x91\x82"=>'O',"\xF0\x9D\x91\x83"=>'P',"\xF0\x9D\x91\x84"=>'Q',
  "\xF0\x9D\x91\x85"=>'R',"\xF0\x9D\x91\x86"=>'S',"\xF0\x9D\x91\x87"=>'T',"\xF0\x9D\x91\x88"=>'U',
  "\xF0\x9D\x91\x89"=>'V',"\xF0\x9D\x91\x8A"=>'W',"\xF0\x9D\x91\x8B"=>'X',"\xF0\x9D\x91\x8C"=>'Y',
  "\xF0\x9D\x91\x8D"=>'Z',"\xF0\x9D\x91\x8E"=>'a',"\xF0\x9D\x91\x8F"=>'b',"\xF0\x9D\x91\x90"=>'c',
  "\xF0\x9D\x91\x91"=>'d',"\xF0\x9D\x91\x92"=>'e',"\xF0\x9D\x91\x93"=>'f',"\xF0\x9D\x91\x94"=>'g',
  "\xF0\x9D\x91\x96"=>'i',"\xF0\x9D\x91\x97"=>'j',"\xF0\x9D\x91\x98"=>'k',"\xF0\x9D\x91\x99"=>'l',
  "\xF0\x9D\x91\x9A"=>'m',"\xF0\x9D\x91\x9B"=>'n',"\xF0\x9D\x91\x9C"=>'o',"\xF0\x9D\x91\x9D"=>'p',
  "\xF0\x9D\x91\x9E"=>'q',"\xF0\x9D\x91\x9F"=>'r',"\xF0\x9D\x91\xA0"=>'s',"\xF0\x9D\x91\xA1"=>'t',
  "\xF0\x9D\x91\xA2"=>'u',"\xF0\x9D\x91\xA3"=>'v',"\xF0\x9D\x91\xA4"=>'w',"\xF0\x9D\x91\xA5"=>'x',
  "\xF0\x9D\x91\xA6"=>'y',"\xF0\x9D\x91\xA7"=>'z',"\xF0\x9D\x91\xA8"=>'A',"\xF0\x9D\x91\xA9"=>'B',
  "\xF0\x9D\x91\xAA"=>'C',"\xF0\x9D\x91\xAB"=>'D',"\xF0\x9D\x91\xAC"=>'E',"\xF0\x9D\x91\xAD"=>'F',
  "\xF0\x9D\x91\xAE"=>'G',"\xF0\x9D\x91\xAF"=>'H',"\xF0\x9D\x91\xB0"=>'I',"\xF0\x9D\x91\xB1"=>'J',
  "\xF0\x9D\x91\xB2"=>'K',"\xF0\x9D\x91\xB3"=>'L',"\xF0\x9D\x91\xB4"=>'M',"\xF0\x9D\x91\xB5"=>'N',
  "\xF0\x9D\x91\xB6"=>'O',"\xF0\x9D\x91\xB7"=>'P',"\xF0\x9D\x91\xB8"=>'Q',"\xF0\x9D\x91\xB9"=>'R',
  "\xF0\x9D\x91\xBA"=>'S',"\xF0\x9D\x91\xBB"=>'T',"\xF0\x9D\x91\xBC"=>'U',"\xF0\x9D\x91\xBD"=>'V',
  "\xF0\x9D\x91\xBE"=>'W',"\xF0\x9D\x91\xBF"=>'X',"\xF0\x9D\x92\x80"=>'Y',"\xF0\x9D\x92\x81"=>'Z',
  "\xF0\x9D\x92\x82"=>'a',"\xF0\x9D\x92\x83"=>'b',"\xF0\x9D\x92\x84"=>'c',"\xF0\x9D\x92\x85"=>'d',
  "\xF0\x9D\x92\x86"=>'e',"\xF0\x9D\x92\x87"=>'f',"\xF0\x9D\x92\x88"=>'g',"\xF0\x9D\x92\x89"=>'h',
  "\xF0\x9D\x92\x8A"=>'i',"\xF0\x9D\x92\x8B"=>'j',"\xF0\x9D\x92\x8C"=>'k',"\xF0\x9D\x92\x8D"=>'l',
  "\xF0\x9D\x92\x8E"=>'m',"\xF0\x9D\x92\x8F"=>'n',"\xF0\x9D\x92\x90"=>'o',"\xF0\x9D\x92\x91"=>'p',
  "\xF0\x9D\x92\x92"=>'q',"\xF0\x9D\x92\x93"=>'r',"\xF0\x9D\x92\x94"=>'s',"\xF0\x9D\x92\x95"=>'t',
  "\xF0\x9D\x92\x96"=>'u',"\xF0\x9D\x92\x97"=>'v',"\xF0\x9D\x92\x98"=>'w',"\xF0\x9D\x92\x99"=>'x',
  "\xF0\x9D\x92\x9A"=>'y',"\xF0\x9D\x92\x9B"=>'z',"\xF0\x9D\x92\x9C"=>'A',"\xF0\x9D\x92\x9E"=>'C',
  "\xF0\x9D\x92\x9F"=>'D',"\xF0\x9D\x92\xA2"=>'G',"\xF0\x9D\x92\xA5"=>'J',"\xF0\x9D\x92\xA6"=>'K',
  "\xF0\x9D\x92\xA9"=>'N',"\xF0\x9D\x92\xAA"=>'O',"\xF0\x9D\x92\xAB"=>'P',"\xF0\x9D\x92\xAC"=>'Q',
  "\xF0\x9D\x92\xAE"=>'S',"\xF0\x9D\x92\xAF"=>'T',"\xF0\x9D\x92\xB0"=>'U',"\xF0\x9D\x92\xB1"=>'V',
  "\xF0\x9D\x92\xB2"=>'W',"\xF0\x9D\x92\xB3"=>'X',"\xF0\x9D\x92\xB4"=>'Y',"\xF0\x9D\x92\xB5"=>'Z',
  "\xF0\x9D\x92\xB6"=>'a',"\xF0\x9D\x92\xB7"=>'b',"\xF0\x9D\x92\xB8"=>'c',"\xF0\x9D\x92\xB9"=>'d',
  "\xF0\x9D\x92\xBB"=>'f',"\xF0\x9D\x92\xBD"=>'h',"\xF0\x9D\x92\xBE"=>'i',"\xF0\x9D\x92\xBF"=>'j',
  "\xF0\x9D\x93\x80"=>'k',"\xF0\x9D\x93\x81"=>'l',"\xF0\x9D\x93\x82"=>'m',"\xF0\x9D\x93\x83"=>'n',
  "\xF0\x9D\x93\x85"=>'p',"\xF0\x9D\x93\x86"=>'q',"\xF0\x9D\x93\x87"=>'r',"\xF0\x9D\x93\x88"=>'s',
  "\xF0\x9D\x93\x89"=>'t',"\xF0\x9D\x93\x8A"=>'u',"\xF0\x9D\x93\x8B"=>'v',"\xF0\x9D\x93\x8C"=>'w',
  "\xF0\x9D\x93\x8D"=>'x',"\xF0\x9D\x93\x8E"=>'y',"\xF0\x9D\x93\x8F"=>'z',"\xF0\x9D\x93\x90"=>'A',
  "\xF0\x9D\x93\x91"=>'B',"\xF0\x9D\x93\x92"=>'C',"\xF0\x9D\x93\x93"=>'D',"\xF0\x9D\x93\x94"=>'E',
  "\xF0\x9D\x93\x95"=>'F',"\xF0\x9D\x93\x96"=>'G',"\xF0\x9D\x93\x97"=>'H',"\xF0\x9D\x93\x98"=>'I',
  "\xF0\x9D\x93\x99"=>'J',"\xF0\x9D\x93\x9A"=>'K',"\xF0\x9D\x93\x9B"=>'L',"\xF0\x9D\x93\x9C"=>'M',
  "\xF0\x9D\x93\x9D"=>'N',"\xF0\x9D\x93\x9E"=>'O',"\xF0\x9D\x93\x9F"=>'P',"\xF0\x9D\x93\xA0"=>'Q',
  "\xF0\x9D\x93\xA1"=>'R',"\xF0\x9D\x93\xA2"=>'S',"\xF0\x9D\x93\xA3"=>'T',"\xF0\x9D\x93\xA4"=>'U',
  "\xF0\x9D\x93\xA5"=>'V',"\xF0\x9D\x93\xA6"=>'W',"\xF0\x9D\x93\xA7"=>'X',"\xF0\x9D\x93\xA8"=>'Y',
  "\xF0\x9D\x93\xA9"=>'Z',"\xF0\x9D\x93\xAA"=>'a',"\xF0\x9D\x93\xAB"=>'b',"\xF0\x9D\x93\xAC"=>'c',
  "\xF0\x9D\x93\xAD"=>'d',"\xF0\x9D\x93\xAE"=>'e',"\xF0\x9D\x93\xAF"=>'f',"\xF0\x9D\x93\xB0"=>'g',
  "\xF0\x9D\x93\xB1"=>'h',"\xF0\x9D\x93\xB2"=>'i',"\xF0\x9D\x93\xB3"=>'j',"\xF0\x9D\x93\xB4"=>'k',
  "\xF0\x9D\x93\xB5"=>'l',"\xF0\x9D\x93\xB6"=>'m',"\xF0\x9D\x93\xB7"=>'n',"\xF0\x9D\x93\xB8"=>'o',
  "\xF0\x9D\x93\xB9"=>'p',"\xF0\x9D\x93\xBA"=>'q',"\xF0\x9D\x93\xBB"=>'r',"\xF0\x9D\x93\xBC"=>'s',
  "\xF0\x9D\x93\xBD"=>'t',"\xF0\x9D\x93\xBE"=>'u',"\xF0\x9D\x93\xBF"=>'v',"\xF0\x9D\x94\x80"=>'w',
  "\xF0\x9D\x94\x81"=>'x',"\xF0\x9D\x94\x82"=>'y',"\xF0\x9D\x94\x83"=>'z',"\xF0\x9D\x94\x84"=>'A',
  "\xF0\x9D\x94\x85"=>'B',"\xF0\x9D\x94\x87"=>'D',"\xF0\x9D\x94\x88"=>'E',"\xF0\x9D\x94\x89"=>'F',
  "\xF0\x9D\x94\x8A"=>'G',"\xF0\x9D\x94\x8D"=>'J',"\xF0\x9D\x94\x8E"=>'K',"\xF0\x9D\x94\x8F"=>'L',
  "\xF0\x9D\x94\x90"=>'M',"\xF0\x9D\x94\x91"=>'N',"\xF0\x9D\x94\x92"=>'O',"\xF0\x9D\x94\x93"=>'P',
  "\xF0\x9D\x94\x94"=>'Q',"\xF0\x9D\x94\x96"=>'S',"\xF0\x9D\x94\x97"=>'T',"\xF0\x9D\x94\x98"=>'U',
  "\xF0\x9D\x94\x99"=>'V',"\xF0\x9D\x94\x9A"=>'W',"\xF0\x9D\x94\x9B"=>'X',"\xF0\x9D\x94\x9C"=>'Y',
  "\xF0\x9D\x94\x9E"=>'a',"\xF0\x9D\x94\x9F"=>'b',"\xF0\x9D\x94\xA0"=>'c',"\xF0\x9D\x94\xA1"=>'d',
  "\xF0\x9D\x94\xA2"=>'e',"\xF0\x9D\x94\xA3"=>'f',"\xF0\x9D\x94\xA4"=>'g',"\xF0\x9D\x94\xA5"=>'h',
  "\xF0\x9D\x94\xA6"=>'i',"\xF0\x9D\x94\xA7"=>'j',"\xF0\x9D\x94\xA8"=>'k',"\xF0\x9D\x94\xA9"=>'l',
  "\xF0\x9D\x94\xAA"=>'m',"\xF0\x9D\x94\xAB"=>'n',"\xF0\x9D\x94\xAC"=>'o',"\xF0\x9D\x94\xAD"=>'p',
  "\xF0\x9D\x94\xAE"=>'q',"\xF0\x9D\x94\xAF"=>'r',"\xF0\x9D\x94\xB0"=>'s',"\xF0\x9D\x94\xB1"=>'t',
  "\xF0\x9D\x94\xB2"=>'u',"\xF0\x9D\x94\xB3"=>'v',"\xF0\x9D\x94\xB4"=>'w',"\xF0\x9D\x94\xB5"=>'x',
  "\xF0\x9D\x94\xB6"=>'y',"\xF0\x9D\x94\xB7"=>'z',"\xF0\x9D\x94\xB8"=>'A',"\xF0\x9D\x94\xB9"=>'B',
  "\xF0\x9D\x94\xBB"=>'D',"\xF0\x9D\x94\xBC"=>'E',"\xF0\x9D\x94\xBD"=>'F',"\xF0\x9D\x94\xBE"=>'G',
  "\xF0\x9D\x95\x80"=>'I',"\xF0\x9D\x95\x81"=>'J',"\xF0\x9D\x95\x82"=>'K',"\xF0\x9D\x95\x83"=>'L',
  "\xF0\x9D\x95\x84"=>'M',"\xF0\x9D\x95\x86"=>'O',"\xF0\x9D\x95\x8A"=>'S',"\xF0\x9D\x95\x8B"=>'T',
  "\xF0\x9D\x95\x8C"=>'U',"\xF0\x9D\x95\x8D"=>'V',"\xF0\x9D\x95\x8E"=>'W',"\xF0\x9D\x95\x8F"=>'X',
  "\xF0\x9D\x95\x90"=>'Y',"\xF0\x9D\x95\x92"=>'a',"\xF0\x9D\x95\x93"=>'b',"\xF0\x9D\x95\x94"=>'c',
  "\xF0\x9D\x95\x95"=>'d',"\xF0\x9D\x95\x96"=>'e',"\xF0\x9D\x95\x97"=>'f',"\xF0\x9D\x95\x98"=>'g',
  "\xF0\x9D\x95\x99"=>'h',"\xF0\x9D\x95\x9A"=>'i',"\xF0\x9D\x95\x9B"=>'j',"\xF0\x9D\x95\x9C"=>'k',
  "\xF0\x9D\x95\x9D"=>'l',"\xF0\x9D\x95\x9E"=>'m',"\xF0\x9D\x95\x9F"=>'n',"\xF0\x9D\x95\xA0"=>'o',
  "\xF0\x9D\x95\xA1"=>'p',"\xF0\x9D\x95\xA2"=>'q',"\xF0\x9D\x95\xA3"=>'r',"\xF0\x9D\x95\xA4"=>'s',
  "\xF0\x9D\x95\xA5"=>'t',"\xF0\x9D\x95\xA6"=>'u',"\xF0\x9D\x95\xA7"=>'v',"\xF0\x9D\x95\xA8"=>'w',
  "\xF0\x9D\x95\xA9"=>'x',"\xF0\x9D\x95\xAA"=>'y',"\xF0\x9D\x95\xAB"=>'z',"\xF0\x9D\x95\xAC"=>'A',
  "\xF0\x9D\x95\xAD"=>'B',"\xF0\x9D\x95\xAE"=>'C',"\xF0\x9D\x95\xAF"=>'D',"\xF0\x9D\x95\xB0"=>'E',
  "\xF0\x9D\x95\xB1"=>'F',"\xF0\x9D\x95\xB2"=>'G',"\xF0\x9D\x95\xB3"=>'H',"\xF0\x9D\x95\xB4"=>'I',
  "\xF0\x9D\x95\xB5"=>'J',"\xF0\x9D\x95\xB6"=>'K',"\xF0\x9D\x95\xB7"=>'L',"\xF0\x9D\x95\xB8"=>'M',
  "\xF0\x9D\x95\xB9"=>'N',"\xF0\x9D\x95\xBA"=>'O',"\xF0\x9D\x95\xBB"=>'P',"\xF0\x9D\x95\xBC"=>'Q',
  "\xF0\x9D\x95\xBD"=>'R',"\xF0\x9D\x95\xBE"=>'S',"\xF0\x9D\x95\xBF"=>'T',"\xF0\x9D\x96\x80"=>'U',
  "\xF0\x9D\x96\x81"=>'V',"\xF0\x9D\x96\x82"=>'W',"\xF0\x9D\x96\x83"=>'X',"\xF0\x9D\x96\x84"=>'Y',
  "\xF0\x9D\x96\x85"=>'Z',"\xF0\x9D\x96\x86"=>'a',"\xF0\x9D\x96\x87"=>'b',"\xF0\x9D\x96\x88"=>'c',
  "\xF0\x9D\x96\x89"=>'d',"\xF0\x9D\x96\x8A"=>'e',"\xF0\x9D\x96\x8B"=>'f',"\xF0\x9D\x96\x8C"=>'g',
  "\xF0\x9D\x96\x8D"=>'h',"\xF0\x9D\x96\x8E"=>'i',"\xF0\x9D\x96\x8F"=>'j',"\xF0\x9D\x96\x90"=>'k',
  "\xF0\x9D\x96\x91"=>'l',"\xF0\x9D\x96\x92"=>'m',"\xF0\x9D\x96\x93"=>'n',"\xF0\x9D\x96\x94"=>'o',
  "\xF0\x9D\x96\x95"=>'p',"\xF0\x9D\x96\x96"=>'q',"\xF0\x9D\x96\x97"=>'r',"\xF0\x9D\x96\x98"=>'s',
  "\xF0\x9D\x96\x99"=>'t',"\xF0\x9D\x96\x9A"=>'u',"\xF0\x9D\x96\x9B"=>'v',"\xF0\x9D\x96\x9C"=>'w',
  "\xF0\x9D\x96\x9D"=>'x',"\xF0\x9D\x96\x9E"=>'y',"\xF0\x9D\x96\x9F"=>'z',"\xF0\x9D\x96\xA0"=>'A',
  "\xF0\x9D\x96\xA1"=>'B',"\xF0\x9D\x96\xA2"=>'C',"\xF0\x9D\x96\xA3"=>'D',"\xF0\x9D\x96\xA4"=>'E',
  "\xF0\x9D\x96\xA5"=>'F',"\xF0\x9D\x96\xA6"=>'G',"\xF0\x9D\x96\xA7"=>'H',"\xF0\x9D\x96\xA8"=>'I',
  "\xF0\x9D\x96\xA9"=>'J',"\xF0\x9D\x96\xAA"=>'K',"\xF0\x9D\x96\xAB"=>'L',"\xF0\x9D\x96\xAC"=>'M',
  "\xF0\x9D\x96\xAD"=>'N',"\xF0\x9D\x96\xAE"=>'O',"\xF0\x9D\x96\xAF"=>'P',"\xF0\x9D\x96\xB0"=>'Q',
  "\xF0\x9D\x96\xB1"=>'R',"\xF0\x9D\x96\xB2"=>'S',"\xF0\x9D\x96\xB3"=>'T',"\xF0\x9D\x96\xB4"=>'U',
  "\xF0\x9D\x96\xB5"=>'V',"\xF0\x9D\x96\xB6"=>'W',"\xF0\x9D\x96\xB7"=>'X',"\xF0\x9D\x96\xB8"=>'Y',
  "\xF0\x9D\x96\xB9"=>'Z',"\xF0\x9D\x96\xBA"=>'a',"\xF0\x9D\x96\xBB"=>'b',"\xF0\x9D\x96\xBC"=>'c',
  "\xF0\x9D\x96\xBD"=>'d',"\xF0\x9D\x96\xBE"=>'e',"\xF0\x9D\x96\xBF"=>'f',"\xF0\x9D\x97\x80"=>'g',
  "\xF0\x9D\x97\x81"=>'h',"\xF0\x9D\x97\x82"=>'i',"\xF0\x9D\x97\x83"=>'j',"\xF0\x9D\x97\x84"=>'k',
  "\xF0\x9D\x97\x85"=>'l',"\xF0\x9D\x97\x86"=>'m',"\xF0\x9D\x97\x87"=>'n',"\xF0\x9D\x97\x88"=>'o',
  "\xF0\x9D\x97\x89"=>'p',"\xF0\x9D\x97\x8A"=>'q',"\xF0\x9D\x97\x8B"=>'r',"\xF0\x9D\x97\x8C"=>'s',
  "\xF0\x9D\x97\x8D"=>'t',"\xF0\x9D\x97\x8E"=>'u',"\xF0\x9D\x97\x8F"=>'v',"\xF0\x9D\x97\x90"=>'w',
  "\xF0\x9D\x97\x91"=>'x',"\xF0\x9D\x97\x92"=>'y',"\xF0\x9D\x97\x93"=>'z',"\xF0\x9D\x97\x94"=>'A',
  "\xF0\x9D\x97\x95"=>'B',"\xF0\x9D\x97\x96"=>'C',"\xF0\x9D\x97\x97"=>'D',"\xF0\x9D\x97\x98"=>'E',
  "\xF0\x9D\x97\x99"=>'F',"\xF0\x9D\x97\x9A"=>'G',"\xF0\x9D\x97\x9B"=>'H',"\xF0\x9D\x97\x9C"=>'I',
  "\xF0\x9D\x97\x9D"=>'J',"\xF0\x9D\x97\x9E"=>'K',"\xF0\x9D\x97\x9F"=>'L',"\xF0\x9D\x97\xA0"=>'M',
  "\xF0\x9D\x97\xA1"=>'N',"\xF0\x9D\x97\xA2"=>'O',"\xF0\x9D\x97\xA3"=>'P',"\xF0\x9D\x97\xA4"=>'Q',
  "\xF0\x9D\x97\xA5"=>'R',"\xF0\x9D\x97\xA6"=>'S',"\xF0\x9D\x97\xA7"=>'T',"\xF0\x9D\x97\xA8"=>'U',
  "\xF0\x9D\x97\xA9"=>'V',"\xF0\x9D\x97\xAA"=>'W',"\xF0\x9D\x97\xAB"=>'X',"\xF0\x9D\x97\xAC"=>'Y',
  "\xF0\x9D\x97\xAD"=>'Z',"\xF0\x9D\x97\xAE"=>'a',"\xF0\x9D\x97\xAF"=>'b',"\xF0\x9D\x97\xB0"=>'c',
  "\xF0\x9D\x97\xB1"=>'d',"\xF0\x9D\x97\xB2"=>'e',"\xF0\x9D\x97\xB3"=>'f',"\xF0\x9D\x97\xB4"=>'g',
  "\xF0\x9D\x97\xB5"=>'h',"\xF0\x9D\x97\xB6"=>'i',"\xF0\x9D\x97\xB7"=>'j',"\xF0\x9D\x97\xB8"=>'k',
  "\xF0\x9D\x97\xB9"=>'l',"\xF0\x9D\x97\xBA"=>'m',"\xF0\x9D\x97\xBB"=>'n',"\xF0\x9D\x97\xBC"=>'o',
  "\xF0\x9D\x97\xBD"=>'p',"\xF0\x9D\x97\xBE"=>'q',"\xF0\x9D\x97\xBF"=>'r',"\xF0\x9D\x98\x80"=>'s',
  "\xF0\x9D\x98\x81"=>'t',"\xF0\x9D\x98\x82"=>'u',"\xF0\x9D\x98\x83"=>'v',"\xF0\x9D\x98\x84"=>'w',
  "\xF0\x9D\x98\x85"=>'x',"\xF0\x9D\x98\x86"=>'y',"\xF0\x9D\x98\x87"=>'z',"\xF0\x9D\x98\x88"=>'A',
  "\xF0\x9D\x98\x89"=>'B',"\xF0\x9D\x98\x8A"=>'C',"\xF0\x9D\x98\x8B"=>'D',"\xF0\x9D\x98\x8C"=>'E',
  "\xF0\x9D\x98\x8D"=>'F',"\xF0\x9D\x98\x8E"=>'G',"\xF0\x9D\x98\x8F"=>'H',"\xF0\x9D\x98\x90"=>'I',
  "\xF0\x9D\x98\x91"=>'J',"\xF0\x9D\x98\x92"=>'K',"\xF0\x9D\x98\x93"=>'L',"\xF0\x9D\x98\x94"=>'M',
  "\xF0\x9D\x98\x95"=>'N',"\xF0\x9D\x98\x96"=>'O',"\xF0\x9D\x98\x97"=>'P',"\xF0\x9D\x98\x98"=>'Q',
  "\xF0\x9D\x98\x99"=>'R',"\xF0\x9D\x98\x9A"=>'S',"\xF0\x9D\x98\x9B"=>'T',"\xF0\x9D\x98\x9C"=>'U',
  "\xF0\x9D\x98\x9D"=>'V',"\xF0\x9D\x98\x9E"=>'W',"\xF0\x9D\x98\x9F"=>'X',"\xF0\x9D\x98\xA0"=>'Y',
  "\xF0\x9D\x98\xA1"=>'Z',"\xF0\x9D\x98\xA2"=>'a',"\xF0\x9D\x98\xA3"=>'b',"\xF0\x9D\x98\xA4"=>'c',
  "\xF0\x9D\x98\xA5"=>'d',"\xF0\x9D\x98\xA6"=>'e',"\xF0\x9D\x98\xA7"=>'f',"\xF0\x9D\x98\xA8"=>'g',
  "\xF0\x9D\x98\xA9"=>'h',"\xF0\x9D\x98\xAA"=>'i',"\xF0\x9D\x98\xAB"=>'j',"\xF0\x9D\x98\xAC"=>'k',
  "\xF0\x9D\x98\xAD"=>'l',"\xF0\x9D\x98\xAE"=>'m',"\xF0\x9D\x98\xAF"=>'n',"\xF0\x9D\x98\xB0"=>'o',
  "\xF0\x9D\x98\xB1"=>'p',"\xF0\x9D\x98\xB2"=>'q',"\xF0\x9D\x98\xB3"=>'r',"\xF0\x9D\x98\xB4"=>'s',
  "\xF0\x9D\x98\xB5"=>'t',"\xF0\x9D\x98\xB6"=>'u',"\xF0\x9D\x98\xB7"=>'v',"\xF0\x9D\x98\xB8"=>'w',
  "\xF0\x9D\x98\xB9"=>'x',"\xF0\x9D\x98\xBA"=>'y',"\xF0\x9D\x98\xBB"=>'z',"\xF0\x9D\x98\xBC"=>'A',
  "\xF0\x9D\x98\xBD"=>'B',"\xF0\x9D\x98\xBE"=>'C',"\xF0\x9D\x98\xBF"=>'D',"\xF0\x9D\x99\x80"=>'E',
  "\xF0\x9D\x99\x81"=>'F',"\xF0\x9D\x99\x82"=>'G',"\xF0\x9D\x99\x83"=>'H',"\xF0\x9D\x99\x84"=>'I',
  "\xF0\x9D\x99\x85"=>'J',"\xF0\x9D\x99\x86"=>'K',"\xF0\x9D\x99\x87"=>'L',"\xF0\x9D\x99\x88"=>'M',
  "\xF0\x9D\x99\x89"=>'N',"\xF0\x9D\x99\x8A"=>'O',"\xF0\x9D\x99\x8B"=>'P',"\xF0\x9D\x99\x8C"=>'Q',
  "\xF0\x9D\x99\x8D"=>'R',"\xF0\x9D\x99\x8E"=>'S',"\xF0\x9D\x99\x8F"=>'T',"\xF0\x9D\x99\x90"=>'U',
  "\xF0\x9D\x99\x91"=>'V',"\xF0\x9D\x99\x92"=>'W',"\xF0\x9D\x99\x93"=>'X',"\xF0\x9D\x99\x94"=>'Y',
  "\xF0\x9D\x99\x95"=>'Z',"\xF0\x9D\x99\x96"=>'a',"\xF0\x9D\x99\x97"=>'b',"\xF0\x9D\x99\x98"=>'c',
  "\xF0\x9D\x99\x99"=>'d',"\xF0\x9D\x99\x9A"=>'e',"\xF0\x9D\x99\x9B"=>'f',"\xF0\x9D\x99\x9C"=>'g',
  "\xF0\x9D\x99\x9D"=>'h',"\xF0\x9D\x99\x9E"=>'i',"\xF0\x9D\x99\x9F"=>'j',"\xF0\x9D\x99\xA0"=>'k',
  "\xF0\x9D\x99\xA1"=>'l',"\xF0\x9D\x99\xA2"=>'m',"\xF0\x9D\x99\xA3"=>'n',"\xF0\x9D\x99\xA4"=>'o',
  "\xF0\x9D\x99\xA5"=>'p',"\xF0\x9D\x99\xA6"=>'q',"\xF0\x9D\x99\xA7"=>'r',"\xF0\x9D\x99\xA8"=>'s',
  "\xF0\x9D\x99\xA9"=>'t',"\xF0\x9D\x99\xAA"=>'u',"\xF0\x9D\x99\xAB"=>'v',"\xF0\x9D\x99\xAC"=>'w',
  "\xF0\x9D\x99\xAD"=>'x',"\xF0\x9D\x99\xAE"=>'y',"\xF0\x9D\x99\xAF"=>'z',"\xF0\x9D\x99\xB0"=>'A',
  "\xF0\x9D\x99\xB1"=>'B',"\xF0\x9D\x99\xB2"=>'C',"\xF0\x9D\x99\xB3"=>'D',"\xF0\x9D\x99\xB4"=>'E',
  "\xF0\x9D\x99\xB5"=>'F',"\xF0\x9D\x99\xB6"=>'G',"\xF0\x9D\x99\xB7"=>'H',"\xF0\x9D\x99\xB8"=>'I',
  "\xF0\x9D\x99\xB9"=>'J',"\xF0\x9D\x99\xBA"=>'K',"\xF0\x9D\x99\xBB"=>'L',"\xF0\x9D\x99\xBC"=>'M',
  "\xF0\x9D\x99\xBD"=>'N',"\xF0\x9D\x99\xBE"=>'O',"\xF0\x9D\x99\xBF"=>'P',"\xF0\x9D\x9A\x80"=>'Q',
  "\xF0\x9D\x9A\x81"=>'R',"\xF0\x9D\x9A\x82"=>'S',"\xF0\x9D\x9A\x83"=>'T',"\xF0\x9D\x9A\x84"=>'U',
  "\xF0\x9D\x9A\x85"=>'V',"\xF0\x9D\x9A\x86"=>'W',"\xF0\x9D\x9A\x87"=>'X',"\xF0\x9D\x9A\x88"=>'Y',
  "\xF0\x9D\x9A\x89"=>'Z',"\xF0\x9D\x9A\x8A"=>'a',"\xF0\x9D\x9A\x8B"=>'b',"\xF0\x9D\x9A\x8C"=>'c',
  "\xF0\x9D\x9A\x8D"=>'d',"\xF0\x9D\x9A\x8E"=>'e',"\xF0\x9D\x9A\x8F"=>'f',"\xF0\x9D\x9A\x90"=>'g',
  "\xF0\x9D\x9A\x91"=>'h',"\xF0\x9D\x9A\x92"=>'i',"\xF0\x9D\x9A\x93"=>'j',"\xF0\x9D\x9A\x94"=>'k',
  "\xF0\x9D\x9A\x95"=>'l',"\xF0\x9D\x9A\x96"=>'m',"\xF0\x9D\x9A\x97"=>'n',"\xF0\x9D\x9A\x98"=>'o',
  "\xF0\x9D\x9A\x99"=>'p',"\xF0\x9D\x9A\x9A"=>'q',"\xF0\x9D\x9A\x9B"=>'r',"\xF0\x9D\x9A\x9C"=>'s',
  "\xF0\x9D\x9A\x9D"=>'t',"\xF0\x9D\x9A\x9E"=>'u',"\xF0\x9D\x9A\x9F"=>'v',"\xF0\x9D\x9A\xA0"=>'w',
  "\xF0\x9D\x9A\xA1"=>'x',"\xF0\x9D\x9A\xA2"=>'y',"\xF0\x9D\x9A\xA3"=>'z',"\xF0\x9D\x9F\x8F"=>'1',
  "\xF0\x9D\x9F\x90"=>'2',"\xF0\x9D\x9F\x91"=>'3',"\xF0\x9D\x9F\x92"=>'4',"\xF0\x9D\x9F\x93"=>'5',
  "\xF0\x9D\x9F\x94"=>'6',"\xF0\x9D\x9F\x95"=>'7',"\xF0\x9D\x9F\x96"=>'8',"\xF0\x9D\x9F\x97"=>'9',
  "\xF0\x9D\x9F\x99"=>'1',"\xF0\x9D\x9F\x9A"=>'2',"\xF0\x9D\x9F\x9B"=>'3',"\xF0\x9D\x9F\x9C"=>'4',
  "\xF0\x9D\x9F\x9D"=>'5',"\xF0\x9D\x9F\x9E"=>'6',"\xF0\x9D\x9F\x9F"=>'7',"\xF0\x9D\x9F\xA0"=>'8',
  "\xF0\x9D\x9F\xA1"=>'9',"\xF0\x9D\x9F\xA3"=>'1',"\xF0\x9D\x9F\xA4"=>'2',"\xF0\x9D\x9F\xA5"=>'3',
  "\xF0\x9D\x9F\xA6"=>'4',"\xF0\x9D\x9F\xA7"=>'5',"\xF0\x9D\x9F\xA8"=>'6',"\xF0\x9D\x9F\xA9"=>'7',
  "\xF0\x9D\x9F\xAA"=>'8',"\xF0\x9D\x9F\xAB"=>'9',"\xF0\x9D\x9F\xAD"=>'1',"\xF0\x9D\x9F\xAE"=>'2',
  "\xF0\x9D\x9F\xAF"=>'3',"\xF0\x9D\x9F\xB0"=>'4',"\xF0\x9D\x9F\xB1"=>'5',"\xF0\x9D\x9F\xB2"=>'6',
  "\xF0\x9D\x9F\xB3"=>'7',"\xF0\x9D\x9F\xB4"=>'8',"\xF0\x9D\x9F\xB5"=>'9',"\xF0\x9D\x9F\xB7"=>'1',
  "\xF0\x9D\x9F\xB8"=>'2',"\xF0\x9D\x9F\xB9"=>'3',"\xF0\x9D\x9F\xBA"=>'4',"\xF0\x9D\x9F\xBB"=>'5',
  "\xF0\x9D\x9F\xBC"=>'6',"\xF0\x9D\x9F\xBD"=>'7',"\xF0\x9D\x9F\xBE"=>'8',"\xF0\x9D\x9F\xBF"=>'9',
  "\xF0\x9F\x84\x80"=>'0.',"\xF0\x9F\x84\x81"=>'0,',"\xF0\x9F\x84\x82"=>'1,',"\xF0\x9F\x84\x83"=>'2,',
  "\xF0\x9F\x84\x84"=>'3,',"\xF0\x9F\x84\x85"=>'4,',"\xF0\x9F\x84\x86"=>'5,',"\xF0\x9F\x84\x87"=>'6,',
  "\xF0\x9F\x84\x88"=>'7,',"\xF0\x9F\x84\x89"=>'8,',"\xF0\x9F\x84\x8A"=>'9,',"\xF0\x9F\x84\x90"=>'(A)',
  "\xF0\x9F\x84\x91"=>'(B)',"\xF0\x9F\x84\x92"=>'(C)',"\xF0\x9F\x84\x93"=>'(D)',"\xF0\x9F\x84\x94"=>'(E)',
  "\xF0\x9F\x84\x95"=>'(F)',"\xF0\x9F\x84\x96"=>'(G)',"\xF0\x9F\x84\x97"=>'(H)',"\xF0\x9F\x84\x98"=>'(I)',
  "\xF0\x9F\x84\x99"=>'(J)',"\xF0\x9F\x84\x9A"=>'(K)',"\xF0\x9F\x84\x9B"=>'(L)',"\xF0\x9F\x84\x9C"=>'(M)',
  "\xF0\x9F\x84\x9D"=>'(N)',"\xF0\x9F\x84\x9E"=>'(O)',"\xF0\x9F\x84\x9F"=>'(P)',"\xF0\x9F\x84\xA0"=>'(Q)',
  "\xF0\x9F\x84\xA1"=>'(R)',"\xF0\x9F\x84\xA2"=>'(S)',"\xF0\x9F\x84\xA3"=>'(T)',"\xF0\x9F\x84\xA4"=>'(U)',
  "\xF0\x9F\x84\xA5"=>'(V)',"\xF0\x9F\x84\xA6"=>'(W)',"\xF0\x9F\x84\xA7"=>'(X)',"\xF0\x9F\x84\xA8"=>'(Y)',
  "\xF0\x9F\x84\xA9"=>'(Z)',"\xF0\x9F\x84\xAA"=>'S',"\xF0\x9F\x84\xAB"=>'C',"\xF0\x9F\x84\xAC"=>'R',
  "\xF0\x9F\x84\xAD"=>'CD',"\xF0\x9F\x84\xAE"=>'WZ',"\xF0\x9F\x84\xB0"=>'A',"\xF0\x9F\x84\xB1"=>'B',
  "\xF0\x9F\x84\xB2"=>'C',"\xF0\x9F\x84\xB3"=>'D',"\xF0\x9F\x84\xB4"=>'E',"\xF0\x9F\x84\xB5"=>'F',
  "\xF0\x9F\x84\xB6"=>'G',"\xF0\x9F\x84\xB7"=>'H',"\xF0\x9F\x84\xB8"=>'I',"\xF0\x9F\x84\xB9"=>'J',
  "\xF0\x9F\x84\xBA"=>'K',"\xF0\x9F\x84\xBB"=>'L',"\xF0\x9F\x84\xBC"=>'M',"\xF0\x9F\x84\xBD"=>'N',
  "\xF0\x9F\x84\xBE"=>'O',"\xF0\x9F\x84\xBF"=>'P',"\xF0\x9F\x85\x80"=>'Q',"\xF0\x9F\x85\x81"=>'R',
  "\xF0\x9F\x85\x82"=>'S',"\xF0\x9F\x85\x83"=>'T',"\xF0\x9F\x85\x84"=>'U',"\xF0\x9F\x85\x85"=>'V',
  "\xF0\x9F\x85\x86"=>'W',"\xF0\x9F\x85\x87"=>'X',"\xF0\x9F\x85\x88"=>'Y',"\xF0\x9F\x85\x89"=>'Z',
  "\xF0\x9F\x85\x8A"=>'HV',"\xF0\x9F\x85\x8B"=>'MV',"\xF0\x9F\x85\x8C"=>'SD',"\xF0\x9F\x85\x8D"=>'SS',
  "\xF0\x9F\x85\x8E"=>'PPV',"\xF0\x9F\x85\x8F"=>'WC',"\xF0\x9F\x86\x90"=>'DJ');
}


if (!USE_MBSTRING) {
    /** UTF-8 mapping from uppercase to lowercase
     *
     * This array derived from CaseFolding-6.0.0.txt
     * (part of the Unicode Character Database).
     * It maps UTF-8 encoded upper case codepoints to the
     * corresponding lower case codepoint using status
     * 'C' (common case folding) and 'F' (full case folding).
     * The special case for Turkic is not considered.
     *
     */
    global $UTF8_UPPER_LOWER;
    if (empty($UTF8_UPPER_LOWER)) { $UTF8_UPPER_LOWER = array(
'A'=>'a','B'=>'b','C'=>'c','D'=>'d','E'=>'e','F'=>'f','G'=>'g','H'=>'h','I'=>'i','J'=>'j','K'=>'k','L'=>'l','M'=>'m',
'N'=>'n','O'=>'o','P'=>'p','Q'=>'q','R'=>'r','S'=>'s','T'=>'t','U'=>'u','V'=>'v','W'=>'w','X'=>'x','Y'=>'y','Z'=>'z',
"\xC2\xB5"=>"\xCE\xBC","\xC3\x80"=>"\xC3\xA0","\xC3\x81"=>"\xC3\xA1","\xC3\x82"=>"\xC3\xA2","\xC3\x83"=>"\xC3\xA3",
"\xC3\x84"=>"\xC3\xA4","\xC3\x85"=>"\xC3\xA5","\xC3\x86"=>"\xC3\xA6","\xC3\x87"=>"\xC3\xA7","\xC3\x88"=>"\xC3\xA8",
"\xC3\x89"=>"\xC3\xA9","\xC3\x8A"=>"\xC3\xAA","\xC3\x8B"=>"\xC3\xAB","\xC3\x8C"=>"\xC3\xAC","\xC3\x8D"=>"\xC3\xAD",
"\xC3\x8E"=>"\xC3\xAE","\xC3\x8F"=>"\xC3\xAF","\xC3\x90"=>"\xC3\xB0","\xC3\x91"=>"\xC3\xB1","\xC3\x92"=>"\xC3\xB2",
"\xC3\x93"=>"\xC3\xB3","\xC3\x94"=>"\xC3\xB4","\xC3\x95"=>"\xC3\xB5","\xC3\x96"=>"\xC3\xB6","\xC3\x98"=>"\xC3\xB8",
"\xC3\x99"=>"\xC3\xB9","\xC3\x9A"=>"\xC3\xBA","\xC3\x9B"=>"\xC3\xBB","\xC3\x9C"=>"\xC3\xBC","\xC3\x9D"=>"\xC3\xBD",
"\xC3\x9E"=>"\xC3\xBE","\xC3\x9F"=>'ss',"\xC4\x80"=>"\xC4\x81","\xC4\x82"=>"\xC4\x83","\xC4\x84"=>"\xC4\x85",
"\xC4\x86"=>"\xC4\x87","\xC4\x88"=>"\xC4\x89","\xC4\x8A"=>"\xC4\x8B","\xC4\x8C"=>"\xC4\x8D","\xC4\x8E"=>"\xC4\x8F",
"\xC4\x90"=>"\xC4\x91","\xC4\x92"=>"\xC4\x93","\xC4\x94"=>"\xC4\x95","\xC4\x96"=>"\xC4\x97","\xC4\x98"=>"\xC4\x99",
"\xC4\x9A"=>"\xC4\x9B","\xC4\x9C"=>"\xC4\x9D","\xC4\x9E"=>"\xC4\x9F","\xC4\xA0"=>"\xC4\xA1","\xC4\xA2"=>"\xC4\xA3",
"\xC4\xA4"=>"\xC4\xA5","\xC4\xA6"=>"\xC4\xA7","\xC4\xA8"=>"\xC4\xA9","\xC4\xAA"=>"\xC4\xAB","\xC4\xAC"=>"\xC4\xAD",
"\xC4\xAE"=>"\xC4\xAF","\xC4\xB0"=>"i\xCC\x87","\xC4\xB2"=>"\xC4\xB3","\xC4\xB4"=>"\xC4\xB5","\xC4\xB6"=>"\xC4\xB7",
"\xC4\xB9"=>"\xC4\xBA","\xC4\xBB"=>"\xC4\xBC","\xC4\xBD"=>"\xC4\xBE","\xC4\xBF"=>"\xC5\x80","\xC5\x81"=>"\xC5\x82",
"\xC5\x83"=>"\xC5\x84","\xC5\x85"=>"\xC5\x86","\xC5\x87"=>"\xC5\x88","\xC5\x89"=>"\xCA\xBCn","\xC5\x8A"=>"\xC5\x8B",
"\xC5\x8C"=>"\xC5\x8D","\xC5\x8E"=>"\xC5\x8F","\xC5\x90"=>"\xC5\x91","\xC5\x92"=>"\xC5\x93","\xC5\x94"=>"\xC5\x95",
"\xC5\x96"=>"\xC5\x97","\xC5\x98"=>"\xC5\x99","\xC5\x9A"=>"\xC5\x9B","\xC5\x9C"=>"\xC5\x9D","\xC5\x9E"=>"\xC5\x9F",
"\xC5\xA0"=>"\xC5\xA1","\xC5\xA2"=>"\xC5\xA3","\xC5\xA4"=>"\xC5\xA5","\xC5\xA6"=>"\xC5\xA7","\xC5\xA8"=>"\xC5\xA9",
"\xC5\xAA"=>"\xC5\xAB","\xC5\xAC"=>"\xC5\xAD","\xC5\xAE"=>"\xC5\xAF","\xC5\xB0"=>"\xC5\xB1","\xC5\xB2"=>"\xC5\xB3",
"\xC5\xB4"=>"\xC5\xB5","\xC5\xB6"=>"\xC5\xB7","\xC5\xB8"=>"\xC3\xBF","\xC5\xB9"=>"\xC5\xBA","\xC5\xBB"=>"\xC5\xBC",
"\xC5\xBD"=>"\xC5\xBE","\xC5\xBF"=>'s',"\xC6\x81"=>"\xC9\x93","\xC6\x82"=>"\xC6\x83","\xC6\x84"=>"\xC6\x85",
"\xC6\x86"=>"\xC9\x94","\xC6\x87"=>"\xC6\x88","\xC6\x89"=>"\xC9\x96","\xC6\x8A"=>"\xC9\x97","\xC6\x8B"=>"\xC6\x8C",
"\xC6\x8E"=>"\xC7\x9D","\xC6\x8F"=>"\xC9\x99","\xC6\x90"=>"\xC9\x9B","\xC6\x91"=>"\xC6\x92","\xC6\x93"=>"\xC9\xA0",
"\xC6\x94"=>"\xC9\xA3","\xC6\x96"=>"\xC9\xA9","\xC6\x97"=>"\xC9\xA8","\xC6\x98"=>"\xC6\x99","\xC6\x9C"=>"\xC9\xAF",
"\xC6\x9D"=>"\xC9\xB2","\xC6\x9F"=>"\xC9\xB5","\xC6\xA0"=>"\xC6\xA1","\xC6\xA2"=>"\xC6\xA3","\xC6\xA4"=>"\xC6\xA5",
"\xC6\xA6"=>"\xCA\x80","\xC6\xA7"=>"\xC6\xA8","\xC6\xA9"=>"\xCA\x83","\xC6\xAC"=>"\xC6\xAD","\xC6\xAE"=>"\xCA\x88",
"\xC6\xAF"=>"\xC6\xB0","\xC6\xB1"=>"\xCA\x8A","\xC6\xB2"=>"\xCA\x8B","\xC6\xB3"=>"\xC6\xB4","\xC6\xB5"=>"\xC6\xB6",
"\xC6\xB7"=>"\xCA\x92","\xC6\xB8"=>"\xC6\xB9","\xC6\xBC"=>"\xC6\xBD","\xC7\x84"=>"\xC7\x86","\xC7\x85"=>"\xC7\x86",
"\xC7\x87"=>"\xC7\x89","\xC7\x88"=>"\xC7\x89","\xC7\x8A"=>"\xC7\x8C","\xC7\x8B"=>"\xC7\x8C","\xC7\x8D"=>"\xC7\x8E",
"\xC7\x8F"=>"\xC7\x90","\xC7\x91"=>"\xC7\x92","\xC7\x93"=>"\xC7\x94","\xC7\x95"=>"\xC7\x96","\xC7\x97"=>"\xC7\x98",
"\xC7\x99"=>"\xC7\x9A","\xC7\x9B"=>"\xC7\x9C","\xC7\x9E"=>"\xC7\x9F","\xC7\xA0"=>"\xC7\xA1","\xC7\xA2"=>"\xC7\xA3",
"\xC7\xA4"=>"\xC7\xA5","\xC7\xA6"=>"\xC7\xA7","\xC7\xA8"=>"\xC7\xA9","\xC7\xAA"=>"\xC7\xAB","\xC7\xAC"=>"\xC7\xAD",
"\xC7\xAE"=>"\xC7\xAF","\xC7\xB0"=>"j\xCC\x8C","\xC7\xB1"=>"\xC7\xB3","\xC7\xB2"=>"\xC7\xB3","\xC7\xB4"=>"\xC7\xB5",
"\xC7\xB6"=>"\xC6\x95","\xC7\xB7"=>"\xC6\xBF","\xC7\xB8"=>"\xC7\xB9","\xC7\xBA"=>"\xC7\xBB","\xC7\xBC"=>"\xC7\xBD",
"\xC7\xBE"=>"\xC7\xBF","\xC8\x80"=>"\xC8\x81","\xC8\x82"=>"\xC8\x83","\xC8\x84"=>"\xC8\x85","\xC8\x86"=>"\xC8\x87",
"\xC8\x88"=>"\xC8\x89","\xC8\x8A"=>"\xC8\x8B","\xC8\x8C"=>"\xC8\x8D","\xC8\x8E"=>"\xC8\x8F","\xC8\x90"=>"\xC8\x91",
"\xC8\x92"=>"\xC8\x93","\xC8\x94"=>"\xC8\x95","\xC8\x96"=>"\xC8\x97","\xC8\x98"=>"\xC8\x99","\xC8\x9A"=>"\xC8\x9B",
"\xC8\x9C"=>"\xC8\x9D","\xC8\x9E"=>"\xC8\x9F","\xC8\xA0"=>"\xC6\x9E","\xC8\xA2"=>"\xC8\xA3","\xC8\xA4"=>"\xC8\xA5",
"\xC8\xA6"=>"\xC8\xA7","\xC8\xA8"=>"\xC8\xA9","\xC8\xAA"=>"\xC8\xAB","\xC8\xAC"=>"\xC8\xAD","\xC8\xAE"=>"\xC8\xAF",
"\xC8\xB0"=>"\xC8\xB1","\xC8\xB2"=>"\xC8\xB3","\xC8\xBA"=>"\xE2\xB1\xA5","\xC8\xBB"=>"\xC8\xBC","\xC8\xBD"=>"\xC6\x9A",
"\xC8\xBE"=>"\xE2\xB1\xA6","\xC9\x81"=>"\xC9\x82","\xC9\x83"=>"\xC6\x80","\xC9\x84"=>"\xCA\x89","\xC9\x85"=>"\xCA\x8C",
"\xC9\x86"=>"\xC9\x87","\xC9\x88"=>"\xC9\x89","\xC9\x8A"=>"\xC9\x8B","\xC9\x8C"=>"\xC9\x8D","\xC9\x8E"=>"\xC9\x8F",
"\xCD\x85"=>"\xCE\xB9","\xCD\xB0"=>"\xCD\xB1","\xCD\xB2"=>"\xCD\xB3","\xCD\xB6"=>"\xCD\xB7","\xCE\x86"=>"\xCE\xAC",
"\xCE\x88"=>"\xCE\xAD","\xCE\x89"=>"\xCE\xAE","\xCE\x8A"=>"\xCE\xAF","\xCE\x8C"=>"\xCF\x8C","\xCE\x8E"=>"\xCF\x8D",
"\xCE\x8F"=>"\xCF\x8E","\xCE\x90"=>"\xCE\xB9\xCC\x88\xCC\x81","\xCE\x91"=>"\xCE\xB1","\xCE\x92"=>"\xCE\xB2",
"\xCE\x93"=>"\xCE\xB3","\xCE\x94"=>"\xCE\xB4","\xCE\x95"=>"\xCE\xB5","\xCE\x96"=>"\xCE\xB6","\xCE\x97"=>"\xCE\xB7",
"\xCE\x98"=>"\xCE\xB8","\xCE\x99"=>"\xCE\xB9","\xCE\x9A"=>"\xCE\xBA","\xCE\x9B"=>"\xCE\xBB","\xCE\x9C"=>"\xCE\xBC",
"\xCE\x9D"=>"\xCE\xBD","\xCE\x9E"=>"\xCE\xBE","\xCE\x9F"=>"\xCE\xBF","\xCE\xA0"=>"\xCF\x80","\xCE\xA1"=>"\xCF\x81",
"\xCE\xA3"=>"\xCF\x83","\xCE\xA4"=>"\xCF\x84","\xCE\xA5"=>"\xCF\x85","\xCE\xA6"=>"\xCF\x86","\xCE\xA7"=>"\xCF\x87",
"\xCE\xA8"=>"\xCF\x88","\xCE\xA9"=>"\xCF\x89","\xCE\xAA"=>"\xCF\x8A","\xCE\xAB"=>"\xCF\x8B",
"\xCE\xB0"=>"\xCF\x85\xCC\x88\xCC\x81","\xCF\x82"=>"\xCF\x83","\xCF\x8F"=>"\xCF\x97","\xCF\x90"=>"\xCE\xB2",
"\xCF\x91"=>"\xCE\xB8","\xCF\x95"=>"\xCF\x86","\xCF\x96"=>"\xCF\x80","\xCF\x98"=>"\xCF\x99","\xCF\x9A"=>"\xCF\x9B",
"\xCF\x9C"=>"\xCF\x9D","\xCF\x9E"=>"\xCF\x9F","\xCF\xA0"=>"\xCF\xA1","\xCF\xA2"=>"\xCF\xA3","\xCF\xA4"=>"\xCF\xA5",
"\xCF\xA6"=>"\xCF\xA7","\xCF\xA8"=>"\xCF\xA9","\xCF\xAA"=>"\xCF\xAB","\xCF\xAC"=>"\xCF\xAD","\xCF\xAE"=>"\xCF\xAF",
"\xCF\xB0"=>"\xCE\xBA","\xCF\xB1"=>"\xCF\x81","\xCF\xB4"=>"\xCE\xB8","\xCF\xB5"=>"\xCE\xB5","\xCF\xB7"=>"\xCF\xB8",
"\xCF\xB9"=>"\xCF\xB2","\xCF\xBA"=>"\xCF\xBB","\xCF\xBD"=>"\xCD\xBB","\xCF\xBE"=>"\xCD\xBC","\xCF\xBF"=>"\xCD\xBD",
"\xD0\x80"=>"\xD1\x90","\xD0\x81"=>"\xD1\x91","\xD0\x82"=>"\xD1\x92","\xD0\x83"=>"\xD1\x93","\xD0\x84"=>"\xD1\x94",
"\xD0\x85"=>"\xD1\x95","\xD0\x86"=>"\xD1\x96","\xD0\x87"=>"\xD1\x97","\xD0\x88"=>"\xD1\x98","\xD0\x89"=>"\xD1\x99",
"\xD0\x8A"=>"\xD1\x9A","\xD0\x8B"=>"\xD1\x9B","\xD0\x8C"=>"\xD1\x9C","\xD0\x8D"=>"\xD1\x9D","\xD0\x8E"=>"\xD1\x9E",
"\xD0\x8F"=>"\xD1\x9F","\xD0\x90"=>"\xD0\xB0","\xD0\x91"=>"\xD0\xB1","\xD0\x92"=>"\xD0\xB2","\xD0\x93"=>"\xD0\xB3",
"\xD0\x94"=>"\xD0\xB4","\xD0\x95"=>"\xD0\xB5","\xD0\x96"=>"\xD0\xB6","\xD0\x97"=>"\xD0\xB7","\xD0\x98"=>"\xD0\xB8",
"\xD0\x99"=>"\xD0\xB9","\xD0\x9A"=>"\xD0\xBA","\xD0\x9B"=>"\xD0\xBB","\xD0\x9C"=>"\xD0\xBC","\xD0\x9D"=>"\xD0\xBD",
"\xD0\x9E"=>"\xD0\xBE","\xD0\x9F"=>"\xD0\xBF","\xD0\xA0"=>"\xD1\x80","\xD0\xA1"=>"\xD1\x81","\xD0\xA2"=>"\xD1\x82",
"\xD0\xA3"=>"\xD1\x83","\xD0\xA4"=>"\xD1\x84","\xD0\xA5"=>"\xD1\x85","\xD0\xA6"=>"\xD1\x86","\xD0\xA7"=>"\xD1\x87",
"\xD0\xA8"=>"\xD1\x88","\xD0\xA9"=>"\xD1\x89","\xD0\xAA"=>"\xD1\x8A","\xD0\xAB"=>"\xD1\x8B","\xD0\xAC"=>"\xD1\x8C",
"\xD0\xAD"=>"\xD1\x8D","\xD0\xAE"=>"\xD1\x8E","\xD0\xAF"=>"\xD1\x8F","\xD1\xA0"=>"\xD1\xA1","\xD1\xA2"=>"\xD1\xA3",
"\xD1\xA4"=>"\xD1\xA5","\xD1\xA6"=>"\xD1\xA7","\xD1\xA8"=>"\xD1\xA9","\xD1\xAA"=>"\xD1\xAB","\xD1\xAC"=>"\xD1\xAD",
"\xD1\xAE"=>"\xD1\xAF","\xD1\xB0"=>"\xD1\xB1","\xD1\xB2"=>"\xD1\xB3","\xD1\xB4"=>"\xD1\xB5","\xD1\xB6"=>"\xD1\xB7",
"\xD1\xB8"=>"\xD1\xB9","\xD1\xBA"=>"\xD1\xBB","\xD1\xBC"=>"\xD1\xBD","\xD1\xBE"=>"\xD1\xBF","\xD2\x80"=>"\xD2\x81",
"\xD2\x8A"=>"\xD2\x8B","\xD2\x8C"=>"\xD2\x8D","\xD2\x8E"=>"\xD2\x8F","\xD2\x90"=>"\xD2\x91","\xD2\x92"=>"\xD2\x93",
"\xD2\x94"=>"\xD2\x95","\xD2\x96"=>"\xD2\x97","\xD2\x98"=>"\xD2\x99","\xD2\x9A"=>"\xD2\x9B","\xD2\x9C"=>"\xD2\x9D",
"\xD2\x9E"=>"\xD2\x9F","\xD2\xA0"=>"\xD2\xA1","\xD2\xA2"=>"\xD2\xA3","\xD2\xA4"=>"\xD2\xA5","\xD2\xA6"=>"\xD2\xA7",
"\xD2\xA8"=>"\xD2\xA9","\xD2\xAA"=>"\xD2\xAB","\xD2\xAC"=>"\xD2\xAD","\xD2\xAE"=>"\xD2\xAF","\xD2\xB0"=>"\xD2\xB1",
"\xD2\xB2"=>"\xD2\xB3","\xD2\xB4"=>"\xD2\xB5","\xD2\xB6"=>"\xD2\xB7","\xD2\xB8"=>"\xD2\xB9","\xD2\xBA"=>"\xD2\xBB",
"\xD2\xBC"=>"\xD2\xBD","\xD2\xBE"=>"\xD2\xBF","\xD3\x80"=>"\xD3\x8F","\xD3\x81"=>"\xD3\x82","\xD3\x83"=>"\xD3\x84",
"\xD3\x85"=>"\xD3\x86","\xD3\x87"=>"\xD3\x88","\xD3\x89"=>"\xD3\x8A","\xD3\x8B"=>"\xD3\x8C","\xD3\x8D"=>"\xD3\x8E",
"\xD3\x90"=>"\xD3\x91","\xD3\x92"=>"\xD3\x93","\xD3\x94"=>"\xD3\x95","\xD3\x96"=>"\xD3\x97","\xD3\x98"=>"\xD3\x99",
"\xD3\x9A"=>"\xD3\x9B","\xD3\x9C"=>"\xD3\x9D","\xD3\x9E"=>"\xD3\x9F","\xD3\xA0"=>"\xD3\xA1","\xD3\xA2"=>"\xD3\xA3",
"\xD3\xA4"=>"\xD3\xA5","\xD3\xA6"=>"\xD3\xA7","\xD3\xA8"=>"\xD3\xA9","\xD3\xAA"=>"\xD3\xAB","\xD3\xAC"=>"\xD3\xAD",
"\xD3\xAE"=>"\xD3\xAF","\xD3\xB0"=>"\xD3\xB1","\xD3\xB2"=>"\xD3\xB3","\xD3\xB4"=>"\xD3\xB5","\xD3\xB6"=>"\xD3\xB7",
"\xD3\xB8"=>"\xD3\xB9","\xD3\xBA"=>"\xD3\xBB","\xD3\xBC"=>"\xD3\xBD","\xD3\xBE"=>"\xD3\xBF","\xD4\x80"=>"\xD4\x81",
"\xD4\x82"=>"\xD4\x83","\xD4\x84"=>"\xD4\x85","\xD4\x86"=>"\xD4\x87","\xD4\x88"=>"\xD4\x89","\xD4\x8A"=>"\xD4\x8B",
"\xD4\x8C"=>"\xD4\x8D","\xD4\x8E"=>"\xD4\x8F","\xD4\x90"=>"\xD4\x91","\xD4\x92"=>"\xD4\x93","\xD4\x94"=>"\xD4\x95",
"\xD4\x96"=>"\xD4\x97","\xD4\x98"=>"\xD4\x99","\xD4\x9A"=>"\xD4\x9B","\xD4\x9C"=>"\xD4\x9D","\xD4\x9E"=>"\xD4\x9F",
"\xD4\xA0"=>"\xD4\xA1","\xD4\xA2"=>"\xD4\xA3","\xD4\xA4"=>"\xD4\xA5","\xD4\xA6"=>"\xD4\xA7","\xD4\xB1"=>"\xD5\xA1",
"\xD4\xB2"=>"\xD5\xA2","\xD4\xB3"=>"\xD5\xA3","\xD4\xB4"=>"\xD5\xA4","\xD4\xB5"=>"\xD5\xA5","\xD4\xB6"=>"\xD5\xA6",
"\xD4\xB7"=>"\xD5\xA7","\xD4\xB8"=>"\xD5\xA8","\xD4\xB9"=>"\xD5\xA9","\xD4\xBA"=>"\xD5\xAA","\xD4\xBB"=>"\xD5\xAB",
"\xD4\xBC"=>"\xD5\xAC","\xD4\xBD"=>"\xD5\xAD","\xD4\xBE"=>"\xD5\xAE","\xD4\xBF"=>"\xD5\xAF","\xD5\x80"=>"\xD5\xB0",
"\xD5\x81"=>"\xD5\xB1","\xD5\x82"=>"\xD5\xB2","\xD5\x83"=>"\xD5\xB3","\xD5\x84"=>"\xD5\xB4","\xD5\x85"=>"\xD5\xB5",
"\xD5\x86"=>"\xD5\xB6","\xD5\x87"=>"\xD5\xB7","\xD5\x88"=>"\xD5\xB8","\xD5\x89"=>"\xD5\xB9","\xD5\x8A"=>"\xD5\xBA",
"\xD5\x8B"=>"\xD5\xBB","\xD5\x8C"=>"\xD5\xBC","\xD5\x8D"=>"\xD5\xBD","\xD5\x8E"=>"\xD5\xBE","\xD5\x8F"=>"\xD5\xBF",
"\xD5\x90"=>"\xD6\x80","\xD5\x91"=>"\xD6\x81","\xD5\x92"=>"\xD6\x82","\xD5\x93"=>"\xD6\x83","\xD5\x94"=>"\xD6\x84",
"\xD5\x95"=>"\xD6\x85","\xD5\x96"=>"\xD6\x86","\xD6\x87"=>"\xD5\xA5\xD6\x82","\xE1\x82\xA0"=>"\xE2\xB4\x80",
"\xE1\x82\xA1"=>"\xE2\xB4\x81","\xE1\x82\xA2"=>"\xE2\xB4\x82","\xE1\x82\xA3"=>"\xE2\xB4\x83",
"\xE1\x82\xA4"=>"\xE2\xB4\x84","\xE1\x82\xA5"=>"\xE2\xB4\x85","\xE1\x82\xA6"=>"\xE2\xB4\x86",
"\xE1\x82\xA7"=>"\xE2\xB4\x87","\xE1\x82\xA8"=>"\xE2\xB4\x88","\xE1\x82\xA9"=>"\xE2\xB4\x89",
"\xE1\x82\xAA"=>"\xE2\xB4\x8A","\xE1\x82\xAB"=>"\xE2\xB4\x8B","\xE1\x82\xAC"=>"\xE2\xB4\x8C",
"\xE1\x82\xAD"=>"\xE2\xB4\x8D","\xE1\x82\xAE"=>"\xE2\xB4\x8E","\xE1\x82\xAF"=>"\xE2\xB4\x8F",
"\xE1\x82\xB0"=>"\xE2\xB4\x90","\xE1\x82\xB1"=>"\xE2\xB4\x91","\xE1\x82\xB2"=>"\xE2\xB4\x92",
"\xE1\x82\xB3"=>"\xE2\xB4\x93","\xE1\x82\xB4"=>"\xE2\xB4\x94","\xE1\x82\xB5"=>"\xE2\xB4\x95",
"\xE1\x82\xB6"=>"\xE2\xB4\x96","\xE1\x82\xB7"=>"\xE2\xB4\x97","\xE1\x82\xB8"=>"\xE2\xB4\x98",
"\xE1\x82\xB9"=>"\xE2\xB4\x99","\xE1\x82\xBA"=>"\xE2\xB4\x9A","\xE1\x82\xBB"=>"\xE2\xB4\x9B",
"\xE1\x82\xBC"=>"\xE2\xB4\x9C","\xE1\x82\xBD"=>"\xE2\xB4\x9D","\xE1\x82\xBE"=>"\xE2\xB4\x9E",
"\xE1\x82\xBF"=>"\xE2\xB4\x9F","\xE1\x83\x80"=>"\xE2\xB4\xA0","\xE1\x83\x81"=>"\xE2\xB4\xA1",
"\xE1\x83\x82"=>"\xE2\xB4\xA2","\xE1\x83\x83"=>"\xE2\xB4\xA3","\xE1\x83\x84"=>"\xE2\xB4\xA4",
"\xE1\x83\x85"=>"\xE2\xB4\xA5","\xE1\xB8\x80"=>"\xE1\xB8\x81","\xE1\xB8\x82"=>"\xE1\xB8\x83",
"\xE1\xB8\x84"=>"\xE1\xB8\x85","\xE1\xB8\x86"=>"\xE1\xB8\x87","\xE1\xB8\x88"=>"\xE1\xB8\x89",
"\xE1\xB8\x8A"=>"\xE1\xB8\x8B","\xE1\xB8\x8C"=>"\xE1\xB8\x8D","\xE1\xB8\x8E"=>"\xE1\xB8\x8F",
"\xE1\xB8\x90"=>"\xE1\xB8\x91","\xE1\xB8\x92"=>"\xE1\xB8\x93","\xE1\xB8\x94"=>"\xE1\xB8\x95",
"\xE1\xB8\x96"=>"\xE1\xB8\x97","\xE1\xB8\x98"=>"\xE1\xB8\x99","\xE1\xB8\x9A"=>"\xE1\xB8\x9B",
"\xE1\xB8\x9C"=>"\xE1\xB8\x9D","\xE1\xB8\x9E"=>"\xE1\xB8\x9F","\xE1\xB8\xA0"=>"\xE1\xB8\xA1",
"\xE1\xB8\xA2"=>"\xE1\xB8\xA3","\xE1\xB8\xA4"=>"\xE1\xB8\xA5","\xE1\xB8\xA6"=>"\xE1\xB8\xA7",
"\xE1\xB8\xA8"=>"\xE1\xB8\xA9","\xE1\xB8\xAA"=>"\xE1\xB8\xAB","\xE1\xB8\xAC"=>"\xE1\xB8\xAD",
"\xE1\xB8\xAE"=>"\xE1\xB8\xAF","\xE1\xB8\xB0"=>"\xE1\xB8\xB1","\xE1\xB8\xB2"=>"\xE1\xB8\xB3",
"\xE1\xB8\xB4"=>"\xE1\xB8\xB5","\xE1\xB8\xB6"=>"\xE1\xB8\xB7","\xE1\xB8\xB8"=>"\xE1\xB8\xB9",
"\xE1\xB8\xBA"=>"\xE1\xB8\xBB","\xE1\xB8\xBC"=>"\xE1\xB8\xBD","\xE1\xB8\xBE"=>"\xE1\xB8\xBF",
"\xE1\xB9\x80"=>"\xE1\xB9\x81","\xE1\xB9\x82"=>"\xE1\xB9\x83","\xE1\xB9\x84"=>"\xE1\xB9\x85",
"\xE1\xB9\x86"=>"\xE1\xB9\x87","\xE1\xB9\x88"=>"\xE1\xB9\x89","\xE1\xB9\x8A"=>"\xE1\xB9\x8B",
"\xE1\xB9\x8C"=>"\xE1\xB9\x8D","\xE1\xB9\x8E"=>"\xE1\xB9\x8F","\xE1\xB9\x90"=>"\xE1\xB9\x91",
"\xE1\xB9\x92"=>"\xE1\xB9\x93","\xE1\xB9\x94"=>"\xE1\xB9\x95","\xE1\xB9\x96"=>"\xE1\xB9\x97",
"\xE1\xB9\x98"=>"\xE1\xB9\x99","\xE1\xB9\x9A"=>"\xE1\xB9\x9B","\xE1\xB9\x9C"=>"\xE1\xB9\x9D",
"\xE1\xB9\x9E"=>"\xE1\xB9\x9F","\xE1\xB9\xA0"=>"\xE1\xB9\xA1","\xE1\xB9\xA2"=>"\xE1\xB9\xA3",
"\xE1\xB9\xA4"=>"\xE1\xB9\xA5","\xE1\xB9\xA6"=>"\xE1\xB9\xA7","\xE1\xB9\xA8"=>"\xE1\xB9\xA9",
"\xE1\xB9\xAA"=>"\xE1\xB9\xAB","\xE1\xB9\xAC"=>"\xE1\xB9\xAD","\xE1\xB9\xAE"=>"\xE1\xB9\xAF",
"\xE1\xB9\xB0"=>"\xE1\xB9\xB1","\xE1\xB9\xB2"=>"\xE1\xB9\xB3","\xE1\xB9\xB4"=>"\xE1\xB9\xB5",
"\xE1\xB9\xB6"=>"\xE1\xB9\xB7","\xE1\xB9\xB8"=>"\xE1\xB9\xB9","\xE1\xB9\xBA"=>"\xE1\xB9\xBB",
"\xE1\xB9\xBC"=>"\xE1\xB9\xBD","\xE1\xB9\xBE"=>"\xE1\xB9\xBF","\xE1\xBA\x80"=>"\xE1\xBA\x81",
"\xE1\xBA\x82"=>"\xE1\xBA\x83","\xE1\xBA\x84"=>"\xE1\xBA\x85","\xE1\xBA\x86"=>"\xE1\xBA\x87",
"\xE1\xBA\x88"=>"\xE1\xBA\x89","\xE1\xBA\x8A"=>"\xE1\xBA\x8B","\xE1\xBA\x8C"=>"\xE1\xBA\x8D",
"\xE1\xBA\x8E"=>"\xE1\xBA\x8F","\xE1\xBA\x90"=>"\xE1\xBA\x91","\xE1\xBA\x92"=>"\xE1\xBA\x93",
"\xE1\xBA\x94"=>"\xE1\xBA\x95","\xE1\xBA\x96"=>"h\xCC\xB1","\xE1\xBA\x97"=>"t\xCC\x88","\xE1\xBA\x98"=>"w\xCC\x8A",
"\xE1\xBA\x99"=>"y\xCC\x8A","\xE1\xBA\x9A"=>"a\xCA\xBE","\xE1\xBA\x9B"=>"\xE1\xB9\xA1","\xE1\xBA\x9E"=>'ss',
"\xE1\xBA\xA0"=>"\xE1\xBA\xA1","\xE1\xBA\xA2"=>"\xE1\xBA\xA3","\xE1\xBA\xA4"=>"\xE1\xBA\xA5",
"\xE1\xBA\xA6"=>"\xE1\xBA\xA7","\xE1\xBA\xA8"=>"\xE1\xBA\xA9","\xE1\xBA\xAA"=>"\xE1\xBA\xAB",
"\xE1\xBA\xAC"=>"\xE1\xBA\xAD","\xE1\xBA\xAE"=>"\xE1\xBA\xAF","\xE1\xBA\xB0"=>"\xE1\xBA\xB1",
"\xE1\xBA\xB2"=>"\xE1\xBA\xB3","\xE1\xBA\xB4"=>"\xE1\xBA\xB5","\xE1\xBA\xB6"=>"\xE1\xBA\xB7",
"\xE1\xBA\xB8"=>"\xE1\xBA\xB9","\xE1\xBA\xBA"=>"\xE1\xBA\xBB","\xE1\xBA\xBC"=>"\xE1\xBA\xBD",
"\xE1\xBA\xBE"=>"\xE1\xBA\xBF","\xE1\xBB\x80"=>"\xE1\xBB\x81","\xE1\xBB\x82"=>"\xE1\xBB\x83",
"\xE1\xBB\x84"=>"\xE1\xBB\x85","\xE1\xBB\x86"=>"\xE1\xBB\x87","\xE1\xBB\x88"=>"\xE1\xBB\x89",
"\xE1\xBB\x8A"=>"\xE1\xBB\x8B","\xE1\xBB\x8C"=>"\xE1\xBB\x8D","\xE1\xBB\x8E"=>"\xE1\xBB\x8F",
"\xE1\xBB\x90"=>"\xE1\xBB\x91","\xE1\xBB\x92"=>"\xE1\xBB\x93","\xE1\xBB\x94"=>"\xE1\xBB\x95",
"\xE1\xBB\x96"=>"\xE1\xBB\x97","\xE1\xBB\x98"=>"\xE1\xBB\x99","\xE1\xBB\x9A"=>"\xE1\xBB\x9B",
"\xE1\xBB\x9C"=>"\xE1\xBB\x9D","\xE1\xBB\x9E"=>"\xE1\xBB\x9F","\xE1\xBB\xA0"=>"\xE1\xBB\xA1",
"\xE1\xBB\xA2"=>"\xE1\xBB\xA3","\xE1\xBB\xA4"=>"\xE1\xBB\xA5","\xE1\xBB\xA6"=>"\xE1\xBB\xA7",
"\xE1\xBB\xA8"=>"\xE1\xBB\xA9","\xE1\xBB\xAA"=>"\xE1\xBB\xAB","\xE1\xBB\xAC"=>"\xE1\xBB\xAD",
"\xE1\xBB\xAE"=>"\xE1\xBB\xAF","\xE1\xBB\xB0"=>"\xE1\xBB\xB1","\xE1\xBB\xB2"=>"\xE1\xBB\xB3",
"\xE1\xBB\xB4"=>"\xE1\xBB\xB5","\xE1\xBB\xB6"=>"\xE1\xBB\xB7","\xE1\xBB\xB8"=>"\xE1\xBB\xB9",
"\xE1\xBB\xBA"=>"\xE1\xBB\xBB","\xE1\xBB\xBC"=>"\xE1\xBB\xBD","\xE1\xBB\xBE"=>"\xE1\xBB\xBF",
"\xE1\xBC\x88"=>"\xE1\xBC\x80","\xE1\xBC\x89"=>"\xE1\xBC\x81","\xE1\xBC\x8A"=>"\xE1\xBC\x82",
"\xE1\xBC\x8B"=>"\xE1\xBC\x83","\xE1\xBC\x8C"=>"\xE1\xBC\x84","\xE1\xBC\x8D"=>"\xE1\xBC\x85",
"\xE1\xBC\x8E"=>"\xE1\xBC\x86","\xE1\xBC\x8F"=>"\xE1\xBC\x87","\xE1\xBC\x98"=>"\xE1\xBC\x90",
"\xE1\xBC\x99"=>"\xE1\xBC\x91","\xE1\xBC\x9A"=>"\xE1\xBC\x92","\xE1\xBC\x9B"=>"\xE1\xBC\x93",
"\xE1\xBC\x9C"=>"\xE1\xBC\x94","\xE1\xBC\x9D"=>"\xE1\xBC\x95","\xE1\xBC\xA8"=>"\xE1\xBC\xA0",
"\xE1\xBC\xA9"=>"\xE1\xBC\xA1","\xE1\xBC\xAA"=>"\xE1\xBC\xA2","\xE1\xBC\xAB"=>"\xE1\xBC\xA3",
"\xE1\xBC\xAC"=>"\xE1\xBC\xA4","\xE1\xBC\xAD"=>"\xE1\xBC\xA5","\xE1\xBC\xAE"=>"\xE1\xBC\xA6",
"\xE1\xBC\xAF"=>"\xE1\xBC\xA7","\xE1\xBC\xB8"=>"\xE1\xBC\xB0","\xE1\xBC\xB9"=>"\xE1\xBC\xB1",
"\xE1\xBC\xBA"=>"\xE1\xBC\xB2","\xE1\xBC\xBB"=>"\xE1\xBC\xB3","\xE1\xBC\xBC"=>"\xE1\xBC\xB4",
"\xE1\xBC\xBD"=>"\xE1\xBC\xB5","\xE1\xBC\xBE"=>"\xE1\xBC\xB6","\xE1\xBC\xBF"=>"\xE1\xBC\xB7",
"\xE1\xBD\x88"=>"\xE1\xBD\x80","\xE1\xBD\x89"=>"\xE1\xBD\x81","\xE1\xBD\x8A"=>"\xE1\xBD\x82",
"\xE1\xBD\x8B"=>"\xE1\xBD\x83","\xE1\xBD\x8C"=>"\xE1\xBD\x84","\xE1\xBD\x8D"=>"\xE1\xBD\x85",
"\xE1\xBD\x90"=>"\xCF\x85\xCC\x93","\xE1\xBD\x92"=>"\xCF\x85\xCC\x93\xCC\x80",
"\xE1\xBD\x94"=>"\xCF\x85\xCC\x93\xCC\x81","\xE1\xBD\x96"=>"\xCF\x85\xCC\x93\xCD\x82","\xE1\xBD\x99"=>"\xE1\xBD\x91",
"\xE1\xBD\x9B"=>"\xE1\xBD\x93","\xE1\xBD\x9D"=>"\xE1\xBD\x95","\xE1\xBD\x9F"=>"\xE1\xBD\x97",
"\xE1\xBD\xA8"=>"\xE1\xBD\xA0","\xE1\xBD\xA9"=>"\xE1\xBD\xA1","\xE1\xBD\xAA"=>"\xE1\xBD\xA2",
"\xE1\xBD\xAB"=>"\xE1\xBD\xA3","\xE1\xBD\xAC"=>"\xE1\xBD\xA4","\xE1\xBD\xAD"=>"\xE1\xBD\xA5",
"\xE1\xBD\xAE"=>"\xE1\xBD\xA6","\xE1\xBD\xAF"=>"\xE1\xBD\xA7","\xE1\xBE\x80"=>"\xE1\xBC\x80\xCE\xB9",
"\xE1\xBE\x81"=>"\xE1\xBC\x81\xCE\xB9","\xE1\xBE\x82"=>"\xE1\xBC\x82\xCE\xB9","\xE1\xBE\x83"=>"\xE1\xBC\x83\xCE\xB9",
"\xE1\xBE\x84"=>"\xE1\xBC\x84\xCE\xB9","\xE1\xBE\x85"=>"\xE1\xBC\x85\xCE\xB9","\xE1\xBE\x86"=>"\xE1\xBC\x86\xCE\xB9",
"\xE1\xBE\x87"=>"\xE1\xBC\x87\xCE\xB9","\xE1\xBE\x88"=>"\xE1\xBC\x80\xCE\xB9","\xE1\xBE\x89"=>"\xE1\xBC\x81\xCE\xB9",
"\xE1\xBE\x8A"=>"\xE1\xBC\x82\xCE\xB9","\xE1\xBE\x8B"=>"\xE1\xBC\x83\xCE\xB9","\xE1\xBE\x8C"=>"\xE1\xBC\x84\xCE\xB9",
"\xE1\xBE\x8D"=>"\xE1\xBC\x85\xCE\xB9","\xE1\xBE\x8E"=>"\xE1\xBC\x86\xCE\xB9","\xE1\xBE\x8F"=>"\xE1\xBC\x87\xCE\xB9",
"\xE1\xBE\x90"=>"\xE1\xBC\xA0\xCE\xB9","\xE1\xBE\x91"=>"\xE1\xBC\xA1\xCE\xB9","\xE1\xBE\x92"=>"\xE1\xBC\xA2\xCE\xB9",
"\xE1\xBE\x93"=>"\xE1\xBC\xA3\xCE\xB9","\xE1\xBE\x94"=>"\xE1\xBC\xA4\xCE\xB9","\xE1\xBE\x95"=>"\xE1\xBC\xA5\xCE\xB9",
"\xE1\xBE\x96"=>"\xE1\xBC\xA6\xCE\xB9","\xE1\xBE\x97"=>"\xE1\xBC\xA7\xCE\xB9","\xE1\xBE\x98"=>"\xE1\xBC\xA0\xCE\xB9",
"\xE1\xBE\x99"=>"\xE1\xBC\xA1\xCE\xB9","\xE1\xBE\x9A"=>"\xE1\xBC\xA2\xCE\xB9","\xE1\xBE\x9B"=>"\xE1\xBC\xA3\xCE\xB9",
"\xE1\xBE\x9C"=>"\xE1\xBC\xA4\xCE\xB9","\xE1\xBE\x9D"=>"\xE1\xBC\xA5\xCE\xB9","\xE1\xBE\x9E"=>"\xE1\xBC\xA6\xCE\xB9",
"\xE1\xBE\x9F"=>"\xE1\xBC\xA7\xCE\xB9","\xE1\xBE\xA0"=>"\xE1\xBD\xA0\xCE\xB9","\xE1\xBE\xA1"=>"\xE1\xBD\xA1\xCE\xB9",
"\xE1\xBE\xA2"=>"\xE1\xBD\xA2\xCE\xB9","\xE1\xBE\xA3"=>"\xE1\xBD\xA3\xCE\xB9","\xE1\xBE\xA4"=>"\xE1\xBD\xA4\xCE\xB9",
"\xE1\xBE\xA5"=>"\xE1\xBD\xA5\xCE\xB9","\xE1\xBE\xA6"=>"\xE1\xBD\xA6\xCE\xB9","\xE1\xBE\xA7"=>"\xE1\xBD\xA7\xCE\xB9",
"\xE1\xBE\xA8"=>"\xE1\xBD\xA0\xCE\xB9","\xE1\xBE\xA9"=>"\xE1\xBD\xA1\xCE\xB9","\xE1\xBE\xAA"=>"\xE1\xBD\xA2\xCE\xB9",
"\xE1\xBE\xAB"=>"\xE1\xBD\xA3\xCE\xB9","\xE1\xBE\xAC"=>"\xE1\xBD\xA4\xCE\xB9","\xE1\xBE\xAD"=>"\xE1\xBD\xA5\xCE\xB9",
"\xE1\xBE\xAE"=>"\xE1\xBD\xA6\xCE\xB9","\xE1\xBE\xAF"=>"\xE1\xBD\xA7\xCE\xB9","\xE1\xBE\xB2"=>"\xE1\xBD\xB0\xCE\xB9",
"\xE1\xBE\xB3"=>"\xCE\xB1\xCE\xB9","\xE1\xBE\xB4"=>"\xCE\xAC\xCE\xB9","\xE1\xBE\xB6"=>"\xCE\xB1\xCD\x82",
"\xE1\xBE\xB7"=>"\xCE\xB1\xCD\x82\xCE\xB9","\xE1\xBE\xB8"=>"\xE1\xBE\xB0","\xE1\xBE\xB9"=>"\xE1\xBE\xB1",
"\xE1\xBE\xBA"=>"\xE1\xBD\xB0","\xE1\xBE\xBB"=>"\xE1\xBD\xB1","\xE1\xBE\xBC"=>"\xCE\xB1\xCE\xB9",
"\xE1\xBE\xBE"=>"\xCE\xB9","\xE1\xBF\x82"=>"\xE1\xBD\xB4\xCE\xB9","\xE1\xBF\x83"=>"\xCE\xB7\xCE\xB9",
"\xE1\xBF\x84"=>"\xCE\xAE\xCE\xB9","\xE1\xBF\x86"=>"\xCE\xB7\xCD\x82","\xE1\xBF\x87"=>"\xCE\xB7\xCD\x82\xCE\xB9",
"\xE1\xBF\x88"=>"\xE1\xBD\xB2","\xE1\xBF\x89"=>"\xE1\xBD\xB3","\xE1\xBF\x8A"=>"\xE1\xBD\xB4",
"\xE1\xBF\x8B"=>"\xE1\xBD\xB5","\xE1\xBF\x8C"=>"\xCE\xB7\xCE\xB9","\xE1\xBF\x92"=>"\xCE\xB9\xCC\x88\xCC\x80",
"\xE1\xBF\x93"=>"\xCE\xB9\xCC\x88\xCC\x81","\xE1\xBF\x96"=>"\xCE\xB9\xCD\x82",
"\xE1\xBF\x97"=>"\xCE\xB9\xCC\x88\xCD\x82","\xE1\xBF\x98"=>"\xE1\xBF\x90","\xE1\xBF\x99"=>"\xE1\xBF\x91",
"\xE1\xBF\x9A"=>"\xE1\xBD\xB6","\xE1\xBF\x9B"=>"\xE1\xBD\xB7","\xE1\xBF\xA2"=>"\xCF\x85\xCC\x88\xCC\x80",
"\xE1\xBF\xA3"=>"\xCF\x85\xCC\x88\xCC\x81","\xE1\xBF\xA4"=>"\xCF\x81\xCC\x93","\xE1\xBF\xA6"=>"\xCF\x85\xCD\x82",
"\xE1\xBF\xA7"=>"\xCF\x85\xCC\x88\xCD\x82","\xE1\xBF\xA8"=>"\xE1\xBF\xA0","\xE1\xBF\xA9"=>"\xE1\xBF\xA1",
"\xE1\xBF\xAA"=>"\xE1\xBD\xBA","\xE1\xBF\xAB"=>"\xE1\xBD\xBB","\xE1\xBF\xAC"=>"\xE1\xBF\xA5",
"\xE1\xBF\xB2"=>"\xE1\xBD\xBC\xCE\xB9","\xE1\xBF\xB3"=>"\xCF\x89\xCE\xB9","\xE1\xBF\xB4"=>"\xCF\x8E\xCE\xB9",
"\xE1\xBF\xB6"=>"\xCF\x89\xCD\x82","\xE1\xBF\xB7"=>"\xCF\x89\xCD\x82\xCE\xB9","\xE1\xBF\xB8"=>"\xE1\xBD\xB8",
"\xE1\xBF\xB9"=>"\xE1\xBD\xB9","\xE1\xBF\xBA"=>"\xE1\xBD\xBC","\xE1\xBF\xBB"=>"\xE1\xBD\xBD",
"\xE1\xBF\xBC"=>"\xCF\x89\xCE\xB9","\xE2\x84\xA6"=>"\xCF\x89","\xE2\x84\xAA"=>'k',"\xE2\x84\xAB"=>"\xC3\xA5",
"\xE2\x84\xB2"=>"\xE2\x85\x8E","\xE2\x85\xA0"=>"\xE2\x85\xB0","\xE2\x85\xA1"=>"\xE2\x85\xB1",
"\xE2\x85\xA2"=>"\xE2\x85\xB2","\xE2\x85\xA3"=>"\xE2\x85\xB3","\xE2\x85\xA4"=>"\xE2\x85\xB4",
"\xE2\x85\xA5"=>"\xE2\x85\xB5","\xE2\x85\xA6"=>"\xE2\x85\xB6","\xE2\x85\xA7"=>"\xE2\x85\xB7",
"\xE2\x85\xA8"=>"\xE2\x85\xB8","\xE2\x85\xA9"=>"\xE2\x85\xB9","\xE2\x85\xAA"=>"\xE2\x85\xBA",
"\xE2\x85\xAB"=>"\xE2\x85\xBB","\xE2\x85\xAC"=>"\xE2\x85\xBC","\xE2\x85\xAD"=>"\xE2\x85\xBD",
"\xE2\x85\xAE"=>"\xE2\x85\xBE","\xE2\x85\xAF"=>"\xE2\x85\xBF","\xE2\x86\x83"=>"\xE2\x86\x84",
"\xE2\x92\xB6"=>"\xE2\x93\x90","\xE2\x92\xB7"=>"\xE2\x93\x91","\xE2\x92\xB8"=>"\xE2\x93\x92",
"\xE2\x92\xB9"=>"\xE2\x93\x93","\xE2\x92\xBA"=>"\xE2\x93\x94","\xE2\x92\xBB"=>"\xE2\x93\x95",
"\xE2\x92\xBC"=>"\xE2\x93\x96","\xE2\x92\xBD"=>"\xE2\x93\x97","\xE2\x92\xBE"=>"\xE2\x93\x98",
"\xE2\x92\xBF"=>"\xE2\x93\x99","\xE2\x93\x80"=>"\xE2\x93\x9A","\xE2\x93\x81"=>"\xE2\x93\x9B",
"\xE2\x93\x82"=>"\xE2\x93\x9C","\xE2\x93\x83"=>"\xE2\x93\x9D","\xE2\x93\x84"=>"\xE2\x93\x9E",
"\xE2\x93\x85"=>"\xE2\x93\x9F","\xE2\x93\x86"=>"\xE2\x93\xA0","\xE2\x93\x87"=>"\xE2\x93\xA1",
"\xE2\x93\x88"=>"\xE2\x93\xA2","\xE2\x93\x89"=>"\xE2\x93\xA3","\xE2\x93\x8A"=>"\xE2\x93\xA4",
"\xE2\x93\x8B"=>"\xE2\x93\xA5","\xE2\x93\x8C"=>"\xE2\x93\xA6","\xE2\x93\x8D"=>"\xE2\x93\xA7",
"\xE2\x93\x8E"=>"\xE2\x93\xA8","\xE2\x93\x8F"=>"\xE2\x93\xA9","\xE2\xB0\x80"=>"\xE2\xB0\xB0",
"\xE2\xB0\x81"=>"\xE2\xB0\xB1","\xE2\xB0\x82"=>"\xE2\xB0\xB2","\xE2\xB0\x83"=>"\xE2\xB0\xB3",
"\xE2\xB0\x84"=>"\xE2\xB0\xB4","\xE2\xB0\x85"=>"\xE2\xB0\xB5","\xE2\xB0\x86"=>"\xE2\xB0\xB6",
"\xE2\xB0\x87"=>"\xE2\xB0\xB7","\xE2\xB0\x88"=>"\xE2\xB0\xB8","\xE2\xB0\x89"=>"\xE2\xB0\xB9",
"\xE2\xB0\x8A"=>"\xE2\xB0\xBA","\xE2\xB0\x8B"=>"\xE2\xB0\xBB","\xE2\xB0\x8C"=>"\xE2\xB0\xBC",
"\xE2\xB0\x8D"=>"\xE2\xB0\xBD","\xE2\xB0\x8E"=>"\xE2\xB0\xBE","\xE2\xB0\x8F"=>"\xE2\xB0\xBF",
"\xE2\xB0\x90"=>"\xE2\xB1\x80","\xE2\xB0\x91"=>"\xE2\xB1\x81","\xE2\xB0\x92"=>"\xE2\xB1\x82",
"\xE2\xB0\x93"=>"\xE2\xB1\x83","\xE2\xB0\x94"=>"\xE2\xB1\x84","\xE2\xB0\x95"=>"\xE2\xB1\x85",
"\xE2\xB0\x96"=>"\xE2\xB1\x86","\xE2\xB0\x97"=>"\xE2\xB1\x87","\xE2\xB0\x98"=>"\xE2\xB1\x88",
"\xE2\xB0\x99"=>"\xE2\xB1\x89","\xE2\xB0\x9A"=>"\xE2\xB1\x8A","\xE2\xB0\x9B"=>"\xE2\xB1\x8B",
"\xE2\xB0\x9C"=>"\xE2\xB1\x8C","\xE2\xB0\x9D"=>"\xE2\xB1\x8D","\xE2\xB0\x9E"=>"\xE2\xB1\x8E",
"\xE2\xB0\x9F"=>"\xE2\xB1\x8F","\xE2\xB0\xA0"=>"\xE2\xB1\x90","\xE2\xB0\xA1"=>"\xE2\xB1\x91",
"\xE2\xB0\xA2"=>"\xE2\xB1\x92","\xE2\xB0\xA3"=>"\xE2\xB1\x93","\xE2\xB0\xA4"=>"\xE2\xB1\x94",
"\xE2\xB0\xA5"=>"\xE2\xB1\x95","\xE2\xB0\xA6"=>"\xE2\xB1\x96","\xE2\xB0\xA7"=>"\xE2\xB1\x97",
"\xE2\xB0\xA8"=>"\xE2\xB1\x98","\xE2\xB0\xA9"=>"\xE2\xB1\x99","\xE2\xB0\xAA"=>"\xE2\xB1\x9A",
"\xE2\xB0\xAB"=>"\xE2\xB1\x9B","\xE2\xB0\xAC"=>"\xE2\xB1\x9C","\xE2\xB0\xAD"=>"\xE2\xB1\x9D",
"\xE2\xB0\xAE"=>"\xE2\xB1\x9E","\xE2\xB1\xA0"=>"\xE2\xB1\xA1","\xE2\xB1\xA2"=>"\xC9\xAB",
"\xE2\xB1\xA3"=>"\xE1\xB5\xBD","\xE2\xB1\xA4"=>"\xC9\xBD","\xE2\xB1\xA7"=>"\xE2\xB1\xA8",
"\xE2\xB1\xA9"=>"\xE2\xB1\xAA","\xE2\xB1\xAB"=>"\xE2\xB1\xAC","\xE2\xB1\xAD"=>"\xC9\x91","\xE2\xB1\xAE"=>"\xC9\xB1",
"\xE2\xB1\xAF"=>"\xC9\x90","\xE2\xB1\xB0"=>"\xC9\x92","\xE2\xB1\xB2"=>"\xE2\xB1\xB3","\xE2\xB1\xB5"=>"\xE2\xB1\xB6",
"\xE2\xB1\xBE"=>"\xC8\xBF","\xE2\xB1\xBF"=>"\xC9\x80","\xE2\xB2\x80"=>"\xE2\xB2\x81","\xE2\xB2\x82"=>"\xE2\xB2\x83",
"\xE2\xB2\x84"=>"\xE2\xB2\x85","\xE2\xB2\x86"=>"\xE2\xB2\x87","\xE2\xB2\x88"=>"\xE2\xB2\x89",
"\xE2\xB2\x8A"=>"\xE2\xB2\x8B","\xE2\xB2\x8C"=>"\xE2\xB2\x8D","\xE2\xB2\x8E"=>"\xE2\xB2\x8F",
"\xE2\xB2\x90"=>"\xE2\xB2\x91","\xE2\xB2\x92"=>"\xE2\xB2\x93","\xE2\xB2\x94"=>"\xE2\xB2\x95",
"\xE2\xB2\x96"=>"\xE2\xB2\x97","\xE2\xB2\x98"=>"\xE2\xB2\x99","\xE2\xB2\x9A"=>"\xE2\xB2\x9B",
"\xE2\xB2\x9C"=>"\xE2\xB2\x9D","\xE2\xB2\x9E"=>"\xE2\xB2\x9F","\xE2\xB2\xA0"=>"\xE2\xB2\xA1",
"\xE2\xB2\xA2"=>"\xE2\xB2\xA3","\xE2\xB2\xA4"=>"\xE2\xB2\xA5","\xE2\xB2\xA6"=>"\xE2\xB2\xA7",
"\xE2\xB2\xA8"=>"\xE2\xB2\xA9","\xE2\xB2\xAA"=>"\xE2\xB2\xAB","\xE2\xB2\xAC"=>"\xE2\xB2\xAD",
"\xE2\xB2\xAE"=>"\xE2\xB2\xAF","\xE2\xB2\xB0"=>"\xE2\xB2\xB1","\xE2\xB2\xB2"=>"\xE2\xB2\xB3",
"\xE2\xB2\xB4"=>"\xE2\xB2\xB5","\xE2\xB2\xB6"=>"\xE2\xB2\xB7","\xE2\xB2\xB8"=>"\xE2\xB2\xB9",
"\xE2\xB2\xBA"=>"\xE2\xB2\xBB","\xE2\xB2\xBC"=>"\xE2\xB2\xBD","\xE2\xB2\xBE"=>"\xE2\xB2\xBF",
"\xE2\xB3\x80"=>"\xE2\xB3\x81","\xE2\xB3\x82"=>"\xE2\xB3\x83","\xE2\xB3\x84"=>"\xE2\xB3\x85",
"\xE2\xB3\x86"=>"\xE2\xB3\x87","\xE2\xB3\x88"=>"\xE2\xB3\x89","\xE2\xB3\x8A"=>"\xE2\xB3\x8B",
"\xE2\xB3\x8C"=>"\xE2\xB3\x8D","\xE2\xB3\x8E"=>"\xE2\xB3\x8F","\xE2\xB3\x90"=>"\xE2\xB3\x91",
"\xE2\xB3\x92"=>"\xE2\xB3\x93","\xE2\xB3\x94"=>"\xE2\xB3\x95","\xE2\xB3\x96"=>"\xE2\xB3\x97",
"\xE2\xB3\x98"=>"\xE2\xB3\x99","\xE2\xB3\x9A"=>"\xE2\xB3\x9B","\xE2\xB3\x9C"=>"\xE2\xB3\x9D",
"\xE2\xB3\x9E"=>"\xE2\xB3\x9F","\xE2\xB3\xA0"=>"\xE2\xB3\xA1","\xE2\xB3\xA2"=>"\xE2\xB3\xA3",
"\xE2\xB3\xAB"=>"\xE2\xB3\xAC","\xE2\xB3\xAD"=>"\xE2\xB3\xAE","\xEA\x99\x80"=>"\xEA\x99\x81",
"\xEA\x99\x82"=>"\xEA\x99\x83","\xEA\x99\x84"=>"\xEA\x99\x85","\xEA\x99\x86"=>"\xEA\x99\x87",
"\xEA\x99\x88"=>"\xEA\x99\x89","\xEA\x99\x8A"=>"\xEA\x99\x8B","\xEA\x99\x8C"=>"\xEA\x99\x8D",
"\xEA\x99\x8E"=>"\xEA\x99\x8F","\xEA\x99\x90"=>"\xEA\x99\x91","\xEA\x99\x92"=>"\xEA\x99\x93",
"\xEA\x99\x94"=>"\xEA\x99\x95","\xEA\x99\x96"=>"\xEA\x99\x97","\xEA\x99\x98"=>"\xEA\x99\x99",
"\xEA\x99\x9A"=>"\xEA\x99\x9B","\xEA\x99\x9C"=>"\xEA\x99\x9D","\xEA\x99\x9E"=>"\xEA\x99\x9F",
"\xEA\x99\xA0"=>"\xEA\x99\xA1","\xEA\x99\xA2"=>"\xEA\x99\xA3","\xEA\x99\xA4"=>"\xEA\x99\xA5",
"\xEA\x99\xA6"=>"\xEA\x99\xA7","\xEA\x99\xA8"=>"\xEA\x99\xA9","\xEA\x99\xAA"=>"\xEA\x99\xAB",
"\xEA\x99\xAC"=>"\xEA\x99\xAD","\xEA\x9A\x80"=>"\xEA\x9A\x81","\xEA\x9A\x82"=>"\xEA\x9A\x83",
"\xEA\x9A\x84"=>"\xEA\x9A\x85","\xEA\x9A\x86"=>"\xEA\x9A\x87","\xEA\x9A\x88"=>"\xEA\x9A\x89",
"\xEA\x9A\x8A"=>"\xEA\x9A\x8B","\xEA\x9A\x8C"=>"\xEA\x9A\x8D","\xEA\x9A\x8E"=>"\xEA\x9A\x8F",
"\xEA\x9A\x90"=>"\xEA\x9A\x91","\xEA\x9A\x92"=>"\xEA\x9A\x93","\xEA\x9A\x94"=>"\xEA\x9A\x95",
"\xEA\x9A\x96"=>"\xEA\x9A\x97","\xEA\x9C\xA2"=>"\xEA\x9C\xA3","\xEA\x9C\xA4"=>"\xEA\x9C\xA5",
"\xEA\x9C\xA6"=>"\xEA\x9C\xA7","\xEA\x9C\xA8"=>"\xEA\x9C\xA9","\xEA\x9C\xAA"=>"\xEA\x9C\xAB",
"\xEA\x9C\xAC"=>"\xEA\x9C\xAD","\xEA\x9C\xAE"=>"\xEA\x9C\xAF","\xEA\x9C\xB2"=>"\xEA\x9C\xB3",
"\xEA\x9C\xB4"=>"\xEA\x9C\xB5","\xEA\x9C\xB6"=>"\xEA\x9C\xB7","\xEA\x9C\xB8"=>"\xEA\x9C\xB9",
"\xEA\x9C\xBA"=>"\xEA\x9C\xBB","\xEA\x9C\xBC"=>"\xEA\x9C\xBD","\xEA\x9C\xBE"=>"\xEA\x9C\xBF",
"\xEA\x9D\x80"=>"\xEA\x9D\x81","\xEA\x9D\x82"=>"\xEA\x9D\x83","\xEA\x9D\x84"=>"\xEA\x9D\x85",
"\xEA\x9D\x86"=>"\xEA\x9D\x87","\xEA\x9D\x88"=>"\xEA\x9D\x89","\xEA\x9D\x8A"=>"\xEA\x9D\x8B",
"\xEA\x9D\x8C"=>"\xEA\x9D\x8D","\xEA\x9D\x8E"=>"\xEA\x9D\x8F","\xEA\x9D\x90"=>"\xEA\x9D\x91",
"\xEA\x9D\x92"=>"\xEA\x9D\x93","\xEA\x9D\x94"=>"\xEA\x9D\x95","\xEA\x9D\x96"=>"\xEA\x9D\x97",
"\xEA\x9D\x98"=>"\xEA\x9D\x99","\xEA\x9D\x9A"=>"\xEA\x9D\x9B","\xEA\x9D\x9C"=>"\xEA\x9D\x9D",
"\xEA\x9D\x9E"=>"\xEA\x9D\x9F","\xEA\x9D\xA0"=>"\xEA\x9D\xA1","\xEA\x9D\xA2"=>"\xEA\x9D\xA3",
"\xEA\x9D\xA4"=>"\xEA\x9D\xA5","\xEA\x9D\xA6"=>"\xEA\x9D\xA7","\xEA\x9D\xA8"=>"\xEA\x9D\xA9",
"\xEA\x9D\xAA"=>"\xEA\x9D\xAB","\xEA\x9D\xAC"=>"\xEA\x9D\xAD","\xEA\x9D\xAE"=>"\xEA\x9D\xAF",
"\xEA\x9D\xB9"=>"\xEA\x9D\xBA","\xEA\x9D\xBB"=>"\xEA\x9D\xBC","\xEA\x9D\xBD"=>"\xE1\xB5\xB9",
"\xEA\x9D\xBE"=>"\xEA\x9D\xBF","\xEA\x9E\x80"=>"\xEA\x9E\x81","\xEA\x9E\x82"=>"\xEA\x9E\x83",
"\xEA\x9E\x84"=>"\xEA\x9E\x85","\xEA\x9E\x86"=>"\xEA\x9E\x87","\xEA\x9E\x8B"=>"\xEA\x9E\x8C",
"\xEA\x9E\x8D"=>"\xC9\xA5","\xEA\x9E\x90"=>"\xEA\x9E\x91","\xEA\x9E\xA0"=>"\xEA\x9E\xA1",
"\xEA\x9E\xA2"=>"\xEA\x9E\xA3","\xEA\x9E\xA4"=>"\xEA\x9E\xA5","\xEA\x9E\xA6"=>"\xEA\x9E\xA7",
"\xEA\x9E\xA8"=>"\xEA\x9E\xA9","\xEF\xAC\x80"=>'ff',"\xEF\xAC\x81"=>'fi',"\xEF\xAC\x82"=>'fl',"\xEF\xAC\x83"=>'ffi',
"\xEF\xAC\x84"=>'ffl',"\xEF\xAC\x85"=>'st',"\xEF\xAC\x86"=>'st',"\xEF\xAC\x93"=>"\xD5\xB4\xD5\xB6",
"\xEF\xAC\x94"=>"\xD5\xB4\xD5\xA5","\xEF\xAC\x95"=>"\xD5\xB4\xD5\xAB","\xEF\xAC\x96"=>"\xD5\xBE\xD5\xB6",
"\xEF\xAC\x97"=>"\xD5\xB4\xD5\xAD","\xEF\xBC\xA1"=>"\xEF\xBD\x81","\xEF\xBC\xA2"=>"\xEF\xBD\x82",
"\xEF\xBC\xA3"=>"\xEF\xBD\x83","\xEF\xBC\xA4"=>"\xEF\xBD\x84","\xEF\xBC\xA5"=>"\xEF\xBD\x85",
"\xEF\xBC\xA6"=>"\xEF\xBD\x86","\xEF\xBC\xA7"=>"\xEF\xBD\x87","\xEF\xBC\xA8"=>"\xEF\xBD\x88",
"\xEF\xBC\xA9"=>"\xEF\xBD\x89","\xEF\xBC\xAA"=>"\xEF\xBD\x8A","\xEF\xBC\xAB"=>"\xEF\xBD\x8B",
"\xEF\xBC\xAC"=>"\xEF\xBD\x8C","\xEF\xBC\xAD"=>"\xEF\xBD\x8D","\xEF\xBC\xAE"=>"\xEF\xBD\x8E",
"\xEF\xBC\xAF"=>"\xEF\xBD\x8F","\xEF\xBC\xB0"=>"\xEF\xBD\x90","\xEF\xBC\xB1"=>"\xEF\xBD\x91",
"\xEF\xBC\xB2"=>"\xEF\xBD\x92","\xEF\xBC\xB3"=>"\xEF\xBD\x93","\xEF\xBC\xB4"=>"\xEF\xBD\x94",
"\xEF\xBC\xB5"=>"\xEF\xBD\x95","\xEF\xBC\xB6"=>"\xEF\xBD\x96","\xEF\xBC\xB7"=>"\xEF\xBD\x97",
"\xEF\xBC\xB8"=>"\xEF\xBD\x98","\xEF\xBC\xB9"=>"\xEF\xBD\x99","\xEF\xBC\xBA"=>"\xEF\xBD\x9A",
"\xF0\x90\x90\x80"=>"\xF0\x90\x90\xA8","\xF0\x90\x90\x81"=>"\xF0\x90\x90\xA9","\xF0\x90\x90\x82"=>"\xF0\x90\x90\xAA",
"\xF0\x90\x90\x83"=>"\xF0\x90\x90\xAB","\xF0\x90\x90\x84"=>"\xF0\x90\x90\xAC","\xF0\x90\x90\x85"=>"\xF0\x90\x90\xAD",
"\xF0\x90\x90\x86"=>"\xF0\x90\x90\xAE","\xF0\x90\x90\x87"=>"\xF0\x90\x90\xAF","\xF0\x90\x90\x88"=>"\xF0\x90\x90\xB0",
"\xF0\x90\x90\x89"=>"\xF0\x90\x90\xB1","\xF0\x90\x90\x8A"=>"\xF0\x90\x90\xB2","\xF0\x90\x90\x8B"=>"\xF0\x90\x90\xB3",
"\xF0\x90\x90\x8C"=>"\xF0\x90\x90\xB4","\xF0\x90\x90\x8D"=>"\xF0\x90\x90\xB5","\xF0\x90\x90\x8E"=>"\xF0\x90\x90\xB6",
"\xF0\x90\x90\x8F"=>"\xF0\x90\x90\xB7","\xF0\x90\x90\x90"=>"\xF0\x90\x90\xB8","\xF0\x90\x90\x91"=>"\xF0\x90\x90\xB9",
"\xF0\x90\x90\x92"=>"\xF0\x90\x90\xBA","\xF0\x90\x90\x93"=>"\xF0\x90\x90\xBB","\xF0\x90\x90\x94"=>"\xF0\x90\x90\xBC",
"\xF0\x90\x90\x95"=>"\xF0\x90\x90\xBD","\xF0\x90\x90\x96"=>"\xF0\x90\x90\xBE","\xF0\x90\x90\x97"=>"\xF0\x90\x90\xBF",
"\xF0\x90\x90\x98"=>"\xF0\x90\x91\x80","\xF0\x90\x90\x99"=>"\xF0\x90\x91\x81","\xF0\x90\x90\x9A"=>"\xF0\x90\x91\x82",
"\xF0\x90\x90\x9B"=>"\xF0\x90\x91\x83","\xF0\x90\x90\x9C"=>"\xF0\x90\x91\x84","\xF0\x90\x90\x9D"=>"\xF0\x90\x91\x85",
"\xF0\x90\x90\x9E"=>"\xF0\x90\x91\x86","\xF0\x90\x90\x9F"=>"\xF0\x90\x91\x87","\xF0\x90\x90\xA0"=>"\xF0\x90\x91\x88",
"\xF0\x90\x90\xA1"=>"\xF0\x90\x91\x89","\xF0\x90\x90\xA2"=>"\xF0\x90\x91\x8A","\xF0\x90\x90\xA3"=>"\xF0\x90\x91\x8B",
"\xF0\x90\x90\xA4"=>"\xF0\x90\x91\x8C","\xF0\x90\x90\xA5"=>"\xF0\x90\x91\x8D","\xF0\x90\x90\xA6"=>"\xF0\x90\x91\x8E",
"\xF0\x90\x90\xA7"=>"\xF0\x90\x91\x8F");
    }

    /** UTF-8 mapping from lowercase to uppercase (sort of)
     *
     * This array derived from CaseFolding-6.0.0.txt
     * (part of the Unicode Character Database) but with a twist.
     * It maps UTF-8 encoded lower case codepoints to the
     * first corresponding upper case codepoint using status
     * 'C' (common case folding) and 'S' (simple case folding).
     * The special case for Turkic is not considered.
     *
     * Basically this is the reverse of the mapping upper=>lower.
     * However, there are some Unicode characters that are neither
     * lowercase nor uppercase, e.g. U+01F2 "Dz" which is titlecase.
     * Because the input was an injective table to lowercase (ie.
     * both "DZ" and "Dz" map to "dz") we cannot completely trust the
     * reversed mapping. The codepoints not stored in the table below are:
     * <code>
     * 017F: 's' : "\xC5\xBF" or 'S' # LATIN SMALL LETTER LONG S
     * 01C5: "\xC7\x86" : "\xC7\x85" or "\xC7\x84" # LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON
     * 01C8: "\xC7\x89" : "\xC7\x88" or "\xC7\x87" # LATIN CAPITAL LETTER L WITH SMALL LETTER J
     * 01CB: "\xC7\x8C" : "\xC7\x8B" or "\xC7\x8A" # LATIN CAPITAL LETTER N WITH SMALL LETTER J
     * 01F2: "\xC7\xB3" : "\xC7\xB2" or "\xC7\xB1" # LATIN CAPITAL LETTER D WITH SMALL LETTER Z
     * 0399: "\xCE\xB9" : "\xCE\x99" or "\xCD\x85" # GREEK CAPITAL LETTER IOTA
     * 039C: "\xCE\xBC" : "\xCE\x9C" or "\xC2\xB5" # GREEK CAPITAL LETTER MU
     * 03C2: "\xCF\x83" : "\xCF\x82" or "\xCE\xA3" # GREEK SMALL LETTER FINAL SIGMA
     * 03D0: "\xCE\xB2" : "\xCF\x90" or "\xCE\x92" # GREEK BETA SYMBOL
     * 03D1: "\xCE\xB8" : "\xCF\x91" or "\xCE\x98" # GREEK THETA SYMBOL
     * 03D5: "\xCF\x86" : "\xCF\x95" or "\xCE\xA6" # GREEK PHI SYMBOL
     * 03D6: "\xCF\x80" : "\xCF\x96" or "\xCE\xA0" # GREEK PI SYMBOL
     * 03F0: "\xCE\xBA" : "\xCF\xB0" or "\xCE\x9A" # GREEK KAPPA SYMBOL
     * 03F1: "\xCF\x81" : "\xCF\xB1" or "\xCE\xA1" # GREEK RHO SYMBOL
     * 03F4: "\xCE\xB8" : "\xCF\xB4" or "\xCE\x98" # GREEK CAPITAL THETA SYMBOL
     * 03F5: "\xCE\xB5" : "\xCF\xB5" or "\xCE\x95" # GREEK LUNATE EPSILON SYMBOL
     * 1E9B: "\xE1\xB9\xA1" : "\xE1\xBA\x9B" or "\xE1\xB9\xA0" # LATIN SMALL LETTER LONG S WITH DOT ABOVE
     * 1FBE: "\xCE\xB9" : "\xE1\xBE\xBE" or "\xCD\x85" # GREEK PROSGEGRAMMENI
     * 2126: "\xCF\x89" : "\xE2\x84\xA6" or "\xCE\xA9" # OHM SIGN
     * 212A: 'k' : "\xE2\x84\xAA" or 'K' # KELVIN SIGN
     * 212B: "\xC3\xA5" : "\xE2\x84\xAB" or "\xC3\x85" # ANGSTROM SIGN
     * </code>
     * This may not be perfect, but it is a reasonable approximation IMHO.
     */
    global $UTF8_LOWER_UPPER;
    if (empty($UTF8_LOWER_UPPER)) { $UTF8_LOWER_UPPER = array(
'a'=>'A','b'=>'B','c'=>'C','d'=>'D','e'=>'E','f'=>'F','g'=>'G','h'=>'H','i'=>'I','j'=>'J','k'=>'K','l'=>'L','m'=>'M',
'n'=>'N','o'=>'O','p'=>'P','q'=>'Q','r'=>'R','s'=>'S','t'=>'T','u'=>'U','v'=>'V','w'=>'W','x'=>'X','y'=>'Y','z'=>'Z',
"\xCE\xBC"=>"\xC2\xB5","\xC3\xA0"=>"\xC3\x80","\xC3\xA1"=>"\xC3\x81","\xC3\xA2"=>"\xC3\x82","\xC3\xA3"=>"\xC3\x83",
"\xC3\xA4"=>"\xC3\x84","\xC3\xA5"=>"\xC3\x85","\xC3\xA6"=>"\xC3\x86","\xC3\xA7"=>"\xC3\x87","\xC3\xA8"=>"\xC3\x88",
"\xC3\xA9"=>"\xC3\x89","\xC3\xAA"=>"\xC3\x8A","\xC3\xAB"=>"\xC3\x8B","\xC3\xAC"=>"\xC3\x8C","\xC3\xAD"=>"\xC3\x8D",
"\xC3\xAE"=>"\xC3\x8E","\xC3\xAF"=>"\xC3\x8F","\xC3\xB0"=>"\xC3\x90","\xC3\xB1"=>"\xC3\x91","\xC3\xB2"=>"\xC3\x92",
"\xC3\xB3"=>"\xC3\x93","\xC3\xB4"=>"\xC3\x94","\xC3\xB5"=>"\xC3\x95","\xC3\xB6"=>"\xC3\x96","\xC3\xB8"=>"\xC3\x98",
"\xC3\xB9"=>"\xC3\x99","\xC3\xBA"=>"\xC3\x9A","\xC3\xBB"=>"\xC3\x9B","\xC3\xBC"=>"\xC3\x9C","\xC3\xBD"=>"\xC3\x9D",
"\xC3\xBE"=>"\xC3\x9E","\xC4\x81"=>"\xC4\x80","\xC4\x83"=>"\xC4\x82","\xC4\x85"=>"\xC4\x84","\xC4\x87"=>"\xC4\x86",
"\xC4\x89"=>"\xC4\x88","\xC4\x8B"=>"\xC4\x8A","\xC4\x8D"=>"\xC4\x8C","\xC4\x8F"=>"\xC4\x8E","\xC4\x91"=>"\xC4\x90",
"\xC4\x93"=>"\xC4\x92","\xC4\x95"=>"\xC4\x94","\xC4\x97"=>"\xC4\x96","\xC4\x99"=>"\xC4\x98","\xC4\x9B"=>"\xC4\x9A",
"\xC4\x9D"=>"\xC4\x9C","\xC4\x9F"=>"\xC4\x9E","\xC4\xA1"=>"\xC4\xA0","\xC4\xA3"=>"\xC4\xA2","\xC4\xA5"=>"\xC4\xA4",
"\xC4\xA7"=>"\xC4\xA6","\xC4\xA9"=>"\xC4\xA8","\xC4\xAB"=>"\xC4\xAA","\xC4\xAD"=>"\xC4\xAC","\xC4\xAF"=>"\xC4\xAE",
"\xC4\xB3"=>"\xC4\xB2","\xC4\xB5"=>"\xC4\xB4","\xC4\xB7"=>"\xC4\xB6","\xC4\xBA"=>"\xC4\xB9","\xC4\xBC"=>"\xC4\xBB",
"\xC4\xBE"=>"\xC4\xBD","\xC5\x80"=>"\xC4\xBF","\xC5\x82"=>"\xC5\x81","\xC5\x84"=>"\xC5\x83","\xC5\x86"=>"\xC5\x85",
"\xC5\x88"=>"\xC5\x87","\xC5\x8B"=>"\xC5\x8A","\xC5\x8D"=>"\xC5\x8C","\xC5\x8F"=>"\xC5\x8E","\xC5\x91"=>"\xC5\x90",
"\xC5\x93"=>"\xC5\x92","\xC5\x95"=>"\xC5\x94","\xC5\x97"=>"\xC5\x96","\xC5\x99"=>"\xC5\x98","\xC5\x9B"=>"\xC5\x9A",
"\xC5\x9D"=>"\xC5\x9C","\xC5\x9F"=>"\xC5\x9E","\xC5\xA1"=>"\xC5\xA0","\xC5\xA3"=>"\xC5\xA2","\xC5\xA5"=>"\xC5\xA4",
"\xC5\xA7"=>"\xC5\xA6","\xC5\xA9"=>"\xC5\xA8","\xC5\xAB"=>"\xC5\xAA","\xC5\xAD"=>"\xC5\xAC","\xC5\xAF"=>"\xC5\xAE",
"\xC5\xB1"=>"\xC5\xB0","\xC5\xB3"=>"\xC5\xB2","\xC5\xB5"=>"\xC5\xB4","\xC5\xB7"=>"\xC5\xB6","\xC3\xBF"=>"\xC5\xB8",
"\xC5\xBA"=>"\xC5\xB9","\xC5\xBC"=>"\xC5\xBB","\xC5\xBE"=>"\xC5\xBD","\xC9\x93"=>"\xC6\x81","\xC6\x83"=>"\xC6\x82",
"\xC6\x85"=>"\xC6\x84","\xC9\x94"=>"\xC6\x86","\xC6\x88"=>"\xC6\x87","\xC9\x96"=>"\xC6\x89","\xC9\x97"=>"\xC6\x8A",
"\xC6\x8C"=>"\xC6\x8B","\xC7\x9D"=>"\xC6\x8E","\xC9\x99"=>"\xC6\x8F","\xC9\x9B"=>"\xC6\x90","\xC6\x92"=>"\xC6\x91",
"\xC9\xA0"=>"\xC6\x93","\xC9\xA3"=>"\xC6\x94","\xC9\xA9"=>"\xC6\x96","\xC9\xA8"=>"\xC6\x97","\xC6\x99"=>"\xC6\x98",
"\xC9\xAF"=>"\xC6\x9C","\xC9\xB2"=>"\xC6\x9D","\xC9\xB5"=>"\xC6\x9F","\xC6\xA1"=>"\xC6\xA0","\xC6\xA3"=>"\xC6\xA2",
"\xC6\xA5"=>"\xC6\xA4","\xCA\x80"=>"\xC6\xA6","\xC6\xA8"=>"\xC6\xA7","\xCA\x83"=>"\xC6\xA9","\xC6\xAD"=>"\xC6\xAC",
"\xCA\x88"=>"\xC6\xAE","\xC6\xB0"=>"\xC6\xAF","\xCA\x8A"=>"\xC6\xB1","\xCA\x8B"=>"\xC6\xB2","\xC6\xB4"=>"\xC6\xB3",
"\xC6\xB6"=>"\xC6\xB5","\xCA\x92"=>"\xC6\xB7","\xC6\xB9"=>"\xC6\xB8","\xC6\xBD"=>"\xC6\xBC","\xC7\x86"=>"\xC7\x84",
"\xC7\x89"=>"\xC7\x87","\xC7\x8C"=>"\xC7\x8A","\xC7\x8E"=>"\xC7\x8D","\xC7\x90"=>"\xC7\x8F","\xC7\x92"=>"\xC7\x91",
"\xC7\x94"=>"\xC7\x93","\xC7\x96"=>"\xC7\x95","\xC7\x98"=>"\xC7\x97","\xC7\x9A"=>"\xC7\x99","\xC7\x9C"=>"\xC7\x9B",
"\xC7\x9F"=>"\xC7\x9E","\xC7\xA1"=>"\xC7\xA0","\xC7\xA3"=>"\xC7\xA2","\xC7\xA5"=>"\xC7\xA4","\xC7\xA7"=>"\xC7\xA6",
"\xC7\xA9"=>"\xC7\xA8","\xC7\xAB"=>"\xC7\xAA","\xC7\xAD"=>"\xC7\xAC","\xC7\xAF"=>"\xC7\xAE","\xC7\xB3"=>"\xC7\xB1",
"\xC7\xB5"=>"\xC7\xB4","\xC6\x95"=>"\xC7\xB6","\xC6\xBF"=>"\xC7\xB7","\xC7\xB9"=>"\xC7\xB8","\xC7\xBB"=>"\xC7\xBA",
"\xC7\xBD"=>"\xC7\xBC","\xC7\xBF"=>"\xC7\xBE","\xC8\x81"=>"\xC8\x80","\xC8\x83"=>"\xC8\x82","\xC8\x85"=>"\xC8\x84",
"\xC8\x87"=>"\xC8\x86","\xC8\x89"=>"\xC8\x88","\xC8\x8B"=>"\xC8\x8A","\xC8\x8D"=>"\xC8\x8C","\xC8\x8F"=>"\xC8\x8E",
"\xC8\x91"=>"\xC8\x90","\xC8\x93"=>"\xC8\x92","\xC8\x95"=>"\xC8\x94","\xC8\x97"=>"\xC8\x96","\xC8\x99"=>"\xC8\x98",
"\xC8\x9B"=>"\xC8\x9A","\xC8\x9D"=>"\xC8\x9C","\xC8\x9F"=>"\xC8\x9E","\xC6\x9E"=>"\xC8\xA0","\xC8\xA3"=>"\xC8\xA2",
"\xC8\xA5"=>"\xC8\xA4","\xC8\xA7"=>"\xC8\xA6","\xC8\xA9"=>"\xC8\xA8","\xC8\xAB"=>"\xC8\xAA","\xC8\xAD"=>"\xC8\xAC",
"\xC8\xAF"=>"\xC8\xAE","\xC8\xB1"=>"\xC8\xB0","\xC8\xB3"=>"\xC8\xB2","\xE2\xB1\xA5"=>"\xC8\xBA","\xC8\xBC"=>"\xC8\xBB",
"\xC6\x9A"=>"\xC8\xBD","\xE2\xB1\xA6"=>"\xC8\xBE","\xC9\x82"=>"\xC9\x81","\xC6\x80"=>"\xC9\x83","\xCA\x89"=>"\xC9\x84",
"\xCA\x8C"=>"\xC9\x85","\xC9\x87"=>"\xC9\x86","\xC9\x89"=>"\xC9\x88","\xC9\x8B"=>"\xC9\x8A","\xC9\x8D"=>"\xC9\x8C",
"\xC9\x8F"=>"\xC9\x8E","\xCE\xB9"=>"\xCD\x85","\xCD\xB1"=>"\xCD\xB0","\xCD\xB3"=>"\xCD\xB2","\xCD\xB7"=>"\xCD\xB6",
"\xCE\xAC"=>"\xCE\x86","\xCE\xAD"=>"\xCE\x88","\xCE\xAE"=>"\xCE\x89","\xCE\xAF"=>"\xCE\x8A","\xCF\x8C"=>"\xCE\x8C",
"\xCF\x8D"=>"\xCE\x8E","\xCF\x8E"=>"\xCE\x8F","\xCE\xB1"=>"\xCE\x91","\xCE\xB2"=>"\xCE\x92","\xCE\xB3"=>"\xCE\x93",
"\xCE\xB4"=>"\xCE\x94","\xCE\xB5"=>"\xCE\x95","\xCE\xB6"=>"\xCE\x96","\xCE\xB7"=>"\xCE\x97","\xCE\xB8"=>"\xCE\x98",
"\xCE\xBA"=>"\xCE\x9A","\xCE\xBB"=>"\xCE\x9B","\xCE\xBD"=>"\xCE\x9D","\xCE\xBE"=>"\xCE\x9E","\xCE\xBF"=>"\xCE\x9F",
"\xCF\x80"=>"\xCE\xA0","\xCF\x81"=>"\xCE\xA1","\xCF\x83"=>"\xCE\xA3","\xCF\x84"=>"\xCE\xA4","\xCF\x85"=>"\xCE\xA5",
"\xCF\x86"=>"\xCE\xA6","\xCF\x87"=>"\xCE\xA7","\xCF\x88"=>"\xCE\xA8","\xCF\x89"=>"\xCE\xA9","\xCF\x8A"=>"\xCE\xAA",
"\xCF\x8B"=>"\xCE\xAB","\xCF\x97"=>"\xCF\x8F","\xCF\x99"=>"\xCF\x98","\xCF\x9B"=>"\xCF\x9A","\xCF\x9D"=>"\xCF\x9C",
"\xCF\x9F"=>"\xCF\x9E","\xCF\xA1"=>"\xCF\xA0","\xCF\xA3"=>"\xCF\xA2","\xCF\xA5"=>"\xCF\xA4","\xCF\xA7"=>"\xCF\xA6",
"\xCF\xA9"=>"\xCF\xA8","\xCF\xAB"=>"\xCF\xAA","\xCF\xAD"=>"\xCF\xAC","\xCF\xAF"=>"\xCF\xAE","\xCF\xB8"=>"\xCF\xB7",
"\xCF\xB2"=>"\xCF\xB9","\xCF\xBB"=>"\xCF\xBA","\xCD\xBB"=>"\xCF\xBD","\xCD\xBC"=>"\xCF\xBE","\xCD\xBD"=>"\xCF\xBF",
"\xD1\x90"=>"\xD0\x80","\xD1\x91"=>"\xD0\x81","\xD1\x92"=>"\xD0\x82","\xD1\x93"=>"\xD0\x83","\xD1\x94"=>"\xD0\x84",
"\xD1\x95"=>"\xD0\x85","\xD1\x96"=>"\xD0\x86","\xD1\x97"=>"\xD0\x87","\xD1\x98"=>"\xD0\x88","\xD1\x99"=>"\xD0\x89",
"\xD1\x9A"=>"\xD0\x8A","\xD1\x9B"=>"\xD0\x8B","\xD1\x9C"=>"\xD0\x8C","\xD1\x9D"=>"\xD0\x8D","\xD1\x9E"=>"\xD0\x8E",
"\xD1\x9F"=>"\xD0\x8F","\xD0\xB0"=>"\xD0\x90","\xD0\xB1"=>"\xD0\x91","\xD0\xB2"=>"\xD0\x92","\xD0\xB3"=>"\xD0\x93",
"\xD0\xB4"=>"\xD0\x94","\xD0\xB5"=>"\xD0\x95","\xD0\xB6"=>"\xD0\x96","\xD0\xB7"=>"\xD0\x97","\xD0\xB8"=>"\xD0\x98",
"\xD0\xB9"=>"\xD0\x99","\xD0\xBA"=>"\xD0\x9A","\xD0\xBB"=>"\xD0\x9B","\xD0\xBC"=>"\xD0\x9C","\xD0\xBD"=>"\xD0\x9D",
"\xD0\xBE"=>"\xD0\x9E","\xD0\xBF"=>"\xD0\x9F","\xD1\x80"=>"\xD0\xA0","\xD1\x81"=>"\xD0\xA1","\xD1\x82"=>"\xD0\xA2",
"\xD1\x83"=>"\xD0\xA3","\xD1\x84"=>"\xD0\xA4","\xD1\x85"=>"\xD0\xA5","\xD1\x86"=>"\xD0\xA6","\xD1\x87"=>"\xD0\xA7",
"\xD1\x88"=>"\xD0\xA8","\xD1\x89"=>"\xD0\xA9","\xD1\x8A"=>"\xD0\xAA","\xD1\x8B"=>"\xD0\xAB","\xD1\x8C"=>"\xD0\xAC",
"\xD1\x8D"=>"\xD0\xAD","\xD1\x8E"=>"\xD0\xAE","\xD1\x8F"=>"\xD0\xAF","\xD1\xA1"=>"\xD1\xA0","\xD1\xA3"=>"\xD1\xA2",
"\xD1\xA5"=>"\xD1\xA4","\xD1\xA7"=>"\xD1\xA6","\xD1\xA9"=>"\xD1\xA8","\xD1\xAB"=>"\xD1\xAA","\xD1\xAD"=>"\xD1\xAC",
"\xD1\xAF"=>"\xD1\xAE","\xD1\xB1"=>"\xD1\xB0","\xD1\xB3"=>"\xD1\xB2","\xD1\xB5"=>"\xD1\xB4","\xD1\xB7"=>"\xD1\xB6",
"\xD1\xB9"=>"\xD1\xB8","\xD1\xBB"=>"\xD1\xBA","\xD1\xBD"=>"\xD1\xBC","\xD1\xBF"=>"\xD1\xBE","\xD2\x81"=>"\xD2\x80",
"\xD2\x8B"=>"\xD2\x8A","\xD2\x8D"=>"\xD2\x8C","\xD2\x8F"=>"\xD2\x8E","\xD2\x91"=>"\xD2\x90","\xD2\x93"=>"\xD2\x92",
"\xD2\x95"=>"\xD2\x94","\xD2\x97"=>"\xD2\x96","\xD2\x99"=>"\xD2\x98","\xD2\x9B"=>"\xD2\x9A","\xD2\x9D"=>"\xD2\x9C",
"\xD2\x9F"=>"\xD2\x9E","\xD2\xA1"=>"\xD2\xA0","\xD2\xA3"=>"\xD2\xA2","\xD2\xA5"=>"\xD2\xA4","\xD2\xA7"=>"\xD2\xA6",
"\xD2\xA9"=>"\xD2\xA8","\xD2\xAB"=>"\xD2\xAA","\xD2\xAD"=>"\xD2\xAC","\xD2\xAF"=>"\xD2\xAE","\xD2\xB1"=>"\xD2\xB0",
"\xD2\xB3"=>"\xD2\xB2","\xD2\xB5"=>"\xD2\xB4","\xD2\xB7"=>"\xD2\xB6","\xD2\xB9"=>"\xD2\xB8","\xD2\xBB"=>"\xD2\xBA",
"\xD2\xBD"=>"\xD2\xBC","\xD2\xBF"=>"\xD2\xBE","\xD3\x8F"=>"\xD3\x80","\xD3\x82"=>"\xD3\x81","\xD3\x84"=>"\xD3\x83",
"\xD3\x86"=>"\xD3\x85","\xD3\x88"=>"\xD3\x87","\xD3\x8A"=>"\xD3\x89","\xD3\x8C"=>"\xD3\x8B","\xD3\x8E"=>"\xD3\x8D",
"\xD3\x91"=>"\xD3\x90","\xD3\x93"=>"\xD3\x92","\xD3\x95"=>"\xD3\x94","\xD3\x97"=>"\xD3\x96","\xD3\x99"=>"\xD3\x98",
"\xD3\x9B"=>"\xD3\x9A","\xD3\x9D"=>"\xD3\x9C","\xD3\x9F"=>"\xD3\x9E","\xD3\xA1"=>"\xD3\xA0","\xD3\xA3"=>"\xD3\xA2",
"\xD3\xA5"=>"\xD3\xA4","\xD3\xA7"=>"\xD3\xA6","\xD3\xA9"=>"\xD3\xA8","\xD3\xAB"=>"\xD3\xAA","\xD3\xAD"=>"\xD3\xAC",
"\xD3\xAF"=>"\xD3\xAE","\xD3\xB1"=>"\xD3\xB0","\xD3\xB3"=>"\xD3\xB2","\xD3\xB5"=>"\xD3\xB4","\xD3\xB7"=>"\xD3\xB6",
"\xD3\xB9"=>"\xD3\xB8","\xD3\xBB"=>"\xD3\xBA","\xD3\xBD"=>"\xD3\xBC","\xD3\xBF"=>"\xD3\xBE","\xD4\x81"=>"\xD4\x80",
"\xD4\x83"=>"\xD4\x82","\xD4\x85"=>"\xD4\x84","\xD4\x87"=>"\xD4\x86","\xD4\x89"=>"\xD4\x88","\xD4\x8B"=>"\xD4\x8A",
"\xD4\x8D"=>"\xD4\x8C","\xD4\x8F"=>"\xD4\x8E","\xD4\x91"=>"\xD4\x90","\xD4\x93"=>"\xD4\x92","\xD4\x95"=>"\xD4\x94",
"\xD4\x97"=>"\xD4\x96","\xD4\x99"=>"\xD4\x98","\xD4\x9B"=>"\xD4\x9A","\xD4\x9D"=>"\xD4\x9C","\xD4\x9F"=>"\xD4\x9E",
"\xD4\xA1"=>"\xD4\xA0","\xD4\xA3"=>"\xD4\xA2","\xD4\xA5"=>"\xD4\xA4","\xD4\xA7"=>"\xD4\xA6","\xD5\xA1"=>"\xD4\xB1",
"\xD5\xA2"=>"\xD4\xB2","\xD5\xA3"=>"\xD4\xB3","\xD5\xA4"=>"\xD4\xB4","\xD5\xA5"=>"\xD4\xB5","\xD5\xA6"=>"\xD4\xB6",
"\xD5\xA7"=>"\xD4\xB7","\xD5\xA8"=>"\xD4\xB8","\xD5\xA9"=>"\xD4\xB9","\xD5\xAA"=>"\xD4\xBA","\xD5\xAB"=>"\xD4\xBB",
"\xD5\xAC"=>"\xD4\xBC","\xD5\xAD"=>"\xD4\xBD","\xD5\xAE"=>"\xD4\xBE","\xD5\xAF"=>"\xD4\xBF","\xD5\xB0"=>"\xD5\x80",
"\xD5\xB1"=>"\xD5\x81","\xD5\xB2"=>"\xD5\x82","\xD5\xB3"=>"\xD5\x83","\xD5\xB4"=>"\xD5\x84","\xD5\xB5"=>"\xD5\x85",
"\xD5\xB6"=>"\xD5\x86","\xD5\xB7"=>"\xD5\x87","\xD5\xB8"=>"\xD5\x88","\xD5\xB9"=>"\xD5\x89","\xD5\xBA"=>"\xD5\x8A",
"\xD5\xBB"=>"\xD5\x8B","\xD5\xBC"=>"\xD5\x8C","\xD5\xBD"=>"\xD5\x8D","\xD5\xBE"=>"\xD5\x8E","\xD5\xBF"=>"\xD5\x8F",
"\xD6\x80"=>"\xD5\x90","\xD6\x81"=>"\xD5\x91","\xD6\x82"=>"\xD5\x92","\xD6\x83"=>"\xD5\x93","\xD6\x84"=>"\xD5\x94",
"\xD6\x85"=>"\xD5\x95","\xD6\x86"=>"\xD5\x96","\xE2\xB4\x80"=>"\xE1\x82\xA0","\xE2\xB4\x81"=>"\xE1\x82\xA1",
"\xE2\xB4\x82"=>"\xE1\x82\xA2","\xE2\xB4\x83"=>"\xE1\x82\xA3","\xE2\xB4\x84"=>"\xE1\x82\xA4",
"\xE2\xB4\x85"=>"\xE1\x82\xA5","\xE2\xB4\x86"=>"\xE1\x82\xA6","\xE2\xB4\x87"=>"\xE1\x82\xA7",
"\xE2\xB4\x88"=>"\xE1\x82\xA8","\xE2\xB4\x89"=>"\xE1\x82\xA9","\xE2\xB4\x8A"=>"\xE1\x82\xAA",
"\xE2\xB4\x8B"=>"\xE1\x82\xAB","\xE2\xB4\x8C"=>"\xE1\x82\xAC","\xE2\xB4\x8D"=>"\xE1\x82\xAD",
"\xE2\xB4\x8E"=>"\xE1\x82\xAE","\xE2\xB4\x8F"=>"\xE1\x82\xAF","\xE2\xB4\x90"=>"\xE1\x82\xB0",
"\xE2\xB4\x91"=>"\xE1\x82\xB1","\xE2\xB4\x92"=>"\xE1\x82\xB2","\xE2\xB4\x93"=>"\xE1\x82\xB3",
"\xE2\xB4\x94"=>"\xE1\x82\xB4","\xE2\xB4\x95"=>"\xE1\x82\xB5","\xE2\xB4\x96"=>"\xE1\x82\xB6",
"\xE2\xB4\x97"=>"\xE1\x82\xB7","\xE2\xB4\x98"=>"\xE1\x82\xB8","\xE2\xB4\x99"=>"\xE1\x82\xB9",
"\xE2\xB4\x9A"=>"\xE1\x82\xBA","\xE2\xB4\x9B"=>"\xE1\x82\xBB","\xE2\xB4\x9C"=>"\xE1\x82\xBC",
"\xE2\xB4\x9D"=>"\xE1\x82\xBD","\xE2\xB4\x9E"=>"\xE1\x82\xBE","\xE2\xB4\x9F"=>"\xE1\x82\xBF",
"\xE2\xB4\xA0"=>"\xE1\x83\x80","\xE2\xB4\xA1"=>"\xE1\x83\x81","\xE2\xB4\xA2"=>"\xE1\x83\x82",
"\xE2\xB4\xA3"=>"\xE1\x83\x83","\xE2\xB4\xA4"=>"\xE1\x83\x84","\xE2\xB4\xA5"=>"\xE1\x83\x85",
"\xE1\xB8\x81"=>"\xE1\xB8\x80","\xE1\xB8\x83"=>"\xE1\xB8\x82","\xE1\xB8\x85"=>"\xE1\xB8\x84",
"\xE1\xB8\x87"=>"\xE1\xB8\x86","\xE1\xB8\x89"=>"\xE1\xB8\x88","\xE1\xB8\x8B"=>"\xE1\xB8\x8A",
"\xE1\xB8\x8D"=>"\xE1\xB8\x8C","\xE1\xB8\x8F"=>"\xE1\xB8\x8E","\xE1\xB8\x91"=>"\xE1\xB8\x90",
"\xE1\xB8\x93"=>"\xE1\xB8\x92","\xE1\xB8\x95"=>"\xE1\xB8\x94","\xE1\xB8\x97"=>"\xE1\xB8\x96",
"\xE1\xB8\x99"=>"\xE1\xB8\x98","\xE1\xB8\x9B"=>"\xE1\xB8\x9A","\xE1\xB8\x9D"=>"\xE1\xB8\x9C",
"\xE1\xB8\x9F"=>"\xE1\xB8\x9E","\xE1\xB8\xA1"=>"\xE1\xB8\xA0","\xE1\xB8\xA3"=>"\xE1\xB8\xA2",
"\xE1\xB8\xA5"=>"\xE1\xB8\xA4","\xE1\xB8\xA7"=>"\xE1\xB8\xA6","\xE1\xB8\xA9"=>"\xE1\xB8\xA8",
"\xE1\xB8\xAB"=>"\xE1\xB8\xAA","\xE1\xB8\xAD"=>"\xE1\xB8\xAC","\xE1\xB8\xAF"=>"\xE1\xB8\xAE",
"\xE1\xB8\xB1"=>"\xE1\xB8\xB0","\xE1\xB8\xB3"=>"\xE1\xB8\xB2","\xE1\xB8\xB5"=>"\xE1\xB8\xB4",
"\xE1\xB8\xB7"=>"\xE1\xB8\xB6","\xE1\xB8\xB9"=>"\xE1\xB8\xB8","\xE1\xB8\xBB"=>"\xE1\xB8\xBA",
"\xE1\xB8\xBD"=>"\xE1\xB8\xBC","\xE1\xB8\xBF"=>"\xE1\xB8\xBE","\xE1\xB9\x81"=>"\xE1\xB9\x80",
"\xE1\xB9\x83"=>"\xE1\xB9\x82","\xE1\xB9\x85"=>"\xE1\xB9\x84","\xE1\xB9\x87"=>"\xE1\xB9\x86",
"\xE1\xB9\x89"=>"\xE1\xB9\x88","\xE1\xB9\x8B"=>"\xE1\xB9\x8A","\xE1\xB9\x8D"=>"\xE1\xB9\x8C",
"\xE1\xB9\x8F"=>"\xE1\xB9\x8E","\xE1\xB9\x91"=>"\xE1\xB9\x90","\xE1\xB9\x93"=>"\xE1\xB9\x92",
"\xE1\xB9\x95"=>"\xE1\xB9\x94","\xE1\xB9\x97"=>"\xE1\xB9\x96","\xE1\xB9\x99"=>"\xE1\xB9\x98",
"\xE1\xB9\x9B"=>"\xE1\xB9\x9A","\xE1\xB9\x9D"=>"\xE1\xB9\x9C","\xE1\xB9\x9F"=>"\xE1\xB9\x9E",
"\xE1\xB9\xA1"=>"\xE1\xB9\xA0","\xE1\xB9\xA3"=>"\xE1\xB9\xA2","\xE1\xB9\xA5"=>"\xE1\xB9\xA4",
"\xE1\xB9\xA7"=>"\xE1\xB9\xA6","\xE1\xB9\xA9"=>"\xE1\xB9\xA8","\xE1\xB9\xAB"=>"\xE1\xB9\xAA",
"\xE1\xB9\xAD"=>"\xE1\xB9\xAC","\xE1\xB9\xAF"=>"\xE1\xB9\xAE","\xE1\xB9\xB1"=>"\xE1\xB9\xB0",
"\xE1\xB9\xB3"=>"\xE1\xB9\xB2","\xE1\xB9\xB5"=>"\xE1\xB9\xB4","\xE1\xB9\xB7"=>"\xE1\xB9\xB6",
"\xE1\xB9\xB9"=>"\xE1\xB9\xB8","\xE1\xB9\xBB"=>"\xE1\xB9\xBA","\xE1\xB9\xBD"=>"\xE1\xB9\xBC",
"\xE1\xB9\xBF"=>"\xE1\xB9\xBE","\xE1\xBA\x81"=>"\xE1\xBA\x80","\xE1\xBA\x83"=>"\xE1\xBA\x82",
"\xE1\xBA\x85"=>"\xE1\xBA\x84","\xE1\xBA\x87"=>"\xE1\xBA\x86","\xE1\xBA\x89"=>"\xE1\xBA\x88",
"\xE1\xBA\x8B"=>"\xE1\xBA\x8A","\xE1\xBA\x8D"=>"\xE1\xBA\x8C","\xE1\xBA\x8F"=>"\xE1\xBA\x8E",
"\xE1\xBA\x91"=>"\xE1\xBA\x90","\xE1\xBA\x93"=>"\xE1\xBA\x92","\xE1\xBA\x95"=>"\xE1\xBA\x94",
"\xC3\x9F"=>"\xE1\xBA\x9E","\xE1\xBA\xA1"=>"\xE1\xBA\xA0","\xE1\xBA\xA3"=>"\xE1\xBA\xA2",
"\xE1\xBA\xA5"=>"\xE1\xBA\xA4","\xE1\xBA\xA7"=>"\xE1\xBA\xA6","\xE1\xBA\xA9"=>"\xE1\xBA\xA8",
"\xE1\xBA\xAB"=>"\xE1\xBA\xAA","\xE1\xBA\xAD"=>"\xE1\xBA\xAC","\xE1\xBA\xAF"=>"\xE1\xBA\xAE",
"\xE1\xBA\xB1"=>"\xE1\xBA\xB0","\xE1\xBA\xB3"=>"\xE1\xBA\xB2","\xE1\xBA\xB5"=>"\xE1\xBA\xB4",
"\xE1\xBA\xB7"=>"\xE1\xBA\xB6","\xE1\xBA\xB9"=>"\xE1\xBA\xB8","\xE1\xBA\xBB"=>"\xE1\xBA\xBA",
"\xE1\xBA\xBD"=>"\xE1\xBA\xBC","\xE1\xBA\xBF"=>"\xE1\xBA\xBE","\xE1\xBB\x81"=>"\xE1\xBB\x80",
"\xE1\xBB\x83"=>"\xE1\xBB\x82","\xE1\xBB\x85"=>"\xE1\xBB\x84","\xE1\xBB\x87"=>"\xE1\xBB\x86",
"\xE1\xBB\x89"=>"\xE1\xBB\x88","\xE1\xBB\x8B"=>"\xE1\xBB\x8A","\xE1\xBB\x8D"=>"\xE1\xBB\x8C",
"\xE1\xBB\x8F"=>"\xE1\xBB\x8E","\xE1\xBB\x91"=>"\xE1\xBB\x90","\xE1\xBB\x93"=>"\xE1\xBB\x92",
"\xE1\xBB\x95"=>"\xE1\xBB\x94","\xE1\xBB\x97"=>"\xE1\xBB\x96","\xE1\xBB\x99"=>"\xE1\xBB\x98",
"\xE1\xBB\x9B"=>"\xE1\xBB\x9A","\xE1\xBB\x9D"=>"\xE1\xBB\x9C","\xE1\xBB\x9F"=>"\xE1\xBB\x9E",
"\xE1\xBB\xA1"=>"\xE1\xBB\xA0","\xE1\xBB\xA3"=>"\xE1\xBB\xA2","\xE1\xBB\xA5"=>"\xE1\xBB\xA4",
"\xE1\xBB\xA7"=>"\xE1\xBB\xA6","\xE1\xBB\xA9"=>"\xE1\xBB\xA8","\xE1\xBB\xAB"=>"\xE1\xBB\xAA",
"\xE1\xBB\xAD"=>"\xE1\xBB\xAC","\xE1\xBB\xAF"=>"\xE1\xBB\xAE","\xE1\xBB\xB1"=>"\xE1\xBB\xB0",
"\xE1\xBB\xB3"=>"\xE1\xBB\xB2","\xE1\xBB\xB5"=>"\xE1\xBB\xB4","\xE1\xBB\xB7"=>"\xE1\xBB\xB6",
"\xE1\xBB\xB9"=>"\xE1\xBB\xB8","\xE1\xBB\xBB"=>"\xE1\xBB\xBA","\xE1\xBB\xBD"=>"\xE1\xBB\xBC",
"\xE1\xBB\xBF"=>"\xE1\xBB\xBE","\xE1\xBC\x80"=>"\xE1\xBC\x88","\xE1\xBC\x81"=>"\xE1\xBC\x89",
"\xE1\xBC\x82"=>"\xE1\xBC\x8A","\xE1\xBC\x83"=>"\xE1\xBC\x8B","\xE1\xBC\x84"=>"\xE1\xBC\x8C",
"\xE1\xBC\x85"=>"\xE1\xBC\x8D","\xE1\xBC\x86"=>"\xE1\xBC\x8E","\xE1\xBC\x87"=>"\xE1\xBC\x8F",
"\xE1\xBC\x90"=>"\xE1\xBC\x98","\xE1\xBC\x91"=>"\xE1\xBC\x99","\xE1\xBC\x92"=>"\xE1\xBC\x9A",
"\xE1\xBC\x93"=>"\xE1\xBC\x9B","\xE1\xBC\x94"=>"\xE1\xBC\x9C","\xE1\xBC\x95"=>"\xE1\xBC\x9D",
"\xE1\xBC\xA0"=>"\xE1\xBC\xA8","\xE1\xBC\xA1"=>"\xE1\xBC\xA9","\xE1\xBC\xA2"=>"\xE1\xBC\xAA",
"\xE1\xBC\xA3"=>"\xE1\xBC\xAB","\xE1\xBC\xA4"=>"\xE1\xBC\xAC","\xE1\xBC\xA5"=>"\xE1\xBC\xAD",
"\xE1\xBC\xA6"=>"\xE1\xBC\xAE","\xE1\xBC\xA7"=>"\xE1\xBC\xAF","\xE1\xBC\xB0"=>"\xE1\xBC\xB8",
"\xE1\xBC\xB1"=>"\xE1\xBC\xB9","\xE1\xBC\xB2"=>"\xE1\xBC\xBA","\xE1\xBC\xB3"=>"\xE1\xBC\xBB",
"\xE1\xBC\xB4"=>"\xE1\xBC\xBC","\xE1\xBC\xB5"=>"\xE1\xBC\xBD","\xE1\xBC\xB6"=>"\xE1\xBC\xBE",
"\xE1\xBC\xB7"=>"\xE1\xBC\xBF","\xE1\xBD\x80"=>"\xE1\xBD\x88","\xE1\xBD\x81"=>"\xE1\xBD\x89",
"\xE1\xBD\x82"=>"\xE1\xBD\x8A","\xE1\xBD\x83"=>"\xE1\xBD\x8B","\xE1\xBD\x84"=>"\xE1\xBD\x8C",
"\xE1\xBD\x85"=>"\xE1\xBD\x8D","\xE1\xBD\x91"=>"\xE1\xBD\x99","\xE1\xBD\x93"=>"\xE1\xBD\x9B",
"\xE1\xBD\x95"=>"\xE1\xBD\x9D","\xE1\xBD\x97"=>"\xE1\xBD\x9F","\xE1\xBD\xA0"=>"\xE1\xBD\xA8",
"\xE1\xBD\xA1"=>"\xE1\xBD\xA9","\xE1\xBD\xA2"=>"\xE1\xBD\xAA","\xE1\xBD\xA3"=>"\xE1\xBD\xAB",
"\xE1\xBD\xA4"=>"\xE1\xBD\xAC","\xE1\xBD\xA5"=>"\xE1\xBD\xAD","\xE1\xBD\xA6"=>"\xE1\xBD\xAE",
"\xE1\xBD\xA7"=>"\xE1\xBD\xAF","\xE1\xBE\x80"=>"\xE1\xBE\x88","\xE1\xBE\x81"=>"\xE1\xBE\x89",
"\xE1\xBE\x82"=>"\xE1\xBE\x8A","\xE1\xBE\x83"=>"\xE1\xBE\x8B","\xE1\xBE\x84"=>"\xE1\xBE\x8C",
"\xE1\xBE\x85"=>"\xE1\xBE\x8D","\xE1\xBE\x86"=>"\xE1\xBE\x8E","\xE1\xBE\x87"=>"\xE1\xBE\x8F",
"\xE1\xBE\x90"=>"\xE1\xBE\x98","\xE1\xBE\x91"=>"\xE1\xBE\x99","\xE1\xBE\x92"=>"\xE1\xBE\x9A",
"\xE1\xBE\x93"=>"\xE1\xBE\x9B","\xE1\xBE\x94"=>"\xE1\xBE\x9C","\xE1\xBE\x95"=>"\xE1\xBE\x9D",
"\xE1\xBE\x96"=>"\xE1\xBE\x9E","\xE1\xBE\x97"=>"\xE1\xBE\x9F","\xE1\xBE\xA0"=>"\xE1\xBE\xA8",
"\xE1\xBE\xA1"=>"\xE1\xBE\xA9","\xE1\xBE\xA2"=>"\xE1\xBE\xAA","\xE1\xBE\xA3"=>"\xE1\xBE\xAB",
"\xE1\xBE\xA4"=>"\xE1\xBE\xAC","\xE1\xBE\xA5"=>"\xE1\xBE\xAD","\xE1\xBE\xA6"=>"\xE1\xBE\xAE",
"\xE1\xBE\xA7"=>"\xE1\xBE\xAF","\xE1\xBE\xB0"=>"\xE1\xBE\xB8","\xE1\xBE\xB1"=>"\xE1\xBE\xB9",
"\xE1\xBD\xB0"=>"\xE1\xBE\xBA","\xE1\xBD\xB1"=>"\xE1\xBE\xBB","\xE1\xBE\xB3"=>"\xE1\xBE\xBC",
"\xE1\xBD\xB2"=>"\xE1\xBF\x88","\xE1\xBD\xB3"=>"\xE1\xBF\x89","\xE1\xBD\xB4"=>"\xE1\xBF\x8A",
"\xE1\xBD\xB5"=>"\xE1\xBF\x8B","\xE1\xBF\x83"=>"\xE1\xBF\x8C","\xE1\xBF\x90"=>"\xE1\xBF\x98",
"\xE1\xBF\x91"=>"\xE1\xBF\x99","\xE1\xBD\xB6"=>"\xE1\xBF\x9A","\xE1\xBD\xB7"=>"\xE1\xBF\x9B",
"\xE1\xBF\xA0"=>"\xE1\xBF\xA8","\xE1\xBF\xA1"=>"\xE1\xBF\xA9","\xE1\xBD\xBA"=>"\xE1\xBF\xAA",
"\xE1\xBD\xBB"=>"\xE1\xBF\xAB","\xE1\xBF\xA5"=>"\xE1\xBF\xAC","\xE1\xBD\xB8"=>"\xE1\xBF\xB8",
"\xE1\xBD\xB9"=>"\xE1\xBF\xB9","\xE1\xBD\xBC"=>"\xE1\xBF\xBA","\xE1\xBD\xBD"=>"\xE1\xBF\xBB",
"\xE1\xBF\xB3"=>"\xE1\xBF\xBC","\xE2\x85\x8E"=>"\xE2\x84\xB2","\xE2\x85\xB0"=>"\xE2\x85\xA0",
"\xE2\x85\xB1"=>"\xE2\x85\xA1","\xE2\x85\xB2"=>"\xE2\x85\xA2","\xE2\x85\xB3"=>"\xE2\x85\xA3",
"\xE2\x85\xB4"=>"\xE2\x85\xA4","\xE2\x85\xB5"=>"\xE2\x85\xA5","\xE2\x85\xB6"=>"\xE2\x85\xA6",
"\xE2\x85\xB7"=>"\xE2\x85\xA7","\xE2\x85\xB8"=>"\xE2\x85\xA8","\xE2\x85\xB9"=>"\xE2\x85\xA9",
"\xE2\x85\xBA"=>"\xE2\x85\xAA","\xE2\x85\xBB"=>"\xE2\x85\xAB","\xE2\x85\xBC"=>"\xE2\x85\xAC",
"\xE2\x85\xBD"=>"\xE2\x85\xAD","\xE2\x85\xBE"=>"\xE2\x85\xAE","\xE2\x85\xBF"=>"\xE2\x85\xAF",
"\xE2\x86\x84"=>"\xE2\x86\x83","\xE2\x93\x90"=>"\xE2\x92\xB6","\xE2\x93\x91"=>"\xE2\x92\xB7",
"\xE2\x93\x92"=>"\xE2\x92\xB8","\xE2\x93\x93"=>"\xE2\x92\xB9","\xE2\x93\x94"=>"\xE2\x92\xBA",
"\xE2\x93\x95"=>"\xE2\x92\xBB","\xE2\x93\x96"=>"\xE2\x92\xBC","\xE2\x93\x97"=>"\xE2\x92\xBD",
"\xE2\x93\x98"=>"\xE2\x92\xBE","\xE2\x93\x99"=>"\xE2\x92\xBF","\xE2\x93\x9A"=>"\xE2\x93\x80",
"\xE2\x93\x9B"=>"\xE2\x93\x81","\xE2\x93\x9C"=>"\xE2\x93\x82","\xE2\x93\x9D"=>"\xE2\x93\x83",
"\xE2\x93\x9E"=>"\xE2\x93\x84","\xE2\x93\x9F"=>"\xE2\x93\x85","\xE2\x93\xA0"=>"\xE2\x93\x86",
"\xE2\x93\xA1"=>"\xE2\x93\x87","\xE2\x93\xA2"=>"\xE2\x93\x88","\xE2\x93\xA3"=>"\xE2\x93\x89",
"\xE2\x93\xA4"=>"\xE2\x93\x8A","\xE2\x93\xA5"=>"\xE2\x93\x8B","\xE2\x93\xA6"=>"\xE2\x93\x8C",
"\xE2\x93\xA7"=>"\xE2\x93\x8D","\xE2\x93\xA8"=>"\xE2\x93\x8E","\xE2\x93\xA9"=>"\xE2\x93\x8F",
"\xE2\xB0\xB0"=>"\xE2\xB0\x80","\xE2\xB0\xB1"=>"\xE2\xB0\x81","\xE2\xB0\xB2"=>"\xE2\xB0\x82",
"\xE2\xB0\xB3"=>"\xE2\xB0\x83","\xE2\xB0\xB4"=>"\xE2\xB0\x84","\xE2\xB0\xB5"=>"\xE2\xB0\x85",
"\xE2\xB0\xB6"=>"\xE2\xB0\x86","\xE2\xB0\xB7"=>"\xE2\xB0\x87","\xE2\xB0\xB8"=>"\xE2\xB0\x88",
"\xE2\xB0\xB9"=>"\xE2\xB0\x89","\xE2\xB0\xBA"=>"\xE2\xB0\x8A","\xE2\xB0\xBB"=>"\xE2\xB0\x8B",
"\xE2\xB0\xBC"=>"\xE2\xB0\x8C","\xE2\xB0\xBD"=>"\xE2\xB0\x8D","\xE2\xB0\xBE"=>"\xE2\xB0\x8E",
"\xE2\xB0\xBF"=>"\xE2\xB0\x8F","\xE2\xB1\x80"=>"\xE2\xB0\x90","\xE2\xB1\x81"=>"\xE2\xB0\x91",
"\xE2\xB1\x82"=>"\xE2\xB0\x92","\xE2\xB1\x83"=>"\xE2\xB0\x93","\xE2\xB1\x84"=>"\xE2\xB0\x94",
"\xE2\xB1\x85"=>"\xE2\xB0\x95","\xE2\xB1\x86"=>"\xE2\xB0\x96","\xE2\xB1\x87"=>"\xE2\xB0\x97",
"\xE2\xB1\x88"=>"\xE2\xB0\x98","\xE2\xB1\x89"=>"\xE2\xB0\x99","\xE2\xB1\x8A"=>"\xE2\xB0\x9A",
"\xE2\xB1\x8B"=>"\xE2\xB0\x9B","\xE2\xB1\x8C"=>"\xE2\xB0\x9C","\xE2\xB1\x8D"=>"\xE2\xB0\x9D",
"\xE2\xB1\x8E"=>"\xE2\xB0\x9E","\xE2\xB1\x8F"=>"\xE2\xB0\x9F","\xE2\xB1\x90"=>"\xE2\xB0\xA0",
"\xE2\xB1\x91"=>"\xE2\xB0\xA1","\xE2\xB1\x92"=>"\xE2\xB0\xA2","\xE2\xB1\x93"=>"\xE2\xB0\xA3",
"\xE2\xB1\x94"=>"\xE2\xB0\xA4","\xE2\xB1\x95"=>"\xE2\xB0\xA5","\xE2\xB1\x96"=>"\xE2\xB0\xA6",
"\xE2\xB1\x97"=>"\xE2\xB0\xA7","\xE2\xB1\x98"=>"\xE2\xB0\xA8","\xE2\xB1\x99"=>"\xE2\xB0\xA9",
"\xE2\xB1\x9A"=>"\xE2\xB0\xAA","\xE2\xB1\x9B"=>"\xE2\xB0\xAB","\xE2\xB1\x9C"=>"\xE2\xB0\xAC",
"\xE2\xB1\x9D"=>"\xE2\xB0\xAD","\xE2\xB1\x9E"=>"\xE2\xB0\xAE","\xE2\xB1\xA1"=>"\xE2\xB1\xA0",
"\xC9\xAB"=>"\xE2\xB1\xA2","\xE1\xB5\xBD"=>"\xE2\xB1\xA3","\xC9\xBD"=>"\xE2\xB1\xA4","\xE2\xB1\xA8"=>"\xE2\xB1\xA7",
"\xE2\xB1\xAA"=>"\xE2\xB1\xA9","\xE2\xB1\xAC"=>"\xE2\xB1\xAB","\xC9\x91"=>"\xE2\xB1\xAD","\xC9\xB1"=>"\xE2\xB1\xAE",
"\xC9\x90"=>"\xE2\xB1\xAF","\xC9\x92"=>"\xE2\xB1\xB0","\xE2\xB1\xB3"=>"\xE2\xB1\xB2","\xE2\xB1\xB6"=>"\xE2\xB1\xB5",
"\xC8\xBF"=>"\xE2\xB1\xBE","\xC9\x80"=>"\xE2\xB1\xBF","\xE2\xB2\x81"=>"\xE2\xB2\x80","\xE2\xB2\x83"=>"\xE2\xB2\x82",
"\xE2\xB2\x85"=>"\xE2\xB2\x84","\xE2\xB2\x87"=>"\xE2\xB2\x86","\xE2\xB2\x89"=>"\xE2\xB2\x88",
"\xE2\xB2\x8B"=>"\xE2\xB2\x8A","\xE2\xB2\x8D"=>"\xE2\xB2\x8C","\xE2\xB2\x8F"=>"\xE2\xB2\x8E",
"\xE2\xB2\x91"=>"\xE2\xB2\x90","\xE2\xB2\x93"=>"\xE2\xB2\x92","\xE2\xB2\x95"=>"\xE2\xB2\x94",
"\xE2\xB2\x97"=>"\xE2\xB2\x96","\xE2\xB2\x99"=>"\xE2\xB2\x98","\xE2\xB2\x9B"=>"\xE2\xB2\x9A",
"\xE2\xB2\x9D"=>"\xE2\xB2\x9C","\xE2\xB2\x9F"=>"\xE2\xB2\x9E","\xE2\xB2\xA1"=>"\xE2\xB2\xA0",
"\xE2\xB2\xA3"=>"\xE2\xB2\xA2","\xE2\xB2\xA5"=>"\xE2\xB2\xA4","\xE2\xB2\xA7"=>"\xE2\xB2\xA6",
"\xE2\xB2\xA9"=>"\xE2\xB2\xA8","\xE2\xB2\xAB"=>"\xE2\xB2\xAA","\xE2\xB2\xAD"=>"\xE2\xB2\xAC",
"\xE2\xB2\xAF"=>"\xE2\xB2\xAE","\xE2\xB2\xB1"=>"\xE2\xB2\xB0","\xE2\xB2\xB3"=>"\xE2\xB2\xB2",
"\xE2\xB2\xB5"=>"\xE2\xB2\xB4","\xE2\xB2\xB7"=>"\xE2\xB2\xB6","\xE2\xB2\xB9"=>"\xE2\xB2\xB8",
"\xE2\xB2\xBB"=>"\xE2\xB2\xBA","\xE2\xB2\xBD"=>"\xE2\xB2\xBC","\xE2\xB2\xBF"=>"\xE2\xB2\xBE",
"\xE2\xB3\x81"=>"\xE2\xB3\x80","\xE2\xB3\x83"=>"\xE2\xB3\x82","\xE2\xB3\x85"=>"\xE2\xB3\x84",
"\xE2\xB3\x87"=>"\xE2\xB3\x86","\xE2\xB3\x89"=>"\xE2\xB3\x88","\xE2\xB3\x8B"=>"\xE2\xB3\x8A",
"\xE2\xB3\x8D"=>"\xE2\xB3\x8C","\xE2\xB3\x8F"=>"\xE2\xB3\x8E","\xE2\xB3\x91"=>"\xE2\xB3\x90",
"\xE2\xB3\x93"=>"\xE2\xB3\x92","\xE2\xB3\x95"=>"\xE2\xB3\x94","\xE2\xB3\x97"=>"\xE2\xB3\x96",
"\xE2\xB3\x99"=>"\xE2\xB3\x98","\xE2\xB3\x9B"=>"\xE2\xB3\x9A","\xE2\xB3\x9D"=>"\xE2\xB3\x9C",
"\xE2\xB3\x9F"=>"\xE2\xB3\x9E","\xE2\xB3\xA1"=>"\xE2\xB3\xA0","\xE2\xB3\xA3"=>"\xE2\xB3\xA2",
"\xE2\xB3\xAC"=>"\xE2\xB3\xAB","\xE2\xB3\xAE"=>"\xE2\xB3\xAD","\xEA\x99\x81"=>"\xEA\x99\x80",
"\xEA\x99\x83"=>"\xEA\x99\x82","\xEA\x99\x85"=>"\xEA\x99\x84","\xEA\x99\x87"=>"\xEA\x99\x86",
"\xEA\x99\x89"=>"\xEA\x99\x88","\xEA\x99\x8B"=>"\xEA\x99\x8A","\xEA\x99\x8D"=>"\xEA\x99\x8C",
"\xEA\x99\x8F"=>"\xEA\x99\x8E","\xEA\x99\x91"=>"\xEA\x99\x90","\xEA\x99\x93"=>"\xEA\x99\x92",
"\xEA\x99\x95"=>"\xEA\x99\x94","\xEA\x99\x97"=>"\xEA\x99\x96","\xEA\x99\x99"=>"\xEA\x99\x98",
"\xEA\x99\x9B"=>"\xEA\x99\x9A","\xEA\x99\x9D"=>"\xEA\x99\x9C","\xEA\x99\x9F"=>"\xEA\x99\x9E",
"\xEA\x99\xA1"=>"\xEA\x99\xA0","\xEA\x99\xA3"=>"\xEA\x99\xA2","\xEA\x99\xA5"=>"\xEA\x99\xA4",
"\xEA\x99\xA7"=>"\xEA\x99\xA6","\xEA\x99\xA9"=>"\xEA\x99\xA8","\xEA\x99\xAB"=>"\xEA\x99\xAA",
"\xEA\x99\xAD"=>"\xEA\x99\xAC","\xEA\x9A\x81"=>"\xEA\x9A\x80","\xEA\x9A\x83"=>"\xEA\x9A\x82",
"\xEA\x9A\x85"=>"\xEA\x9A\x84","\xEA\x9A\x87"=>"\xEA\x9A\x86","\xEA\x9A\x89"=>"\xEA\x9A\x88",
"\xEA\x9A\x8B"=>"\xEA\x9A\x8A","\xEA\x9A\x8D"=>"\xEA\x9A\x8C","\xEA\x9A\x8F"=>"\xEA\x9A\x8E",
"\xEA\x9A\x91"=>"\xEA\x9A\x90","\xEA\x9A\x93"=>"\xEA\x9A\x92","\xEA\x9A\x95"=>"\xEA\x9A\x94",
"\xEA\x9A\x97"=>"\xEA\x9A\x96","\xEA\x9C\xA3"=>"\xEA\x9C\xA2","\xEA\x9C\xA5"=>"\xEA\x9C\xA4",
"\xEA\x9C\xA7"=>"\xEA\x9C\xA6","\xEA\x9C\xA9"=>"\xEA\x9C\xA8","\xEA\x9C\xAB"=>"\xEA\x9C\xAA",
"\xEA\x9C\xAD"=>"\xEA\x9C\xAC","\xEA\x9C\xAF"=>"\xEA\x9C\xAE","\xEA\x9C\xB3"=>"\xEA\x9C\xB2",
"\xEA\x9C\xB5"=>"\xEA\x9C\xB4","\xEA\x9C\xB7"=>"\xEA\x9C\xB6","\xEA\x9C\xB9"=>"\xEA\x9C\xB8",
"\xEA\x9C\xBB"=>"\xEA\x9C\xBA","\xEA\x9C\xBD"=>"\xEA\x9C\xBC","\xEA\x9C\xBF"=>"\xEA\x9C\xBE",
"\xEA\x9D\x81"=>"\xEA\x9D\x80","\xEA\x9D\x83"=>"\xEA\x9D\x82","\xEA\x9D\x85"=>"\xEA\x9D\x84",
"\xEA\x9D\x87"=>"\xEA\x9D\x86","\xEA\x9D\x89"=>"\xEA\x9D\x88","\xEA\x9D\x8B"=>"\xEA\x9D\x8A",
"\xEA\x9D\x8D"=>"\xEA\x9D\x8C","\xEA\x9D\x8F"=>"\xEA\x9D\x8E","\xEA\x9D\x91"=>"\xEA\x9D\x90",
"\xEA\x9D\x93"=>"\xEA\x9D\x92","\xEA\x9D\x95"=>"\xEA\x9D\x94","\xEA\x9D\x97"=>"\xEA\x9D\x96",
"\xEA\x9D\x99"=>"\xEA\x9D\x98","\xEA\x9D\x9B"=>"\xEA\x9D\x9A","\xEA\x9D\x9D"=>"\xEA\x9D\x9C",
"\xEA\x9D\x9F"=>"\xEA\x9D\x9E","\xEA\x9D\xA1"=>"\xEA\x9D\xA0","\xEA\x9D\xA3"=>"\xEA\x9D\xA2",
"\xEA\x9D\xA5"=>"\xEA\x9D\xA4","\xEA\x9D\xA7"=>"\xEA\x9D\xA6","\xEA\x9D\xA9"=>"\xEA\x9D\xA8",
"\xEA\x9D\xAB"=>"\xEA\x9D\xAA","\xEA\x9D\xAD"=>"\xEA\x9D\xAC","\xEA\x9D\xAF"=>"\xEA\x9D\xAE",
"\xEA\x9D\xBA"=>"\xEA\x9D\xB9","\xEA\x9D\xBC"=>"\xEA\x9D\xBB","\xE1\xB5\xB9"=>"\xEA\x9D\xBD",
"\xEA\x9D\xBF"=>"\xEA\x9D\xBE","\xEA\x9E\x81"=>"\xEA\x9E\x80","\xEA\x9E\x83"=>"\xEA\x9E\x82",
"\xEA\x9E\x85"=>"\xEA\x9E\x84","\xEA\x9E\x87"=>"\xEA\x9E\x86","\xEA\x9E\x8C"=>"\xEA\x9E\x8B",
"\xC9\xA5"=>"\xEA\x9E\x8D","\xEA\x9E\x91"=>"\xEA\x9E\x90","\xEA\x9E\xA1"=>"\xEA\x9E\xA0",
"\xEA\x9E\xA3"=>"\xEA\x9E\xA2","\xEA\x9E\xA5"=>"\xEA\x9E\xA4","\xEA\x9E\xA7"=>"\xEA\x9E\xA6",
"\xEA\x9E\xA9"=>"\xEA\x9E\xA8","\xEF\xBD\x81"=>"\xEF\xBC\xA1","\xEF\xBD\x82"=>"\xEF\xBC\xA2",
"\xEF\xBD\x83"=>"\xEF\xBC\xA3","\xEF\xBD\x84"=>"\xEF\xBC\xA4","\xEF\xBD\x85"=>"\xEF\xBC\xA5",
"\xEF\xBD\x86"=>"\xEF\xBC\xA6","\xEF\xBD\x87"=>"\xEF\xBC\xA7","\xEF\xBD\x88"=>"\xEF\xBC\xA8",
"\xEF\xBD\x89"=>"\xEF\xBC\xA9","\xEF\xBD\x8A"=>"\xEF\xBC\xAA","\xEF\xBD\x8B"=>"\xEF\xBC\xAB",
"\xEF\xBD\x8C"=>"\xEF\xBC\xAC","\xEF\xBD\x8D"=>"\xEF\xBC\xAD","\xEF\xBD\x8E"=>"\xEF\xBC\xAE",
"\xEF\xBD\x8F"=>"\xEF\xBC\xAF","\xEF\xBD\x90"=>"\xEF\xBC\xB0","\xEF\xBD\x91"=>"\xEF\xBC\xB1",
"\xEF\xBD\x92"=>"\xEF\xBC\xB2","\xEF\xBD\x93"=>"\xEF\xBC\xB3","\xEF\xBD\x94"=>"\xEF\xBC\xB4",
"\xEF\xBD\x95"=>"\xEF\xBC\xB5","\xEF\xBD\x96"=>"\xEF\xBC\xB6","\xEF\xBD\x97"=>"\xEF\xBC\xB7",
"\xEF\xBD\x98"=>"\xEF\xBC\xB8","\xEF\xBD\x99"=>"\xEF\xBC\xB9","\xEF\xBD\x9A"=>"\xEF\xBC\xBA",
"\xF0\x90\x90\xA8"=>"\xF0\x90\x90\x80","\xF0\x90\x90\xA9"=>"\xF0\x90\x90\x81","\xF0\x90\x90\xAA"=>"\xF0\x90\x90\x82",
"\xF0\x90\x90\xAB"=>"\xF0\x90\x90\x83","\xF0\x90\x90\xAC"=>"\xF0\x90\x90\x84","\xF0\x90\x90\xAD"=>"\xF0\x90\x90\x85",
"\xF0\x90\x90\xAE"=>"\xF0\x90\x90\x86","\xF0\x90\x90\xAF"=>"\xF0\x90\x90\x87","\xF0\x90\x90\xB0"=>"\xF0\x90\x90\x88",
"\xF0\x90\x90\xB1"=>"\xF0\x90\x90\x89","\xF0\x90\x90\xB2"=>"\xF0\x90\x90\x8A","\xF0\x90\x90\xB3"=>"\xF0\x90\x90\x8B",
"\xF0\x90\x90\xB4"=>"\xF0\x90\x90\x8C","\xF0\x90\x90\xB5"=>"\xF0\x90\x90\x8D","\xF0\x90\x90\xB6"=>"\xF0\x90\x90\x8E",
"\xF0\x90\x90\xB7"=>"\xF0\x90\x90\x8F","\xF0\x90\x90\xB8"=>"\xF0\x90\x90\x90","\xF0\x90\x90\xB9"=>"\xF0\x90\x90\x91",
"\xF0\x90\x90\xBA"=>"\xF0\x90\x90\x92","\xF0\x90\x90\xBB"=>"\xF0\x90\x90\x93","\xF0\x90\x90\xBC"=>"\xF0\x90\x90\x94",
"\xF0\x90\x90\xBD"=>"\xF0\x90\x90\x95","\xF0\x90\x90\xBE"=>"\xF0\x90\x90\x96","\xF0\x90\x90\xBF"=>"\xF0\x90\x90\x97",
"\xF0\x90\x91\x80"=>"\xF0\x90\x90\x98","\xF0\x90\x91\x81"=>"\xF0\x90\x90\x99","\xF0\x90\x91\x82"=>"\xF0\x90\x90\x9A",
"\xF0\x90\x91\x83"=>"\xF0\x90\x90\x9B","\xF0\x90\x91\x84"=>"\xF0\x90\x90\x9C","\xF0\x90\x91\x85"=>"\xF0\x90\x90\x9D",
"\xF0\x90\x91\x86"=>"\xF0\x90\x90\x9E","\xF0\x90\x91\x87"=>"\xF0\x90\x90\x9F","\xF0\x90\x91\x88"=>"\xF0\x90\x90\xA0",
"\xF0\x90\x91\x89"=>"\xF0\x90\x90\xA1","\xF0\x90\x91\x8A"=>"\xF0\x90\x90\xA2","\xF0\x90\x91\x8B"=>"\xF0\x90\x90\xA3",
"\xF0\x90\x91\x8C"=>"\xF0\x90\x90\xA4","\xF0\x90\x91\x8D"=>"\xF0\x90\x90\xA5","\xF0\x90\x91\x8E"=>"\xF0\x90\x90\xA6",
"\xF0\x90\x91\x8F"=>"\xF0\x90\x90\xA7");
    }
}


?>